Package: CoSMoS
Type: Package
Title: Complete Stochastic Modelling Solution
Version: 2.2.0
Date: 2026-05-07
Authors@R: c(person("Simon Michael", "Papalexiou", 
                      role = "aut"),
             person("Francesco", "Serinaldi", 
                      role = "aut"),
             person("Filip", "Strnad",
                     role = "aut"),
             person("Yannis", "Markonis", 
                      role = "aut"),
             person("Kevin", "Shook", 
                      role = c("ctb", "cre"),
                      email = "kshook@kshook.ca")
              )
Description: Makes univariate, multivariate, or random fields simulations precise and simple. Just select the desired time series or random fields’ properties and it will do the rest. CoSMoS is based on the framework described in Papalexiou (2018, <doi:10.1016/j.advwatres.2018.02.013>), extended for random fields in Papalexiou and Serinaldi (2020, <doi:10.1029/2019WR026331>), and further advanced in Papalexiou et al. (2021, <doi:10.1029/2020WR029466>) to allow fine-scale space-time simulation of storms (or even cyclone-mimicking fields).
Depends: R (>= 3.5.0), ggplot2, data.table
Imports: utils, methods, stats, grDevices, nloptr, MBA, Matrix, mAr,
        matrixcalc, mvtnorm, patchwork, animation, ggquiver, pracma,
        plot3D, Rcpp
LinkingTo: Rcpp, RcppEigen, RcppNumerical, BH
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: testthat, knitr, rmarkdown
VignetteBuilder: knitr
Author: Simon Michael Papalexiou [aut],
  Francesco Serinaldi [aut],
  Filip Strnad [aut],
  Yannis Markonis [aut],
  Kevin Shook [ctb, cre]
Maintainer: Kevin Shook <kshook@kshook.ca>
License: GPL-3
URL: https://github.com/TycheLab/CoSMoS
NeedsCompilation: yes
Packaged: 2026-05-07 18:24:04 UTC; kevin
Repository: CRAN
Date/Publication: 2026-05-07 20:02:51 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-07 22:29:00 UTC; unix
Archs: CoSMoS.so.dSYM

Package: CoSMoS
Type: Package
Title: Complete Stochastic Modelling Solution
Version: 2.2.0
Date: 2026-05-07
Authors@R: c(person("Simon Michael", "Papalexiou", 
                      role = "aut"),
             person("Francesco", "Serinaldi", 
                      role = "aut"),
             person("Filip", "Strnad",
                     role = "aut"),
             person("Yannis", "Markonis", 
                      role = "aut"),
             person("Kevin", "Shook", 
                      role = c("ctb", "cre"),
                      email = "kshook@kshook.ca")
              )
Description: Makes univariate, multivariate, or random fields simulations precise and simple. Just select the desired time series or random fields’ properties and it will do the rest. CoSMoS is based on the framework described in Papalexiou (2018, <doi:10.1016/j.advwatres.2018.02.013>), extended for random fields in Papalexiou and Serinaldi (2020, <doi:10.1029/2019WR026331>), and further advanced in Papalexiou et al. (2021, <doi:10.1029/2020WR029466>) to allow fine-scale space-time simulation of storms (or even cyclone-mimicking fields).
Depends: R (>= 3.5.0), ggplot2, data.table
Imports: utils, methods, stats, grDevices, nloptr, MBA, Matrix, mAr,
        matrixcalc, mvtnorm, patchwork, animation, ggquiver, pracma,
        plot3D, Rcpp
LinkingTo: Rcpp, RcppEigen, RcppNumerical, BH
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: testthat, knitr, rmarkdown
VignetteBuilder: knitr
Author: Simon Michael Papalexiou [aut],
  Francesco Serinaldi [aut],
  Filip Strnad [aut],
  Yannis Markonis [aut],
  Kevin Shook [ctb, cre]
Maintainer: Kevin Shook <kshook@kshook.ca>
License: GPL-3
URL: https://github.com/TycheLab/CoSMoS
NeedsCompilation: yes
Packaged: 2026-05-07 18:24:04 UTC; kevin
Repository: CRAN
Date/Publication: 2026-05-07 20:02:51 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-07 21:26:21 UTC; unix
Archs: CoSMoS.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,acti)
S3method(plot,checkTS)
S3method(plot,cosmosts)
S3method(plot,fitACS)
S3method(plot,fitDist)
export(acs)
export(actf)
export(actfInv)
export(actfdiscrete)
export(acti)
export(actpnts)
export(actpntsB6)
export(advectionF)
export(advectionFhyperbolic)
export(advectionFradial)
export(advectionFrotation)
export(advectionFspiral)
export(advectionFspiralCE)
export(advectionFuniform)
export(analyzeTS)
export(anisotropyT)
export(anisotropyTaffine)
export(anisotropyTswirl)
export(anisotropyTwave)
export(checkRF)
export(checkTS)
export(dburrIII)
export(dburrXII)
export(dgev)
export(dggamma)
export(dparetoII)
export(fitACS)
export(fitDist)
export(fitVAR)
export(fitactf)
export(generateMTS)
export(generateMTSFast)
export(generateRF)
export(generateRFFast)
export(generateTS)
export(mburrIII)
export(mburrXII)
export(mgev)
export(mggamma)
export(moments)
export(mparetoII)
export(pburrIII)
export(pburrXII)
export(pgev)
export(pggamma)
export(populationstat)
export(pparetoII)
export(qburrIII)
export(qburrXII)
export(qgev)
export(qggamma)
export(qparetoII)
export(quickTSPlot)
export(rburrIII)
export(rburrXII)
export(regenerateTS)
export(reportTS)
export(rgev)
export(rggamma)
export(rparetoII)
export(sample.moments)
export(simulateTS)
export(stcfclayton)
export(stcfgneiting14)
export(stcfgneiting16)
export(stcs)
import(animation)
import(data.table)
import(ggplot2)
import(ggquiver)
import(graphics)
import(mAr)
import(methods)
import(mvtnorm)
import(nloptr)
import(patchwork)
import(plot3D)
import(stats)
import(utils)
importFrom(MBA,mba.surf)
importFrom(Matrix,nearPD)
importFrom(Rcpp,sourceCpp)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,n2mfrow)
importFrom(matrixcalc,is.positive.definite)
importFrom(pracma,distmat)
useDynLib(CoSMoS, .registration = TRUE)

# CoSMoS <img src="man/figures/logo.png" align="right" width="120" />

[![AppVeyor build status](https://ci.appveyor.com/api/projects/status/github/strnda/CoSMoS?branch=master&svg=true)](https://ci.appveyor.com/project/strnda/CoSMoS)
[![CRAN status](https://www.r-pkg.org/badges/version/CoSMoS)](https://cran.r-project.org/package=CoSMoS)
[![CRAN_Download_Badge](http://cranlogs.r-pkg.org/badges/grand-total/CoSMoS)](https://CRAN.R-project.org/package=CoSMoS)
[![license](https://img.shields.io/badge/license-GPL3-lightgrey.svg)](https://choosealicense.com/)


## Complete Stochastic Modelling Solution
CoSMoS is an R package that makes time series generation with desired properties easy. Just choose the characteristics of the time series you want to generate, and it will do the rest.
The generated time series preserve any probability distribution and any linear autocorrelation structure. Users can generate as many and as long time series from processes such as precipitation, wind, temperature, relative humidity etc. It is based on a framework that unified, extended, and improved a modelling strategy that generates time series by transforming “parent” Gaussian time series having specific characteristics (Papalexiou, 2018).

## Install
To install the latest version of the package run:

```r
## copy-paste to get the latest version of CoSMoS

if (!require('devtools')) {install.packages('devtools'); library(devtools)} 

install_github('TycheLab/CoSMoS', upgrade = 'never')

library(CoSMoS)

?`CoSMoS-package`
```

# Funding
The package was partly funded by the Global institute for Water Security (GIWS; https://www.usask.ca/water/) and the Global Water Futures (GWF; https://gwf.usask.ca/) program.

# Authors
Coded by: Filip Strnad and Francesco Serinaldi     
Conceptual design by: Simon Michael Papalexiou     
Tested and documented by: Yannis Markonis     
Maintained by: Kevin Shook

# References
Papalexiou, S.M., 2018. Unified theory for stochastic modelling of hydroclimatic processes: Preserving marginal distributions, correlation structures, and intermittency. Advances in Water Resources 115, 234-252. https://doi.org/10.1016/j.advwatres.2018.02.013

# CoSMoS <img src="man/figures/logo.png" align="right" width="120" />

[![Travis build status](https://travis-ci.org/strnda/CoSMoS.svg?branch=master)](https://app.travis-ci.com/strnda/CoSMoS)
[![AppVeyor build status](https://ci.appveyor.com/api/projects/status/github/strnda/CoSMoS?branch=master&svg=true)](https://ci.appveyor.com/project/strnda/CoSMoS)
[![CRAN status](https://www.r-pkg.org/badges/version/CoSMoS)](https://cran.r-project.org/package=CoSMoS)
[![CRAN_Download_Badge](http://cranlogs.r-pkg.org/badges/grand-total/CoSMoS)](https://CRAN.R-project.org/package=CoSMoS)
[![license](https://img.shields.io/badge/license-GPL3-lightgrey.svg)](https://choosealicense.com/)


## Complete Stochastic Modelling Solution
CoSMoS is an R package that makes time series generation with desired properties easy. Just choose the characteristics of the time series you want to generate, and it will do the rest.
The generated time series preserve any probability distribution and any linear autocorrelation structure. Users can generate as many and as long time series from processes such as precipitation, wind, temperature, relative humidity etc. It is based on a framework that unified, extended, and improved a modelling strategy that generates time series by transforming “parent” Gaussian time series having specific characteristics (Papalexiou, 2018).

## Install
To install the latest version of the package run:

```r
## copy-paste to get the latest version of CoSMoS

if (!require('devtools')) {install.packages('devtools'); library(devtools)} 

install_github('TycheLab/CoSMoS', upgrade = 'never')

library(CoSMoS)

?`CoSMoS-package`
```

# Funding
The package was partly funded by the Global institute for Water Security (GIWS; https://www.usask.ca/water/) and the Global Water Futures (GWF; https://gwf.usask.ca/) program.

# Authors
Coded by: Filip Strnad and Francesco Serinaldi     
Conceptual design by: Simon Michael Papalexiou     
Tested and documented by: Yannis Markonis     
Maintained by: Kevin Shook

# References
Papalexiou, S.M., 2018. Unified theory for stochastic modelling of hydroclimatic processes: Preserving marginal distributions, correlation structures, and intermittency. Advances in Water Resources 115, 234-252. https://doi.org/10.1016/j.advwatres.2018.02.013

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(eval    = TRUE,
                      echo    = TRUE,
                      fig.width  = 7,
                      fig.retina = 1,
                      warning = FALSE,
                      message = FALSE)
library(CoSMoS)
library(plot3D)
library(patchwork)
ggplot2::theme_set(ggplot2::theme_light())

## ----warning=FALSE, message=FALSE---------------------------------------------
## (i) specifying the sample size
no <- 1000
## (ii) defining the type of marginal distribution and its parameters
marginaldist <- "ggamma"
param <- list(scale = 1, shape1 = .8, shape2 = .8)
## (iii) defining the desired autocorrelation
acf.my <- c(1, 0.8)
## (iv) simulating
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf.my)
## and (v) visually checking the generated time series
quickTSPlot(ggamma_sim[[1]])

## ----warning=FALSE, message=FALSE---------------------------------------------
acf <- c(1, 0.6, 0.5, 0.4, 0.3) # up to lag-4
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
quickTSPlot(ggamma_sim[[1]])

## ----warning=FALSE, message=FALSE---------------------------------------------
## specify lag
lags <- 0:10

## get the ACS
f <- acs(id = "fgn", t = lags, H = .75)
b <- acs(id = "burrXII", t = lags, scale = 1, shape1 = .6, shape2 = .4)
w <- acs(id = "weibull", t = lags, scale = 2, shape = 0.8)
p <- acs(id = "paretoII", t = lags, scale = 3, shape = 0.3)

## visualize the ACS
dta <- data.table(lags, f, b, w, p)
m.dta <- melt(data = dta, id.vars = "lags")

ggplot(data = m.dta, mapping = aes(x = lags, y = value, group = variable, colour = variable)) +
  geom_point(size = 2.5) +
  geom_line(lwd = 1) +
  scale_color_manual(
    values = c("steelblue4", "red4", "green4", "darkorange"),
    labels = c("FGN", "Burr XII", "Weibull", "Pareto II"), name = ""
  ) +
  labs(x = bquote(lag ~ tau), y = "ACS") +
  scale_x_continuous(breaks = lags)

## ----warning=FALSE, message=FALSE---------------------------------------------
acf <- acs(id = "paretoII", t = 0:30, scale = 1, shape = .75)
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
dta <- data.frame(time = 1:no, value = ggamma_sim[[1]])

quickTSPlot(dta$value)

## ----warning=FALSE, message=FALSE---------------------------------------------
my_acf <- exp(seq(0, -2, -0.1))
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = my_acf)
quickTSPlot(ggamma_sim[[1]])

## ----fig.height = 7, fig.cap = "Five simulated intermittent Generalized Gamma time series (p0 = 0.9).", warning=FALSE, message=FALSE----
prob_zero <- .9
## the argument `TSn = 5` enables the simulation of 5 timeseries.
ggamma_sim <- generateTS(
  n = no, margdist = marginaldist, margarg = param, acsvalue = acf,
  p0 = prob_zero, TSn = 5
)
plot(x = ggamma_sim, main = "") 

## ----warning=FALSE, message=FALSE---------------------------------------------
checkTS(ggamma_sim)

## ----fig.height = 7, fig.cap = "Clayton-Weibull spatiotemporal correlation structure.", warning=FALSE, message=FALSE----
d <- 51
st <- expand.grid(0:(d - 1), 0:(d - 1))

## get the STCS
wc <- stcfclayton(
  t = st[, 1], s = st[, 2], scfid = "weibull", tcfid = "weibull", copulaarg = 2,
  scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 5.1, shape = 0.8)
)

## visualize the STCS
wc.m <- matrix(data = wc, nrow = d)
j <- tail(which(wc.m[1, ] > 0.15), 1)
i <- tail(which(wc.m[, 1] > 0.15), 1)
wc.m <- wc.m[1:i, 1:j]

persp3D(
  z = wc.m, x = 1:nrow(wc.m), y = 1:ncol(wc.m),
  expand = 1, main = "", scale = TRUE, facets = TRUE,
  xlab = "Time lag", ylab = "Distance", zlab = "STCF", colkey = list(side = 4, length = 0.5),
  phi = 20, theta = 120, resfac = 10, col = gg2.col(100)
)

## ----fig.height = 5, fig.cap = "Five spatiotemporally correlated time series simulated with `generateMTS()`.", warning=FALSE, message=FALSE----
## set a sequence of hypothetical coordinates
d <- 5
coord <- cbind(runif(d) * 30, runif(d) * 30)

## compute VAR model parameters
fit <- fitVAR(
  spacepoints = coord,
  p = 4,
  margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 25, shape = 0.7),
    tcfarg = list(scale = 3.1, shape = 0.8)
  )
)

## generate correlated timeseries
sim <- generateMTS(n = 500, STmodel = fit)

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[, 1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) +
  geom_line() +
  facet_grid(rows = vars(variable), scales = "free_y")

## ----fig.height = 5, fig.cap = "Five spatiotemporally correlated time series simulated with `generateMTSFast()`.", warning=FALSE, message=FALSE----
## set a sequence of hypothetical coordinates
d <- 5
coord <- cbind(runif(d) * 30, runif(d) * 30)

## fit and generate correlated timeseries
sim <- generateMTSFast(
  n = 500,
  spacepoints = coord,
  p0 = 0.7,
  margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull",
    scfarg = list(scale = 25, shape = 0.7),
    tcfarg = list(scale = 3.1, shape = 0.8)
  )
)

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[, 1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) +
  geom_line() +
  facet_grid(rows = vars(variable), scales = "free_y")

## ----warning=FALSE, message=FALSE, error=TRUE---------------------------------
try({
## fit VAR model parameters for a 20x20 grid
fit <- fitVAR(spacepoints = 20, p = 3, margdist = "burrXII",
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton",
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2,
              scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 1.1, shape = 0.8)))

## generate isotropic random fields with nonseparable correlation structure
sim1 <- generateRF(n = 1000, STmodel = fit)

## fast simulation of isotropic random fields with separable correlation structure
sim2 <- generateRFFast(n = 1000, spacepoints = 20, p0 = 0.7, margdist = "paretoII",
                       margarg = list(scale = 1, shape = .3),
                       stcsarg = list(scfid = "weibull", tcfid = "weibull",
                       scfarg = list(scale = 20, shape = 0.7),
                       tcfarg = list(scale = 1.1, shape = 0.8)))
})

## ----fig.height = 6.5, fig.cap = "Validation of simulated random fields: empirical moments (top), correlation functions and CDF (middle), spatial snapshots (bottom).", warning=FALSE, message=FALSE----
## check random fields
checkRF(RF = sim1, nfields = 9*9, method = "stat")
checkRF(RF = sim1, nfields = 9*9, method = "statplot")
checkRF(RF = sim1, nfields = 9*9, method = "field")

## ----fig.width = 4, fig.height = 4, fig.cap = "Swirl-transformed coordinate system (b = 10, α = 1.8π).", warning=FALSE, message=FALSE----
## specify a grid of coordinates
m <- 30
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

## transform the coordinate system
at <- anisotropyTswirl(
  spacepoints = coord, x0 = floor(m / 2), y0 = floor(m / 2),
  b = 10, alpha = 1.8 * pi
)

## visualize transformed coordinate system
aux <- data.frame(lon = at[, 1], lat = at[, 2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
ggplot(aux, aes(x = lon, y = lat)) +
  geom_path(aes(group = id1)) +
  geom_path(aes(group = id2)) +
  geom_point(col = 2)

## ----fig.height = 7, fig.cap = "Simulated swirl-like (cyclone-shaped) random fields over a 30×30 grid.", warning=FALSE, message=FALSE----
## compute VAR model parameters
fit <- fitVAR(
  spacepoints = m, p = 1, margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.1, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 2, shape = 0.7), tcfarg = list(scale = 20.1, shape = 0.8)
  ),
  anisotropyid = "swirl",
  anisotropyarg = list(x0 = floor(m / 2), y0 = floor(m / 2), b = 10, alpha = 1.8 * pi)
)

## generate
set.seed(9)
sim3 <- generateRF(n = 25, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 5 * 5, method = "field")

## ----fig.height = 7, fig.cap = "Simulated Lagrangian random fields with affine anisotropy and uniform north-easterly advection.", warning=FALSE, message=FALSE----
## compute VAR model parameters
fit <- fitVAR(
  spacepoints = 30, p = 1, margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  anisotropyid = "affine",
  anisotropyarg = list(phi1 = 2., phi2 = 0.5, phi12 = 0, theta = -pi / 3),
  advectionid = "uniform", advectionarg = list(u = 2.5, v = 1.5)
)

## generate
set.seed(10) # 1 # 5
sim3 <- generateRF(n = 81, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 9 * 9, method = "field")

## ----fig.height = 3.5, fig.cap = "Bivariate samples (n = 5000) from Gaussian, Student (df = 4), and flipped Bárdossy copulas with identical linear correlation.", warning=FALSE, message=FALSE----
## set the hypothetical coordinates of two locations
coord <- matrix(rep(c(0, 1.65), times = 2), ncol = 2)

## compute VAR model parameters
## Gaussian dependence structure
fit1 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  )
)
## Student dependence structure
fit2 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  dsid = "student", dsarg = 4
)
## Bardossy dependence structure
fit3 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  dsid = "bardossyF", dsarg = 0.5
)
## generate
set.seed(10)
sim1 <- data.frame(generateRF(n = 5000, STmodel = fit1))
sim2 <- data.frame(generateRF(n = 5000, STmodel = fit2))
sim3 <- data.frame(generateRF(n = 5000, STmodel = fit3))

## visual check
sim1 <- cbind(sim1, id = "1 - Gauss copula")
sim2 <- cbind(sim2, id = "2 - Student copula")
sim3 <- cbind(sim3, id = "3 - flipped Bárdossy copula")
dta <- rbind(sim1, sim2, sim3)

ggplot(data = dta) +
  geom_point(mapping = aes(x = X1, y = X2), alpha = 0.07) +
  facet_wrap(facets = ~id, nrow = 1)

## linear correlations
pc <- round(c(
  cor.G = cor(sim1[, 1:2])[1, 2], cor.S = cor(sim2[, 1:2])[1, 2],
  cor.B = cor(sim3[, 1:2])[1, 2]
), 3)
pc

## ----fig.cap = "Observed hourly precipitation record.", warning=FALSE, message=FALSE----
data("precip")
quickTSPlot(precip$value)

## ----eval=FALSE, warning=FALSE, message=FALSE---------------------------------
# precip_ggamma <- analyzeTS(TS = precip, season = "month", dist = "ggamma",
#                            acsID = "weibull", lag.max = 12)

## ----include=FALSE------------------------------------------------------------
precip_ggamma <- readRDS("figures/precip_ggamma.rds")

## ----fig.height = 9, fig.cap = "Fitted Generalized Gamma distribution (top) and Weibull ACS (bottom) against monthly precipitation data.", warning=FALSE, message=FALSE----
reportTS(aTS = precip_ggamma, method = "dist") 
reportTS(aTS = precip_ggamma, method = "acs") 

## ----eval=FALSE, warning=FALSE, message=FALSE---------------------------------
# precip_pareto <- analyzeTS(TS = precip, season = "month", dist = "paretoII", acsID = "fgn", lag.max = 12)

## ----include=FALSE------------------------------------------------------------
precip_pareto <- readRDS("figures/precip_pareto.rds")

## ----fig.height = 9, fig.cap = "Fitted Pareto II distribution (top) and FGN ACS (bottom) against monthly precipitation data: note the poor fit relative to the Generalized Gamma / Weibull combination.", warning=FALSE, message=FALSE----
reportTS(aTS = precip_pareto, method = "dist") 
reportTS(aTS = precip_pareto, method = "acs") 

## ----fig.cap = "Observed (top) and simulated (bottom) hourly precipitation.", warning=FALSE, message=FALSE----
sim_precip <- simulateTS(
  aTS = precip_ggamma, from = as.POSIXct(x = "1978-12-01 00:00:00"),
  to = as.POSIXct(x = "2008-12-01 00:00:00")
)
dta <- precip
dta[, id := "observed"]
sim_precip[, id := "simulated"]
dta <- rbind(dta, sim_precip)

ggplot(data = dta) +
  geom_line(mapping = aes(x = date, y = value)) +
  facet_wrap(facets = ~id, ncol = 1)

## ----eval=FALSE, warning=FALSE, message=FALSE---------------------------------
# data("disch")
# str <- analyzeTS(TS = disch, dist = "lnorm", norm = "N2", acsID = "paretoII",
#                  lag.max = 20, constrain = TRUE, season = "month")

## ----include=FALSE------------------------------------------------------------
data("disch")
str <- readRDS("figures/str.rds")

## ----fig.height = 9, fig.cap = "Fitted log-normal distribution and Pareto II ACS against monthly streamflow data for Nassawango Creek.", warning=FALSE, message=FALSE----
reportTS(aTS = str) 

## ----fig.height = 3.5, fig.cap = "Observed (top) and simulated (bottom) daily streamflow for Nassawango Creek.", warning=FALSE, message=FALSE----
sim_str <- simulateTS(aTS = str)

dta <- disch
dta[, id := "observed"]
sim_str[, id := "simulated"]
dta <- rbind(dta, sim_str)

ggplot(data = dta) +
  geom_line(mapping = aes(x = date, y = value)) +
  facet_wrap(facets = ~id, ncol = 1)

#' Numerical estimation of moments
#'
#' Uses numerical integration to compute the theoretical raw or central moments
#' of the specified distribution.
#'
#' @param dist character; distribution name (e.g. \code{"norm"},
#'   \code{"paretoII"})
#' @param distarg list of distribution arguments
#' @param distbounds numeric vector of length 2; distribution bounds (default
#'   \code{c(-Inf, Inf)})
#' @param p0 numeric; probability zero (default 0)
#' @param raw logical; compute raw moments?
#' @param central logical; compute central moments?
#' @param coef logical; compute standardised coefficients (CV, skewness,
#'   kurtosis)?
#' @param order integer vector; raw moment orders (default \code{1:4})
#'
#' @return a named list with zero or more of:
#'   \describe{
#'     \item{\code{m}}{raw moments}
#'     \item{\code{mu}}{central moments}
#'     \item{\code{coefficients}}{CV, skewness, kurtosis}
#'   }
#'
#' @seealso \code{\link{sample.moments}}, \code{\link{populationstat}}
#'
#' @import stats
#' @export
#'
#' @keywords moments
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## Normal distribution
#' moments("norm", list(mean = 2, sd = 1))
#'
#' ## Pareto type II
#' moments(dist    = "paretoII",
#'         distarg = list(shape = 0.2, scale = 1))
#'
moments <- function(dist, distarg, p0 = 0, raw = TRUE, central = TRUE,
                    coef = TRUE, distbounds = c(-Inf, Inf), order = 1:4) {

  if ((central | coef) & (max(order) < 4)) order <- 1:4

  if (existsFunction(paste0("m", dist), where = asNamespace("CoSMoS"))) {
    m <- lapply(order, function(i) do.call(paste0("m", dist),
                                           args = c(list(r = i), distarg)))
  } else {
    f <- function(x, dist, q, ...) {
      .arg <- c(list(x), ...)
      x^q * do.call(paste0("d", dist), args = .arg)
    }
    m <- vector("list", max(order))
    for (i in order) {
      m[[i]] <- integrate(f = f, distbounds[1], distbounds[2],
                          q = i, distarg, dist = dist,
                          stop.on.error = FALSE)$value
    }
  }

  m        <- lapply(m, function(x) (1 - p0) * x)
  names(m) <- paste0("m", order)

  out <- list()

  if (raw) out$m <- unlist(m)

  if (central | coef) {
    mu <- data.frame(
      mu1 = m$m1,
      mu2 = m$m2 - m$m1^2,
      mu3 = 2 * m$m1^3 - 3 * m$m2 * m$m1 + m$m3,
      mu4 = -3 * m$m1^4 + 6 * m$m2 * m$m1^2 - 4 * m$m3 * m$m1 + m$m4
    )
    if (central) out$mu <- unlist(mu)
  }

  if (coef) {
    coef_df <- data.frame(
      cv = sqrt(mu$mu2) / mu$mu1,
      cs = mu$mu3 / sqrt(mu$mu2)^3,
      ck = mu$mu4 / mu$mu2^2
    )
    out$coefficients <- unlist(coef_df)
  }

  out
}


#' Sample moments
#'
#' Computes raw moments, central moments, and standardised coefficients (CV,
#' skewness, kurtosis) from a numeric sample.
#'
#' @param x numeric vector of values
#' @param na.rm logical; strip \code{NA} values before computation?
#' @inheritParams moments
#'
#' @return a named list with zero or more of:
#'   \describe{
#'     \item{\code{m}}{raw moments}
#'     \item{\code{mu}}{central moments}
#'     \item{\code{coefficients}}{CV, skewness, kurtosis}
#'   }
#'
#' @seealso \code{\link{moments}}, \code{\link{checkTS}}
#'
#' @import stats
#' @export
#'
#' @keywords moments
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' x <- rnorm(1000)
#' sample.moments(x)
#'
#' y <- rparetoII(1000, 10, .1)
#' sample.moments(y)
#'
sample.moments <- function(x, na.rm = FALSE, raw = TRUE, central = TRUE,
                           coef = TRUE, order = 1:4) {

  if ((central | coef) & (max(order) < 4)) order <- 1:4

  n    <- if (na.rm) length(which(!is.na(x))) else length(x)
  xbar <- mean(x, na.rm = na.rm)
  out  <- list()

  if (raw) {
    m        <- lapply(order, function(i) 1 / n * sum(x^i))
    names(m) <- paste0("m", order)
    out$m    <- unlist(m)
  }

  if (central | coef) {
    mu        <- lapply(order, function(i) 1 / n * sum((x - xbar)^i))
    names(mu) <- paste0("mu", order)
    if (central) out$mu <- unlist(mu)
  }

  if (coef) {
    coef_df <- data.frame(
      cv = sqrt(mu$mu2) / mu$mu1,
      cs = mu$mu3 / sqrt(mu$mu2)^3,
      ck = mu$mu4 / mu$mu2^2
    )
    out$coefficients <- unlist(coef_df)
  }

  out
}

test_that("test moments function for normal and pareto type II distributions", {
  expect_equal(
    moments("norm", list(
      mean = 2,
      sd = 1
    )),
    list(
      m = c(m1 = 2, m2 = 5, m3 = 14, m4 = 43),
      mu = c(mu1 = 2, mu2 = 1, mu3 = 0, mu4 = 3),
      coefficients = c(cv = .5, cs = 0, ck = 3)
    ),
    tolerance = .01
  )

  expect_equal(
    moments(
      dist = "paretoII",
      distarg = list(
        scale = 1,
        shape = .3
      )
    ),
    list(
      m = c(m1 = 1.4285, m2 = 7.1428, m3 = 214.2857, m4 = -4285.7142),
      mu = c(mu1 = 1.4285, mu2 = 5.1020, mu3 = 189.5043, mu4 = -5435.2353),
      coefficients = c(cv = 1.5811, cs = 16.4438, ck = -208.8)
    ),
    tolerance = .01
  )
})

test_that("test sample.moments function for normal distributions", {
  expect_equal(
    sample.moments(rnorm(10000,
      mean = 2,
      sd = 1
    )),
    list(
      m = c(m1 = 2, m2 = 5, m3 = 14, m4 = 43),
      mu = c(mu1 = 2, mu2 = 1, mu3 = 0, mu4 = 3),
      coefficients = c(cv = .5, cs = 0, ck = 3)
    ),
    tolerance = 1
  )
})

#' Autoregressive model of order 1
#'
#' Generates a Gaussian time series from a first-order autoregressive (AR(1))
#' model with specified lag-1 autocorrelation, mean, and standard deviation.
#'
#' @param n      Positive integer. Number of values to generate.
#' @param alpha  Numeric in \eqn{(-1, 1)}. Lag-1 autocorrelation coefficient.
#' @param mean   Numeric. Process mean. Default \code{0}.
#' @param sd     Positive numeric. Process standard deviation. Default \code{1}.
#'
#' @return A numeric vector of length \code{n}.
#'
#' @keywords internal
#'
AR1 <- function(n, alpha, mean = 0, sd = 1) {

  ## --- input validation -------------------------------------------------------
  n <- suppressWarnings(as.integer(n))
  if (is.na(n) || n < 1L) {
    stop("'n' must be a positive integer.")
  }
  if (!is.numeric(alpha) || length(alpha) != 1L || abs(alpha) >= 1) {
    stop("'alpha' must be a single numeric value in (-1, 1).")
  }
  if (!is.numeric(sd) || length(sd) != 1L || sd <= 0) {
    stop("'sd' must be a single positive numeric value.")
  }

  ## --- innovation parameters --------------------------------------------------
  noise_mean <- (1 - alpha) * mean
  noise_sd   <- sqrt(1 - alpha^2) * sd

  ## --- initialise output ------------------------------------------------------
  val <- rnorm(n = 1L, mean = mean, sd = sd)

  ## --- AR(1) recursion --------------------------------------------------------
  if (n > 1L) {
    noise <- rnorm(n = n, mean = noise_mean, sd = noise_sd)
    for (i in seq(2L, n)) {
      val[i] <- alpha * val[i - 1L] + noise[i]
    }
  }

  val
}


#' Yule-Walker solver
#'
#' Solves the Yule-Walker equations to compute AR(\eqn{p}) coefficients from
#' an autocorrelation structure (ACS) vector. Used internally by
#' \code{\link{ARp}} and \code{\link{seasonalAR}}.
#'
#' @param ACS Numeric vector. ACS values starting from lag 0
#'   (\code{ACS[1] = 1}). The AR order is \code{length(ACS) - 1}.
#'
#' @return Numeric matrix of AR coefficients, shape \code{(p, 1)}.
#'
#' @keywords internal
#'
#' @seealso \code{\link{ARp}}, \code{\link{seasonalAR}}
#'
YW <- function(ACS) {

  p   <- length(ACS) - 1L
  idx <- seq_len(p)

  ## symmetric Toeplitz covariance matrix: P[i,j] = ACS[|i-j| + 1]
  P   <- outer(idx, idx, function(i, j) ACS[abs(i - j) + 1L])
  rho <- matrix(ACS[seq(2L, p + 1L)], nrow = p, ncol = 1L)

  solve(P, rho)
}


#' Autoregressive model of order \emph{p}
#'
#' Generates a time series from an AR(\eqn{p}) model. The parent Gaussian
#' process is constructed to match a target autocorrelation structure via the
#' Yule-Walker equations, then transformed to a target marginal distribution
#' and optional intermittency (probability of zeros).
#'
#' @param margdist  Character. Name of the target marginal distribution
#'   (e.g. \code{"ggamma"}, \code{"paretoII"}). Must have a quantile function
#'   \code{q<margdist>} available.
#' @param margarg   Named list. Parameters of the marginal distribution passed
#'   to \code{q<margdist>}.
#' @param acsvalue  Numeric vector. Target autocorrelation structure starting
#'   from lag 0 (i.e. \code{acsvalue[1] = 1}).
#' @param actfpara  List returned by \code{\link{fitactf}}, containing the
#'   fitted ACTF coefficients in \code{actfpara$actfcoef}.
#' @param n         Positive integer. Length of the generated time series.
#' @param p         Positive integer or \code{NULL}. AR model order. When
#'   \code{NULL} (default), the order is chosen automatically as the number of
#'   lags where the transformed ACS exceeds 0.01, capped at 1000.
#' @param p0        Numeric in \eqn{[0, 1)}. Probability of zero values
#'   (intermittency). Default \code{0}.
#'
#' @return A numeric vector of length \code{n} with the following attributes:
#' \describe{
#'   \item{margdist}{Target marginal distribution name.}
#'   \item{margarg}{Target marginal distribution parameters.}
#'   \item{acsvalue}{Original (untransformed) target ACS.}
#'   \item{p0}{Probability of zeros.}
#'   \item{ar_coef}{AR(\eqn{p}) coefficients (Yule-Walker solution).}
#'   \item{noise_sd}{Standard deviation of the Gaussian innovations.}
#'   \item{gaussian}{The underlying Gaussian process (length \code{n}).}
#'   \item{transformed_acs}{ACTF-transformed ACS used for the AR model.}
#' }
#'
#' @keywords internal
#'
#' @importFrom matrixcalc is.positive.definite
#'
#' @seealso \code{\link{generateTS}}, \code{\link{actpnts}},
#'   \code{\link{fitactf}}, \code{\link{acs}}
#'
ARp <- function(margdist, margarg, acsvalue, actfpara, n, p = NULL, p0 = 0) {

  ## --- input validation -------------------------------------------------------
  if (!is.character(margdist) || length(margdist) != 1L) {
    stop("'margdist' must be a single character string.")
  }
  if (!is.list(margarg)) {
    stop("'margarg' must be a named list of distribution parameters.")
  }
  if (!is.numeric(acsvalue) || acsvalue[1] != 1) {
    stop("'acsvalue' must be a numeric vector starting with 1 (lag-0 value).")
  }
  n <- suppressWarnings(as.integer(n))
  if (is.na(n) || n < 1L) {
    stop("'n' must be a positive integer.")
  }
  if (!is.numeric(p0) || length(p0) != 1L || p0 < 0 || p0 >= 1) {
    stop("'p0' must be a single numeric value in [0, 1).")
  }

  ## --- transform ACS via ACTF -------------------------------------------------
  b <- actfpara$actfcoef[1]
  c <- actfpara$actfcoef[2]
  transformed_acs <- actf(acsvalue, b = b, c = c)

  ## --- determine AR order p ---------------------------------------------------
  max_order <- 1000L

  if (is.null(p) || p > max_order) {
    n_sig <- length(transformed_acs[transformed_acs > 0.01]) - 1L
    p     <- min(n_sig, max_order)
    message(sprintf('AR order "p" automatically set to %d.', p))
  }

  if (length(transformed_acs) - 1L < p) {
    stop(sprintf(
      "ACS is too short for order p = %d. Supply a longer 'acsvalue' or set p = NULL.",
      p
    ))
  }

  ## --- solve Yule-Walker via shared YW() helper -------------------------------
  rho     <- matrix(transformed_acs[seq(2L, p + 1L)], nrow = p, ncol = 1L)

  ## build Toeplitz covariance matrix for positive-definiteness check
  idx <- seq_len(p)
  P   <- outer(idx, idx, function(i, j) transformed_acs[abs(i - j) + 1L])

  if (!is.positive.definite(P, tol = 1e-8)) {
    stop(paste0(
      "Covariance matrix is not positive definite. ",
      "Try a different ACS (e.g. Pareto II, Weibull, or fGn)."
    ))
  }

  ar_coef  <- YW(transformed_acs[seq_len(p + 1L)])  # uses shared helper
  noise_sd <- sqrt(1 - sum(rho * ar_coef))

  ## --- generate parent Gaussian process ---------------------------------------
  gauss <- AR1(n = p, alpha = rho[1])

  innovations <- rnorm(n = p + n, mean = 0, sd = noise_sd)
  ar_coef_rev <- rev(ar_coef)

  for (i in seq(p + 1L, n + p)) {
    gauss[i] <- sum(gauss[(i - p):(i - 1L)] * ar_coef_rev) + innovations[i]
  }

  ## --- transform to target marginal -------------------------------------------
  gauss_out      <- gauss[-seq_len(p)]
  uval           <- (pnorm(gauss_out) - p0) / (1 - p0)
  uval[uval < 0] <- 0

  out <- do.call(paste0("q", margdist), args = c(list(p = uval), margarg))

  ## --- attach metadata as attributes ------------------------------------------
  structure(
    .Data           = out,
    margdist        = margdist,
    margarg         = margarg,
    acsvalue        = acsvalue,
    p0              = p0,
    ar_coef         = ar_coef,
    noise_sd        = noise_sd,
    gaussian        = gauss_out,
    transformed_acs = transformed_acs
  )
}

test_that('test random values deneration using AR1 and ARp models', {

  expect_equal({set.seed(666);
    AR1(n = 10,
        alpha = .8,
        mean = 0,
        sd = 1)},
    c(0.7533, 0.3895, 1.5285, -0.1072, 0.3692, -0.4883, -0.8721, -1.7730, -1.4436, 0.1350),
    tolerance = .01)


  dist <- 'paretoII'
  distarg <- list(scale = 1,
                  shape = .3)
  p0 <- .5
  order <- 5000

  pnts <- actpnts(margdist = dist,
                  margarg = distarg,
                  p0 = p0)
  fit <- fitactf(pnts)

  acsvalue <- acs(id = 'weibull',
                  t = 0:order,
                  scale = 10,
                  shape = .75)

  set.seed(666)
  val <- ARp(margdist = dist,
             margarg = distarg,
             acsvalue = acsvalue,
             actfpara = fit,
             n = 10000,
             p0 = p0)
  attributes(val) <- NULL

  # Reference values at indices 28:35 (first stable non-zero region with seed 666)
  expect_equal(val[28:35],
               c(0.6953, 0.9566, 1.8351, 0.4491, 1.2782, 1.1348, 2.5966, 2.2523),
    tolerance = .01)
})

test_that('test wrapper for acti and ARp functions', {
  set.seed(666)
  x <- generateTS(margdist = 'burrXII',
                  margarg = list(scale = 1,
                                 shape1 = .75,
                                 shape2 = .25),
                  acsvalue = acs(id = 'weibull',
                                 t = 0:30,
                                 scale = 10,
                                 shape = .75),
                  n = 1000, p = 30, p0 = .5, TSn = 3)

  expect_equal({
      val <- x[[1]];
      attributes(val) <- NULL;
      val[190:200]
    },
    c(0, 0, 0, 0, 0, 0, 0.1546, 0, 0, 0, 0),
    tolerance = .01
  )

  expect_equal({set.seed(666);
    val <- regenerateTS(x);
    unlist(val)[190:200]},
  c(0, 0.0214, 0, 0, 0, 0, 0, 0, 0, 0, 0),
  tolerance = .01)
})

# Internal helper shared by generateRF() and generateMTS()
# Runs the VAR simulation and applies the marginal back-transformation.
# Not exported.
simulate_var <- function(n, STmodel) {
  mm      <- STmodel$mm
  sim.out <- mAr.sim(w = rep(0, mm),
                     A = STmodel$alpha,
                     C = STmodel$res.cov,
                     N = n)
  x <- as.matrix(sim.out)

  if (STmodel$dsid == "gauss") {
    x <- pnorm(x, mean(x), sd(x))
  }
  if (STmodel$dsid == "student") {
    nu <- STmodel$dsarg
    S  <- rchisq(n, nu)
    x  <- pt(x * sqrt(nu / S), nu)
  }
  if (STmodel$dsid == "bardossy") {
    m <- STmodel$dsarg
    set.seed(666)
    x <- apply((x - mean(x) + m)^2, 2,
                function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }
  if (STmodel$dsid == "bardossyF") {
    m <- STmodel$dsarg
    set.seed(666)
    x <- 1 - apply((x - mean(x) + m)^2, 2,
                   function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }

  p0       <- STmodel$p0
  margdist <- STmodel$margdist
  margarg  <- STmodel$margarg

  x[x <= p0] <- 0
  x[x > p0]  <- do.call(paste0("q", margdist),
                         args = c(list(p = (x[x > p0] - p0) / (1 - p0)),
                                  margarg))

  structure(.Data = x, class = "matrix", STmodel = STmodel)
}


#' Simulation of random fields with given marginals and spatiotemporal properties
#'
#' Generates a random field with given marginals and spatiotemporal properties.
#' Provide (1) the output of \code{\link{fitVAR}} and (2) the number of time
#' steps to simulate.
#'
#' @param n number of fields (time steps) to simulate
#' @param STmodel list of arguments from \code{\link{fitVAR}}
#'
#' @return A matrix of class \code{"matrix"} with attribute \code{STmodel}.
#'   Rows correspond to spatial locations and columns to time steps.
#'
#' @seealso \code{\link{fitVAR}}, \code{\link{checkRF}}, \code{\link{generateMTS}}
#'
#' @name generateRF
#'
#' @import mAr
#'
#' @export
#'
#' @details
#' Referring to the documentation of \code{\link{fitVAR}} for details on
#' computational complexity, here we report indicative simulation CPU times,
#' assuming model parameters are already evaluated. CPU times refer to a
#' Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
#' 4-core, 8 logical processors, and 32 GB RAM. \cr
#' CPU time: \cr
#' m = 30, p = 1, n = 1000: ~17s \cr
#' m = 30, p = 1, n = 10000: ~75s \cr
#' m = 30, p = 5, n = 100: ~280s \cr
#' m = 30, p = 5, n = 1000: ~302s \cr
#' m = 50, p = 1, n = 1000: ~160s \cr
#' m = 50, p = 1, n = 10000: ~570s \cr
#' where m denotes the side length of a square field (m x m).
#'
#' @examples
#' ## The example below simulates few random fields of size 10x10 with AR(1)
#' ## temporal correlation for illustration. For reliable performance assessment
#' ## generate a larger number of fields (e.g. 100 or more) of size ~30x30.
#' ## See 'Details' for running times with different settings.
#'
#' fit <- fitVAR(
#'   spacepoints = 10,
#'   p = 1,
#'   margdist = "burrXII",
#'   margarg = list(scale = 3, shape1 = .9, shape2 = .2),
#'   p0 = 0.8,
#'   stcsid = "clayton",
#'   stcsarg = list(scfid = "weibull", tcfid = "weibull",
#'                  copulaarg = 2,
#'                  scfarg = list(scale = 20, shape = 0.7),
#'                  tcfarg = list(scale = 1.1, shape = 0.8))
#' )
#'
#' sim <- generateRF(n = 12, STmodel = fit)
#' checkRF(sim, lags = 10, nfields = 12)
#'
generateRF <- function(n, STmodel) {
  simulate_var(n, STmodel)
}


#' Simulation of multiple time series with given marginals and spatiotemporal
#' properties
#'
#' Generates multiple time series with given marginals and spatiotemporal
#' properties. Provide (1) the output of \code{\link{fitVAR}} and (2) the number
#' of time steps to simulate.
#'
#' @param n number of time steps to simulate
#' @param STmodel list of arguments from \code{\link{fitVAR}}
#'
#' @return A matrix of class \code{"matrix"} with attribute \code{STmodel}.
#'   Rows correspond to time steps and columns to spatial locations.
#'
#' @seealso \code{\link{fitVAR}}, \code{\link{generateRF}}, \code{\link{generateMTSFast}}
#'
#' @name generateMTS
#'
#' @import mAr
#'
#' @export
#'
#' @details
#' Referring to the documentation of \code{\link{fitVAR}} for details on
#' computational complexity, here we report indicative simulation CPU times,
#' assuming model parameters are already evaluated. CPU times refer to a
#' Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
#' 4-core, 8 logical processors, and 32 GB RAM. \cr
#' CPU time: \cr
#' d = 900, p = 1, n = 1000: ~17s \cr
#' d = 900, p = 1, n = 10000: ~75s \cr
#' d = 900, p = 5, n = 100: ~280s \cr
#' d = 900, p = 5, n = 1000: ~302s \cr
#' d = 2500, p = 1, n = 1000: ~160s \cr
#' d = 2500, p = 1, n = 10000: ~570s \cr
#' where \eqn{d} denotes the number of spatial locations.
#'
#' @examples
#' ## Simulation of a 4-dimensional vector with VAR(1) correlation structure
#' coord <- cbind(runif(4) * 30, runif(4) * 30)
#'
#' fit <- fitVAR(
#'   spacepoints = coord,
#'   p = 1,
#'   margdist = "burrXII",
#'   margarg = list(scale = 3,
#'                  shape1 = .9,
#'                  shape2 = .2),
#'   p0 = 0.8,
#'   stcsid = "clayton",
#'   stcsarg = list(scfid = "weibull",
#'                  tcfid = "weibull",
#'                  copulaarg = 2,
#'                  scfarg = list(scale = 20,
#'                                shape = 0.7),
#'                  tcfarg = list(scale = 1.1,
#'                                shape = 0.8))
#' )
#'
#' sim <- generateMTS(n = 100, STmodel = fit)
#'
generateMTS <- function(n, STmodel) {
  simulate_var(n, STmodel)
}


#' Faster simulation of random fields with approximately separable
#' spatiotemporal correlation structure
#'
#' For more details see section 6 in Serinaldi and Kilsby (2018) and section 2.4
#' in Papalexiou and Serinaldi (2020).
#'
#' @param n number of fields (time steps) to simulate
#' @param spacepoints side length m of the square field (m x m)
#' @param margdist target marginal distribution of the field
#' @param margarg list of marginal distribution arguments; consult the
#'   documentation of the selected distribution for the required parameters
#' @param p0 probability zero
#' @param distbounds distribution bounds (default \code{c(-Inf, Inf)})
#' @param stcsarg list of spatiotemporal correlation structure arguments;
#'   consult the documentation of the selected structure for required parameters
#' @param scalefactor factor specifying the distance between pixel centres
#'   (default 1)
#' @param anisotropyid spatial anisotropy ID (\code{"affine"} by default;
#'   \code{"swirl"} or \code{"wave"} also available)
#' @param anisotropyarg list of arguments for \code{\link{anisotropyT}};
#'   isotropic fields by default
#' @param dsid dependence structure ID (\code{"gauss"} by default;
#'   \code{"student"}, \code{"bardossy"}, or \code{"bardossyF"})
#' @param dsarg argument for the dependence structure: \code{NULL} for
#'   \code{"gauss"}, degrees of freedom for \code{"student"}, or parameter
#'   \code{m} in \eqn{(-\infty, \infty)} for \code{"bardossy"}
#'
#' @return A matrix of class \code{c("matrix", "cosmosts")} with attribute
#'   \code{STmodel} containing the fitted model components.
#'
#' @seealso \code{\link{generateRF}}, \code{\link{generateMTSFast}},
#'   \code{\link{checkRF}}, \code{\link{fitVAR}}
#'
#' @name generateRFFast
#'
#' @import mvtnorm
#'
#' @importFrom Matrix nearPD
#' @importFrom matrixcalc is.positive.definite
#'
#' @export
#'
#' @references Serinaldi, F., Kilsby, C.G. (2018). Unsurprising Surprises:
#' The Frequency of Record-breaking and Overthreshold Hydrological Extremes Under
#' Spatial and Temporal Dependence. Water Resources Research, 54(9), 6460-6487,
#' \doi{10.1029/2018WR023055}
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#'
#' @details
#' \code{\link{generateRFFast}} provides faster RF simulation than
#' \code{\link{generateRF}} by exploiting circulant-embedding fast Fourier
#' transformation. This approach is feasible only for approximately separable
#' target spatiotemporal correlation functions.
#' \code{\link{generateRFFast}} combines fitting and simulation in a single call.
#' Indicative CPU times (Windows 10 Pro x64, Intel Core i7-6700HQ, 32 GB RAM): \cr
#' m = 50, n = 1000: ~58s \cr
#' m = 50, n = 10000: ~160s \cr
#' m = 100, n = 1000: ~2955s (~50 min) \cr
#'
#' @examples
#'
#' sim <- generateRFFast(
#'     n = 50,
#'     spacepoints = 3,
#'     p0 = 0.7,
#'     margdist = "paretoII",
#'     margarg = list(scale = 1,
#'                    shape = .3),
#'     stcsarg = list(scfid = "weibull",
#'                    tcfid = "weibull",
#'                    scfarg = list(scale = 20,
#'                                  shape = 0.7),
#'                    tcfarg = list(scale = 1.1,
#'                                  shape = 0.8))
#' )
#'
#' checkRF(sim, lags = 10, nfields = 49)
#'
generateRFFast <- function(n, spacepoints, margdist, margarg, p0,
                           distbounds = c(-Inf, Inf), stcsarg,
                           scalefactor = 1,
                           anisotropyid = "affine",
                           anisotropyarg = list(phi1 = 1, phi2 = 1,
                                                phi12 = 0, theta = 0),
                           dsid = "gauss", dsarg = NULL) {

  if (!is.numeric(spacepoints))
    stop("spacepoints should be an integer")

  mm <- spacepoints^2
  d  <- seq(0, spacepoints - 1, length = spacepoints) * scalefactor
  dd <- expand.grid(d, d)

  anisotropyarg$spacepoints <- dd
  dd  <- anisotropyT2(anisotropyid, anisotropyarg)
  dis <- as.matrix(dist(dd, diag = TRUE, upper = TRUE))

  if (dsid %in% c("gauss", "student")) {
    pnts <- actpnts(margdist = margdist, margarg = margarg,
                    p0 = p0, distbounds = distbounds)
  }
  if (dsid %in% c("bardossy", "bardossyF")) {
    pnts <- actpntsB6(margdist = margdist, margarg = margarg,
                      m = dsarg, p0 = p0)
  }
  actfpara <- fitactf(pnts)

  scf <- actf(acs(id = stcsarg$scfid, t = dis,
                  scale = stcsarg$scfarg$scale,
                  shape = stcsarg$scfarg$shape),
              b = actfpara$actfcoef[1],
              c = actfpara$actfcoef[2])
  if (!is.positive.definite(round(as.matrix(scf), 6)))
    scf <- as.matrix(nearPD(scf, corr = TRUE, conv.tol = 1e-04)$mat)

  tcf <- actf(acs(id = stcsarg$tcfid, t = 0:n,
                  scale = stcsarg$tcfarg$scale,
                  shape = stcsarg$tcfarg$shape),
              b = actfpara$actfcoef[1],
              c = actfpara$actfcoef[2])

  Sj  <- DHMgenSj(tcf)
  rn  <- rmvnorm(2 * length(Sj) - 2, sigma = scf)
  out <- matrix(NA, nrow = mm, ncol = n)
  for (i in 1:mm) {
    out[i, ] <- DHMgenSim(Sj = Sj, rn = rn[, i])
  }

  x <- t(as.matrix(out))

  if (dsid == "gauss") {
    x <- pnorm(x, mean(x), sd(x))
  }
  if (dsid == "student") {
    nu <- dsarg
    S  <- rchisq(n, nu)
    x  <- pt(x * sqrt(nu / S), nu)
  }
  if (dsid == "bardossy") {
    m <- dsarg
    set.seed(666)
    x <- apply((x - mean(x) + m)^2, 2,
               function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }
  if (dsid == "bardossyF") {
    m <- dsarg
    set.seed(666)
    x <- 1 - apply((x - mean(x) + m)^2, 2,
                   function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }

  x[x <= p0] <- 0
  x[x > p0]  <- do.call(paste0("q", margdist),
                         args = c(list(p = (x[x > p0] - p0) / (1 - p0)),
                                  margarg))

  d    <- seq(0, max(dis), length = 200)
  u    <- 0:n
  scf0 <- actf(acs(id = stcsarg$scfid, t = d,
                   scale = stcsarg$scfarg$scale,
                   shape = stcsarg$scfarg$shape),
               b = actfpara$actfcoef[1],
               c = actfpara$actfcoef[2])
  tcf0 <- actf(acs(id = stcsarg$tcfid, t = u,
                   scale = stcsarg$tcfarg$scale,
                   shape = stcsarg$tcfarg$shape),
               b = actfpara$actfcoef[1],
               c = actfpara$actfcoef[2])

  fout <- actfInv(tcf0 %*% t(scf0),
                  b = actfpara$actfcoef[1],
                  c = actfpara$actfcoef[2])

  structure(.Data = x,
            class = c("matrix", "cosmosts"),
            STmodel = list(margdist = margdist,
                           margarg  = margarg,
                           p0       = p0,
                           stcs     = fout,
                           s.lags   = d,
                           t.lags   = u,
                           grid.lags = dis))
}


#' Faster simulation of multiple time series with approximately separable
#' spatiotemporal correlation structure
#'
#' For more details see section 6 in Serinaldi and Kilsby (2018) and section 2.4
#' in Papalexiou and Serinaldi (2020).
#'
#' @param n number of time steps to simulate
#' @param spacepoints matrix (d x 2) of coordinates (e.g. longitude and
#'   latitude) for d spatial locations (e.g. gauge stations)
#' @param margdist target marginal distribution
#' @param margarg list of marginal distribution arguments; consult the
#'   documentation of the selected distribution for the required parameters
#' @param p0 probability zero
#' @param distbounds distribution bounds (default \code{c(-Inf, Inf)})
#' @param stcsarg list of spatiotemporal correlation structure arguments;
#'   consult the documentation of the selected structure for required parameters
#' @param scalefactor factor specifying the distance between pixel centres
#'   (default 1)
#' @param anisotropyid spatial anisotropy ID (\code{"affine"} by default;
#'   \code{"swirl"} or \code{"wave"} also available)
#' @param anisotropyarg list of arguments for \code{\link{anisotropyT}};
#'   isotropic fields by default
#' @param dsid dependence structure ID (\code{"gauss"} by default;
#'   \code{"student"}, \code{"bardossy"}, or \code{"bardossyF"})
#' @param dsarg argument for the dependence structure: \code{NULL} for
#'   \code{"gauss"}, degrees of freedom for \code{"student"}, or parameter
#'   \code{m} in \eqn{(-\infty, \infty)} for \code{"bardossy"}
#'
#' @return A matrix of class \code{c("matrix", "cosmosts")} with attribute
#'   \code{STmodel} containing the fitted model components.
#'
#' @seealso \code{\link{generateMTS}}, \code{\link{generateRFFast}},
#'   \code{\link{fitVAR}}
#'
#' @name generateMTSFast
#'
#' @import mvtnorm
#'
#' @importFrom Matrix nearPD
#' @importFrom matrixcalc is.positive.definite
#'
#' @export
#'
#' @references Serinaldi, F., Kilsby, C.G. (2018). Unsurprising Surprises:
#' The Frequency of Record-breaking and Overthreshold Hydrological Extremes Under
#' Spatial and Temporal Dependence. Water Resources Research, 54(9), 6460-6487,
#' \doi{10.1029/2018WR023055}
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#'
#' @details
#' \code{\link{generateMTSFast}} provides faster multivariate simulation than
#' \code{\link{generateMTS}} by exploiting circulant-embedding fast Fourier
#' transformation. This approach is feasible only for approximately separable
#' target spatiotemporal correlation functions.
#' \code{\link{generateMTSFast}} combines fitting and simulation in a single call.
#' Indicative CPU times (Windows 10 Pro x64, Intel Core i7-6700HQ, 32 GB RAM): \cr
#' d = 2500, n = 1000: ~58s \cr
#' d = 2500, n = 10000: ~160s \cr
#' d = 10000, n = 1000: ~2955s (~50 min) \cr
#' where \eqn{d} denotes the number of spatial locations.
#'
#' @examples
#' coord <- cbind(runif(4) * 30, runif(4) * 30)
#'
#' sim <- generateMTSFast(
#'     n = 50,
#'     spacepoints = coord,
#'     p0 = 0.7,
#'     margdist = "paretoII",
#'     margarg = list(scale = 1,
#'                    shape = .3),
#'     stcsarg = list(scfid = "weibull",
#'                    tcfid = "weibull",
#'                    scfarg = list(scale = 20,
#'                                  shape = 0.7),
#'                    tcfarg = list(scale = 1.1,
#'                                  shape = 0.8))
#' )
#'
generateMTSFast <- function(n, spacepoints, margdist, margarg, p0,
                            distbounds = c(-Inf, Inf), stcsarg,
                            scalefactor = 1,
                            anisotropyid = "affine",
                            anisotropyarg = list(phi1 = 1, phi2 = 1,
                                                 phi12 = 0, theta = 0),
                            dsid = "gauss", dsarg = NULL) {

  if (!(is.matrix(spacepoints) || is.array(spacepoints)))
    stop("spacepoints should be a matrix (d x 2)")

  mm <- nrow(spacepoints)
  dd <- spacepoints

  anisotropyarg$spacepoints <- dd
  dd  <- anisotropyT2(anisotropyid, anisotropyarg)
  dis <- as.matrix(dist(dd, diag = TRUE, upper = TRUE))

  if (dsid %in% c("gauss", "student")) {
    pnts <- actpnts(margdist = margdist, margarg = margarg,
                    p0 = p0, distbounds = distbounds)
  }
  if (dsid %in% c("bardossy", "bardossyF")) {
    pnts <- actpntsB6(margdist = margdist, margarg = margarg,
                      m = dsarg, p0 = p0)
  }
  actfpara <- fitactf(pnts)

  scf <- actf(acs(id = stcsarg$scfid, t = dis,
                  scale = stcsarg$scfarg$scale,
                  shape = stcsarg$scfarg$shape),
              b = actfpara$actfcoef[1],
              c = actfpara$actfcoef[2])
  if (!is.positive.definite(round(as.matrix(scf), 6)))
    scf <- as.matrix(nearPD(scf, corr = TRUE, conv.tol = 1e-04)$mat)

  tcf <- actf(acs(id = stcsarg$tcfid, t = 0:n,
                  scale = stcsarg$tcfarg$scale,
                  shape = stcsarg$tcfarg$shape),
              b = actfpara$actfcoef[1],
              c = actfpara$actfcoef[2])

  Sj  <- DHMgenSj(tcf)
  rn  <- rmvnorm(2 * length(Sj) - 2, sigma = scf)
  out <- matrix(NA, nrow = mm, ncol = n)
  for (i in 1:mm) {
    out[i, ] <- DHMgenSim(Sj = Sj, rn = rn[, i])
  }

  x <- t(as.matrix(out))

  if (dsid == "gauss") {
    x <- pnorm(x, mean(x), sd(x))
  }
  if (dsid == "student") {
    nu <- dsarg
    S  <- rchisq(n, nu)
    x  <- pt(x * sqrt(nu / S), nu)
  }
  if (dsid == "bardossy") {
    m <- dsarg
    set.seed(666)
    x <- apply((x - mean(x) + m)^2, 2,
               function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }
  if (dsid == "bardossyF") {
    m <- dsarg
    set.seed(666)
    x <- 1 - apply((x - mean(x) + m)^2, 2,
                   function(x) jitter(rank(x) / (n + 1), amount = 1 / (n + 1)))
  }

  x[x <= p0] <- 0
  x[x > p0]  <- do.call(paste0("q", margdist),
                         args = c(list(p = (x[x > p0] - p0) / (1 - p0)),
                                  margarg))

  d    <- seq(0, max(dis), length = 200)
  u    <- 0:n
  scf0 <- actf(acs(id = stcsarg$scfid, t = d,
                   scale = stcsarg$scfarg$scale,
                   shape = stcsarg$scfarg$shape),
               b = actfpara$actfcoef[1],
               c = actfpara$actfcoef[2])
  tcf0 <- actf(acs(id = stcsarg$tcfid, t = u,
                   scale = stcsarg$tcfarg$scale,
                   shape = stcsarg$tcfarg$shape),
               b = actfpara$actfcoef[1],
               c = actfpara$actfcoef[2])

  fout <- actfInv(tcf0 %*% t(scf0),
                  b = actfpara$actfcoef[1],
                  c = actfpara$actfcoef[2])

  structure(.Data = x,
            class = c("matrix", "cosmosts"),
            STmodel = list(margdist  = margdist,
                           margarg   = margarg,
                           p0        = p0,
                           stcs      = fout,
                           s.lags    = d,
                           t.lags    = u,
                           grid.lags = dis))
}

## acs.R
## AutoCorrelation Structure dispatcher and parametric acf* functions.
## Group D refactoring — style pass; acf* helpers unexported (BUG-05).


#' AutoCorrelation Structure
#'
#' Provides a parametric function that describes the values of the linear
#' autocorrelation up to desired lags. For more details on the parametric
#' autocorrelation structures see section 3.2 in Papalexiou (2018).
#'
#' @param id autocorrelation structure id.
#' @param ... other arguments (\code{t} as lag and ACS parameters).
#'
#' @return A numeric vector of autocorrelation values at the supplied lags.
#'
#' @name acs
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic
#'   modelling of hydroclimatic processes: Preserving marginal distributions,
#'   correlation structures, and intermittency. Advances in Water Resources,
#'   115, 234-252, \doi{10.1016/j.advwatres.2018.02.013}
#'
#' @seealso \code{\link{actpnts}}, \code{\link{fitACS}}, \code{\link{fitactf}}
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## specify lag
#' t <- 0:10
#'
#' ## get the ACS
#' f <- acs("fgn",     t = t, H = .75)
#' b <- acs("burrXII", t = t, scale = 1, shape1 = .6, shape2 = .4)
#' w <- acs("weibull", t = t, scale = 2, shape = 0.8)
#' p <- acs("paretoII", t = t, scale = 3, shape = 0.3)
#'
#' ## visualize the ACS
#' dta   <- data.table(t, f, b, w, p)
#' m.dta <- melt(dta, id.vars = "t")
#'
#' ggplot(m.dta,
#'        aes(x      = t,
#'            y      = value,
#'            group  = variable,
#'            colour = variable)) +
#'   geom_point(size = 2.5) +
#'   geom_line(lwd = 1) +
#'   scale_color_manual(values = c("steelblue4", "red4", "green4", "darkorange"),
#'                      labels = c("FGN", "Burr XII", "Weibull", "Pareto II"),
#'                      name   = "") +
#'   labs(x = bquote(lag ~ tau),
#'        y = "Acf") +
#'   scale_x_continuous(breaks = t) +
#'   theme_classic()

acs <- function(id, ...) {

  args <- list(...)

  do.call(paste0("acf", id), args = args)
}


#' Parametric autocorrelation structure functions
#'
#' @description
#' Parametric functions used internally by \code{\link{acs}} to evaluate
#' autocorrelation at lag \code{t}.  These are internal helpers and are not
#' part of the public API; call them via \code{acs(id, ...)}.
#'
#' @param t     lag value (non-negative numeric).
#' @param scale scale parameter (must be > 0).
#' @param shape,shape1,shape2   shape parameters (must be > 0).
#' @param H     Hurst exponent for the FGN model (0.5 < H < 1).
#'
#' @return A numeric value (or \code{NaN} if parameters are invalid).
#'
#' @keywords internal
#' @name ACSfunctions

acfburrXII <- function(t, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    (1 + shape2 * (t / scale) ^ shape1) ^ (-1 / (shape1 * shape2))
  }
}

#' @keywords internal
#' @rdname ACSfunctions

acfparetoII <- function(t, scale, shape) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {
    (1 + (shape * t) / scale) ^ (-1 / shape)
  }
}

#' @keywords internal
#' @rdname ACSfunctions

acffgn <- function(t, H) {

  (abs(-1 + t) ^ (2 * H) - 2 * abs(t) ^ (2 * H) + abs(1 + t) ^ (2 * H)) / 2
}

#' @keywords internal
#' @rdname ACSfunctions

acfweibull <- function(t, scale, shape) {

  exp(-(t / scale) ^ shape)
}

test_that("test ACS generation", {
  t <- 2
  H <- .75
  scale <- 1
  shape <- .3
  shape1 <- .6
  shape2 <- .2

  expect_equal(acs("fgn", t, H), 0.27,
    tolerance = 0.01)

  expect_equal(acs('burrXII', t, scale, shape1, shape2),   0.110,
               tolerance = 0.01)

  expect_equal(acs("paretoII", t, scale, shape), 0.209,
    tolerance = 0.01)

  expect_equal(acs("weibull", t, scale, shape), 0.292,
    tolerance = 0.01)

  expect_error(acs("wleibull", t, scale, shape)) ## misspelled distribution name
})

## distributions.R
## Merged from: ggamma.R, paretoII.R, burrXII.R, burrIII.R, gev.R
## Group C refactoring.  No algorithmic changes; style-only pass.
## norm.R belongs in fitting.R (Group F) — not included here.


# ===========================================================================
# Generalized Gamma (GGamma)
# ===========================================================================

#' Generalized Gamma distribution
#'
#' Provides density, distribution function, quantile function, random value
#' generation, and raw moments of order \emph{r} for the generalized gamma
#' distribution.
#'
#' @param x,q     vector of quantiles.
#' @param p       vector of probabilities.
#' @param n       number of observations. If \code{length(n) > 1}, the length
#'   is taken to be the number required.
#' @param r       raw moment order.
#' @param scale,shape1,shape2   scale and shape parameters; the shape
#'   arguments cannot be vectors (must have length one).
#' @param log,log.p   logical; if \code{TRUE}, probabilities \code{p} are
#'   given as \code{log(p)}.
#' @param lower.tail  logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @return \code{dggamma} returns a numeric vector of density values.
#'   \code{pggamma} returns a numeric vector of cumulative probabilities.
#'   \code{qggamma} returns a numeric vector of quantiles.
#'   \code{rggamma} returns a numeric vector of random deviates.
#'   \code{mggamma} returns the raw moment of order \code{r}.
#'
#' @name GGamma
#' @aliases GGamma
#' @aliases dggamma
#'
#' @keywords distribution
#' @concept Univariate
#' @concept Continuous
#'
#' @import stats ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Koutsoyiannis, D. (2012). Entropy based
#'   derivation of probability distributions: A case study to daily rainfall.
#'   Advances in Water Resources, 45, 51-57,
#'   \doi{10.1016/j.advwatres.2011.11.007}
#'
#' @seealso \code{\link{fitDist}}, \code{\link{moments}}
#'
#' @examples
#'
#' ## plot the density
#'
#' ggplot(data.frame(x = c(0, 20)),
#'        aes(x)) +
#'   stat_function(fun = dggamma,
#'                 args = list(scale = 5,
#'                             shape1 = .25,
#'                             shape2 = .75),
#'                 colour = "royalblue4") +
#'   labs(x = "",
#'        y = "Density") +
#'   theme_classic()

dggamma <- function(x, scale, shape1, shape2, log = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    d <- (shape2 * (x / scale) ^ (-1 + shape1)) /
         (exp(x / scale) ^ shape2 * scale * gamma(shape1 / shape2))

    if (log) d <- log(d)

    d
  }
}

#' @rdname GGamma
#' @export

pggamma <- function(q, scale, shape1, shape2,
                    lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    p <- pgamma((q / scale) ^ shape2, scale = 1, shape = shape1 / shape2)

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    p
  }
}

#' @rdname GGamma
#' @export

qggamma <- function(p, scale, shape1, shape2,
                    lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    scale * qgamma(p, scale = 1, shape = shape1 / shape2) ^ (1 / shape2)
  }
}

#' @rdname GGamma
#' @export

rggamma <- function(n, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    qggamma(pnorm(rnorm(n)), scale, shape1, shape2)
  }
}

#' @rdname GGamma
#' @export

mggamma <- function(r, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    (scale ^ r * gamma(r / shape2 + shape1 / shape2)) / gamma(shape1 / shape2)
  }
}


# ===========================================================================
# Pareto Type II (ParetoII)
# ===========================================================================

#' Pareto Type II distribution
#'
#' Provides density, distribution function, quantile function, random value
#' generation, and raw moments of order \emph{r} for the Pareto type II
#' distribution.
#'
#' @param x,q     vector of quantiles.
#' @param p       vector of probabilities.
#' @param n       number of observations. If \code{length(n) > 1}, the length
#'   is taken to be the number required.
#' @param r       raw moment order.
#' @param scale,shape   scale and shape parameters; the shape argument cannot
#'   be a vector (must have length one).
#' @param log,log.p   logical; if \code{TRUE}, probabilities \code{p} are
#'   given as \code{log(p)}.
#' @param lower.tail  logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @return \code{dparetoII} returns a numeric vector of density values.
#'   \code{pparetoII} returns a numeric vector of cumulative probabilities.
#'   \code{qparetoII} returns a numeric vector of quantiles.
#'   \code{rparetoII} returns a numeric vector of random deviates.
#'   \code{mparetoII} returns the raw moment of order \code{r}.
#'
#' @name ParetoII
#' @aliases ParetoII
#' @aliases dparetoII
#'
#' @keywords distribution
#' @concept Univariate
#' @concept Continuous
#'
#' @import stats ggplot2
#' @export
#'
#' @seealso \code{\link{fitDist}}, \code{\link{moments}}
#'
#' @examples
#'
#' ## plot the density
#'
#' ggplot(data.frame(x = c(0, 20)),
#'        aes(x)) +
#'   stat_function(fun = dparetoII,
#'                 args = list(scale = 1,
#'                             shape = .3),
#'                 colour = "royalblue4") +
#'   labs(x = "",
#'        y = "Density") +
#'   theme_classic()

dparetoII <- function(x, scale, shape, log = FALSE) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {

    d <- (1 + (shape * x) / scale) ^ (-1 - 1 / shape) / scale

    if (log) d <- log(d)

    d
  }
}

#' @rdname ParetoII
#' @export

pparetoII <- function(q, scale, shape, lower.tail = TRUE, log.p = FALSE) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {

    p <- 1 - (1 + (shape * q) / scale) ^ (-1 / shape)

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    p
  }
}

#' @rdname ParetoII
#' @export

qparetoII <- function(p, scale, shape, lower.tail = TRUE, log.p = FALSE) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    (scale * (-1 + (1 - p) ^ (-shape))) / shape
  }
}

#' @rdname ParetoII
#' @export

rparetoII <- function(n, scale, shape) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {
    qparetoII(runif(n), scale, shape)
  }
}

#' @rdname ParetoII
#' @export

mparetoII <- function(r, scale, shape) {

  if ((scale <= 0) | (shape <= 0)) {
    NaN
  } else {
    ((scale / shape) ^ r * gamma(1 / shape - r) * gamma(1 + r)) /
      gamma(1 / shape)
  }
}


# ===========================================================================
# Burr Type XII (BurrXII)
# ===========================================================================

#' Burr Type XII distribution
#'
#' Provides density, distribution function, quantile function, random value
#' generation, and raw moments of order \emph{r} for the Burr Type XII
#' distribution.
#'
#' @param x,q     vector of quantiles.
#' @param p       vector of probabilities.
#' @param n       number of observations. If \code{length(n) > 1}, the length
#'   is taken to be the number required.
#' @param r       raw moment order.
#' @param scale,shape1,shape2   scale and shape parameters; the shape
#'   arguments cannot be vectors (must have length one).
#' @param log,log.p   logical; if \code{TRUE}, probabilities \code{p} are
#'   given as \code{log(p)}.
#' @param lower.tail  logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @return \code{dburrXII} returns a numeric vector of density values.
#'   \code{pburrXII} returns a numeric vector of cumulative probabilities.
#'   \code{qburrXII} returns a numeric vector of quantiles.
#'   \code{rburrXII} returns a numeric vector of random deviates.
#'   \code{mburrXII} returns the raw moment of order \code{r}.
#'
#' @name BurrXII
#' @aliases BurrXII
#' @aliases dburrXII
#'
#' @keywords distribution
#' @concept Univariate
#' @concept Continuous
#'
#' @import stats ggplot2
#' @export
#'
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic
#'   modelling of hydroclimatic processes: Preserving marginal distributions,
#'   correlation structures, and intermittency. Advances in Water Resources,
#'   115, 234-252, \doi{10.1016/j.advwatres.2018.02.013}
#'
#' @seealso \code{\link{fitDist}}, \code{\link{moments}}
#'
#' @examples
#'
#' ## plot the density
#'
#' ggplot(data.frame(x = c(0, 10)),
#'        aes(x)) +
#'   stat_function(fun = dburrXII,
#'                 args = list(scale = 5,
#'                             shape1 = .25,
#'                             shape2 = .75),
#'                 colour = "royalblue4") +
#'   labs(x = "",
#'        y = "Density") +
#'   theme_classic()

dburrXII <- function(x, scale, shape1, shape2, log = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    d <- ((x / scale) ^ (-1 + shape1) *
            (1 + shape2 * (x / scale) ^ shape1) ^ (-1 - 1 / (shape1 * shape2))) /
         scale

    if (log) d <- log(d)

    d
  }
}

#' @rdname BurrXII
#' @export

pburrXII <- function(q, scale, shape1, shape2,
                     lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    p <- 1 - (1 + shape2 * (q / scale) ^ shape1) ^ (-1 / (shape1 * shape2))

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    p
  }
}

#' @rdname BurrXII
#' @export

qburrXII <- function(p, scale, shape1, shape2,
                     lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    ## Q(u) = beta * (((1-u)^(-g1*g2) - 1) / g2)^(1/g1)
    scale * (((1 - p) ^ (-(shape1 * shape2)) - 1) / shape2) ^ (1 / shape1)
  }
}

#' @rdname BurrXII
#' @export

rburrXII <- function(n, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    qburrXII(runif(n), scale, shape1, shape2)
  }
}

#' @rdname BurrXII
#' @export

mburrXII <- function(r, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    (scale ^ r *
       shape2 ^ (-1 - r / shape1) *
       beta((r + shape1) / shape1,
            (1 - r * shape2) / (shape1 * shape2))) /
      shape1
  }
}


# ===========================================================================
# Burr Type III (BurrIII)
# ===========================================================================

#' Burr Type III distribution
#'
#' Provides density, distribution function, quantile function, random value
#' generation, and raw moments of order \emph{r} for the Burr Type III
#' distribution.
#'
#' @param x,q     vector of quantiles.
#' @param p       vector of probabilities.
#' @param n       number of observations. If \code{length(n) > 1}, the length
#'   is taken to be the number required.
#' @param r       raw moment order.
#' @param scale,shape1,shape2   scale and shape parameters; the shape
#'   arguments cannot be vectors (must have length one).
#' @param log,log.p   logical; if \code{TRUE}, probabilities \code{p} are
#'   given as \code{log(p)}.
#' @param lower.tail  logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @return \code{dburrIII} returns a numeric vector of density values.
#'   \code{pburrIII} returns a numeric vector of cumulative probabilities.
#'   \code{qburrIII} returns a numeric vector of quantiles.
#'   \code{rburrIII} returns a numeric vector of random deviates.
#'   \code{mburrIII} returns the raw moment of order \code{r}.
#'
#' @name BurrIII
#' @aliases BurrIII
#' @aliases dburrIII
#'
#' @keywords distribution
#' @concept Univariate
#' @concept Continuous
#'
#' @import stats ggplot2
#' @export
#'
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic
#'   modelling of hydroclimatic processes: Preserving marginal distributions,
#'   correlation structures, and intermittency. Advances in Water Resources,
#'   115, 234-252, \doi{10.1016/j.advwatres.2018.02.013}
#'
#' @seealso \code{\link{fitDist}}, \code{\link{moments}}
#'
#' @examples
#'
#' ## plot the density
#'
#' ggplot(data.frame(x = c(1, 15)),
#'        aes(x)) +
#'   stat_function(fun = dburrIII,
#'                 args = list(scale = 5,
#'                             shape1 = .25,
#'                             shape2 = .75),
#'                 colour = "royalblue4") +
#'   labs(x = "",
#'        y = "Density") +
#'   theme_classic()

dburrIII <- function(x, scale, shape1, shape2, log = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    d <- ((x / scale) ^ (-1 - 1 / shape2) *
            (1 + 1 / (shape1 * (x / scale) ^ (1 / shape2))) ^ (-1 - shape1 * shape2)) /
         scale

    if (log) d <- log(d)

    d
  }
}

#' @rdname BurrIII
#' @export

pburrIII <- function(q, scale, shape1, shape2,
                     lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    p <- (1 + 1 / (shape1 * (q / scale) ^ (1 / shape2))) ^ (-(shape1 * shape2))

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    p
  }
}

#' @rdname BurrIII
#' @export

qburrIII <- function(p, scale, shape1, shape2,
                     lower.tail = TRUE, log.p = FALSE) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {

    if (!lower.tail) p <- 1 - p
    if (log.p)       p <- log(p)

    scale * (shape1 * (p ^ (-1 / (shape1 * shape2)) - 1)) ^ (-shape2)
  }
}

#' @rdname BurrIII
#' @export

rburrIII <- function(n, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    qburrIII(runif(n), scale, shape1, shape2)
  }
}

#' @rdname BurrIII
#' @export

mburrIII <- function(r, scale, shape1, shape2) {

  if ((shape1 <= 0) | (shape2 <= 0)) {
    NaN
  } else {
    (scale ^ r *
       gamma((r + shape1) * shape2) *
       gamma(1 - r * shape2)) /
      (shape1 ^ (r * shape2) * gamma(shape1 * shape2))
  }
}


# ===========================================================================
# Generalized Extreme Value (GEV)
# ===========================================================================

#' Generalized Extreme Value distribution
#'
#' Provides density, distribution function, quantile function, random value
#' generation, and raw moments of order \emph{r} for the generalized extreme
#' value distribution.
#'
#' @param x,q     vector of quantiles.
#' @param p       vector of probabilities.
#' @param n       number of observations. If \code{length(n) > 1}, the length
#'   is taken to be the number required.
#' @param r       raw moment order.
#' @param loc,scale,shape   location, scale, and shape parameters.
#' @param log,log.p   logical; if \code{TRUE}, probabilities \code{p} are
#'   given as \code{log(p)}.
#' @param lower.tail  logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.
#'
#' @return \code{dgev} returns a numeric vector of density values.
#'   \code{pgev} returns a numeric vector of cumulative probabilities.
#'   \code{qgev} returns a numeric vector of quantiles.
#'   \code{rgev} returns a numeric vector of random deviates.
#'   \code{mgev} returns the raw moment of order \code{r} (via numerical
#'   integration).
#'
#' @name GEV
#' @aliases GEV
#' @aliases dgev
#'
#' @keywords distribution
#' @concept Univariate
#' @concept Continuous
#'
#' @import stats ggplot2
#' @export
#'
#' @seealso \code{\link{fitDist}}, \code{\link{moments}}
#'
#' @examples
#'
#' ## plot the density
#'
#' ggplot(data.frame(x = c(0, 20)),
#'        aes(x)) +
#'   stat_function(fun = dgev,
#'                 args = list(loc = 1,
#'                             scale = .5,
#'                             shape = .15),
#'                 colour = "royalblue4") +
#'   labs(x = "",
#'        y = "Density") +
#'   theme_classic()

dgev <- function(x, loc, scale, shape, log = FALSE) {

  x <- (x - loc) / scale

  if (shape == 0) {

    d <- log(1 / scale) - x - exp(-x)
  } else {

    n     <- length(x)
    x     <- 1 + shape * x
    i     <- x > 0
    scale <- rep(scale, length.out = n)[i]

    d     <- c()
    d[i]  <- log(1 / scale) - x[i] ^ (-1 / shape) - (1 / shape + 1) * log(x[i])
    d[!i] <- -Inf
  }

  if (!log) d <- exp(d)

  d
}

#' @rdname GEV
#' @export

pgev <- function(q, loc, scale, shape, lower.tail = TRUE, log.p = FALSE) {

  if (shape == 0) {
    p <- (q - loc) / scale
  } else {
    p <- -1 / shape * log(pmax(0, 1 - shape * (q - loc) / scale))
  }

  p <- exp(-exp(-p))

  if (!lower.tail) p <- 1 - p
  if (log.p)       p <- log(p)

  p
}

#' @rdname GEV
#' @export

qgev <- function(p, loc, scale, shape, lower.tail = TRUE, log.p = FALSE) {

  if (!lower.tail) p <- 1 - p
  if (log.p)       p <- log(p)

  if (shape == 0) {
    loc - scale * log(-log(p))
  } else {
    loc + scale / shape * (1 - (-log(p)) ^ shape)
  }
}

#' @rdname GEV
#' @export

rgev <- function(n, loc, scale, shape) {

  qgev(p     = runif(n),
       loc   = loc,
       scale = scale,
       shape = shape)
}

#' @rdname GEV
#' @export

mgev <- function(r, loc, scale, shape) {

  if (scale == 0) {
    NaN
  } else {

    integrand <- function(x, q, ...) {
      x ^ q * dgev(x, loc = loc, scale = scale, shape = shape)
    }

    ## integration bounds vary with the sign of loc (GEV support boundary)
    if (loc == 0) {
      lwr <- -Inf
      upr <- Inf
    } else if (loc > 0) {
      lwr <- loc - scale / shape
      upr <- Inf
    } else {
      lwr <- -Inf
      upr <- loc - scale / shape
    }

    integrate(f             = integrand,
              lower         = lwr,
              upper         = upr,
              q             = r,
              loc           = loc,
              scale         = scale,
              shape         = shape,
              stop.on.error = FALSE)$value
  }
}

test_that("test burr type III dist", {
  scale <- 1
  shape1 <- .8
  shape2 <- .2

  expect_equal(dburrIII(1, scale, shape1, shape2), 0.39,
    tolerance = 0.01
  )

  expect_equal(pburrIII(1, scale, shape1, shape2), 0.878,
    tolerance = 0.01
  )

  set.seed(666)
  expect_equal(rburrIII(1, scale, shape1, shape2), 0.794,
    tolerance = 0.01
  )

  expect_equal(mburrIII(1, scale, shape1, shape2), 0.518,
    tolerance = 0.01
  )
})

test_that("test burr type XII dist", {
  scale <- 1
  shape1 <- .8
  shape2 <- .2

  expect_equal(dburrXII(1, scale, shape1, shape2), 0.267,
    tolerance = 0.01
  )

  expect_equal(pburrXII(1, scale, shape1, shape2), 0.68,
    tolerance = 0.01
  )

  set.seed(666)
  expect_equal(rburrXII(1, scale, shape1, shape2), 1.45,
    tolerance = 0.01
  )

  expect_equal(mburrXII(1, scale, shape1, shape2), 1.1,
    tolerance = 0.01
  )
})

test_that("test generalized gamma dist", {
  scale <- 1
  shape1 <- .8
  shape2 <- .2

  expect_equal(dggamma(1, scale, shape1, shape2), 0.0273,
    tolerance = 0.01
  )

  expect_equal(pggamma(1, scale, shape1, shape2), 0.019,
    tolerance = 0.01
  )

  set.seed(666)
  expect_equal(rggamma(1, scale, shape1, shape2), 4175,
    tolerance = 0.01
  )

  expect_equal(mggamma(1, scale, shape1, shape2), 6720,
    tolerance = 0.01
  )
})

test_that("test pareto type II dist", {
  scale <- 1
  shape <- .3

  expect_equal(dparetoII(1, scale, shape), 0.32,
    tolerance = 0.01
  )

  expect_equal(pparetoII(1, scale, shape), 0.58,
    tolerance = 0.01
  )

  set.seed(666)
  expect_equal(rparetoII(1, scale, shape), 1.88,
    tolerance = 0.01
  )

  expect_equal(mparetoII(1, scale, shape), 1.43,
    tolerance = 0.01
  )
})

#' Numerical and visual check of generated random fields
#'
#' Compares generated random fields sample statistics with the theoretically
#' expected values (similar to \code{\link{checkTS}}). It also returns graphical output for
#' visual check.
#'
#' @param RF output of \code{\link{generateRF}}
#' @param lags number of lags of empirical STCF to be considered in the
#' graphical output (default set to 30)
#' @param nfields number of fields to be used in the numerical and graphical
#' output (default set to 49). As the plots are arranged in a matrix with nrows as close as possible to ncol, we suggest using values such as 3x3, 3x4, 7x8, etc.
#' @param method report method - \code{"stat"} for basic statistical report,
#' \code{"statplot"} for graphical check of lagged SCS, target STCS, and marginal
#' distribution, \code{"field"} for plotting a matrix of the first \code{nfields},
#' and \code{"movie"} to save the first \code{nfields} as a GIF file named "movieRF.gif"
#' in the current working directory
#'
#' @name checkRF
#'
#' @import patchwork animation
#'
#' @importFrom grDevices colorRampPalette n2mfrow
#' @importFrom MBA mba.surf
#'
#' @export
#'
#' @examples
#' ## The example below refers to the fitting and simulation of 10 random fields
#' ## of size 10x10 with AR(1) temporal correlation. As the fitting algorithm has
#' ## O((mxm)^3) complexity for a mxm field, this setting allows for quick fitting
#' ## and simulation (short CPU time). However, for a more effective visualization
#' ## and reliable performance assessment, we suggest to generate a larger number
#' ## of fields (e.g. 100 or more) of size about 30X30. This setting needs more
#' ## CPU time but enables more effective comparison of theoretical and
#' ## empirical statistics. Sizes larger than about 50x50 can be unpractical
#' ##  on standard machines.
#'
#' fit <- fitVAR(
#'   spacepoints = 10,
#'   p = 1,
#'   margdist ="burrXII",
#'   margarg = list(scale = 3, shape1 = .9, shape2 = .2),
#'   p0 = 0.8,
#'   stcsid = "clayton",
#'   stcsarg = list(scfid = "weibull", tcfid = "weibull",
#'                  copulaarg = 2,
#'                  scfarg = list(scale = 20, shape = 0.7),
#'                 tcfarg = list(scale = 1.1, shape = 0.8))
#' )
#'
#' sim <- generateRF(n = 12,
#'                     STmodel = fit)
#' checkRF(RF = sim,
#'           lags = 10,
#'           nfields = 12)
#'
#'
checkRF <- function(RF, lags = 30, nfields = 49, method = "stat") {

  att <- attributes(RF)$STmodel
  margdist <- att$margdist
  margarg <- att$margarg
  p0 <- att$p0
  stcs <- att$stcs
  s.lags <- att$s.lags
  t.lags <- att$t.lags
  grid.lags <- att$grid.lags
  distbounds <-  att$distbounds
  m <- sqrt(ncol(RF))
  if (m^2 < 20){
    npoints <- m^2
  } else {
    npoints <- 20
  }
  stcs.sim <- acf(as.matrix(RF)[ , 1:npoints], lag.max = lags, plot=F)[[1]][ , , 1]


  if(method == "stat") {
    out <- data.frame(mean = c(popmean(margdist, margarg, distbounds = distbounds,
                                       p0 = p0), apply(RF[,1:npoints], 2, mean)), sd = c(popsd(margdist, margarg,
                                                                                               distbounds = distbounds, p0 = p0), apply(RF[,1:npoints], 2, sd)), skew = c(popskew(margdist,
                                                                                                                                                                                  margarg, distbounds = distbounds, p0 = p0), apply(RF[,1:npoints], 2,
                                                                                                                                                                                                                                    function(x) sample.moments(x, raw = F, central = F, coef = T)$coefficients[2])),
                      p0 = c(p0, apply(RF[,1:npoints], 2, function(x) length(which(x == 0))/length(x))))
    row.names(out) <- c("expected", paste0("sample location ", 1:npoints))
    out <- round(out, 2)

    out

  }


  if(method == "statplot") {

    ## STCS  slices

    DataS <- data.table(grid.lags[1:npoints], t(as.matrix(stcs.sim[1:3, ])))
    colnames(DataS) <- c("Lag", "scf1", "scf2", "scf3")
    DataS <- melt(DataS, id.vars = "Lag")

    i.lag <- which(s.lags <= max(grid.lags[1:npoints]) )
    DataT <- data.table(s.lags[i.lag], t(as.matrix(stcs[1:3, i.lag])))
    colnames(DataT) <- c("Lag", "scf1", "scf2", "scf3")
    DataT <- melt(DataT, id.vars = "Lag")

    ran <- range(c(DataT$Lag, DataS$Lag))

    Lag <- value <- variable <- level <- x <- y <- NULL

    p1 <- ggplot() +
      geom_point(data = DataS,
                 aes(x = Lag,
                     y = value,
                     colour = (variable)),
                 alpha = .5) +
      geom_line(data = DataT,
                aes(x = Lag,
                    y = value,
                    colour = (variable)),
                lwd = .5,
                alpha = 1) +
      xlim(ran) +
      scale_color_viridis_d(labels = c(expression(paste(tau, " = 0")),
                                      expression(paste(tau, " = 1")),
                                      expression(paste(tau, " = 2"))
                                      ),
                           option = "D") +
      labs(x = expression(paste("Distance ", delta)),
           y = "SCS",
           title = "Lagged spatial correlation structure",
           color = "Time lag") +
      theme_light() +
      theme(legend.position=c(0.99,0.99),
            legend.justification=c(1,1),
            strip.background = element_rect(fill = "grey5"),
            strip.text = element_text(colour = "grey95"))


    ## STCS plot
    Data0 <- cbind(expand.grid(s.lags, (1:lags) - 1), as.numeric(t(stcs[1:lags, ])) )
    colnames(Data0) <- c("x", "y", "z")

    p2 <- ggplot() +
      stat_contour(data = Data0,
                   aes(x = x, y = y, z = z, colour = after_stat(level)),
                   breaks = seq(0, 1, 0.1),
                   linewidth = 1) +
      scale_color_viridis_c(option = 'D') +
      labs(x = expression(paste('Distance ', delta)),
           y = expression(paste('Time lag ', tau)),
           title = 'Target STCS',
           colour = "STCS") +
      theme_light() +
      theme(strip.background = element_rect(fill = 'grey5'),
            strip.text = element_text(colour = 'grey95'))


    ## CDF plot
    if(p0 != 0) xpos <- as.vector(RF[RF > 0])
    else xpos <- as.vector(RF)
    if(length(xpos) > 1000) xpos <- xpos[1:1000]

    Data.simu <- data.table(x = sort(xpos), cdf = ppoints(xpos,0), type = "simu")

    Data.theo <- data.table(
      x = do.call(paste0("q", margdist),
                  args = c(list(p = seq(0.001, 0.999, by = 0.001)), margarg)),
      cdf = seq(0.001, 0.999, by = 0.001),
      type = "theo")

    Data <- rbind(Data.simu, Data.theo)

    p3 <- ggplot(data = Data) +
      geom_line(aes(x = Data$x,
                    y = Data$cdf,
                    colour = factor(Data$type),
                    linetype  = factor(Data$type)),
                lwd = 1) +
      xlim(range(Data$x)) +
      ## scale_color_viridis_d(labels = c("Simulated", "Fitted")) +
      scale_color_manual(labels = c("Simulated", "Fitted"),
                            values = c("royalblue4", "red4")) +
      scale_linetype(labels = c("Simulated", "Fitted")) +
      labs(x = "Nonzero values",
           y = "Nonexceedence probability",
           title = "Probability distribution fit") +
      guides(color = guide_legend(title=NULL),
             linetype = guide_legend(title=NULL)) +
      theme_light() +
      theme(legend.position=c(0.99,0.01),
            legend.justification=c(1,0),
            strip.background = element_rect(fill = "grey5"),
            strip.text = element_text(colour = "grey95"))

    p1 / (p2 | p3)

  }

  my.colors <- colorRampPalette( c(
    "#ffffff","#610B11", "#80110A", "#AF2F03", "#D05201", "#E87D0A", "#EE9E1B", "#F3C03B",
    "#F7DE68", "#FAF397", "#E8D794", "#DABD77", "#CBA265", "#BC8854", "#A96E48",
    "#965F4A", "#7C4F47", "#322221", "#3D2E2C", "#4F413B", "#60534B", "#72675B",
    "#877E6F", "#A39E8A", "#C0BDA8", "#E2E2CD", "#E0CD99", "#C0AE6F", "#A09253",
    "#81783C", "#645F27", "#545316", "#43440A", "#303401", "#173F3B", "#1D6447",
    "#1E803A", "#169517", "#42B218", "#8DCC3C", "#C8E065"))

  if(method == "field") {

    ## Matrix of fields
    op <- par(no.readonly = TRUE) # the whole list of settable par's.
    par(mar = c(0, 0, 0, 0),
        oma = c(0.5, 0.5, 0.5, 0.5))
    graphics::layout(mat = matrix(data = 1:nfields,
                                  nrow = n2mfrow(nfields)[1],
                                  byrow = TRUE))

    if(p0 > 0 & sample.moments(RF[1:nfields, ], raw = F, central = F, coef = T)$coefficients[2] > 1) {

      aux  <-  sqrt(RF[1:nfields, ])
    }  else {

      aux <- RF[1:nfields, ]
    }
    d <- u <- seq(0, m - 1, length = m)
    ud <- expand.grid(u, d)
    for(i in 1:nfields){
      z <- mba.surf(data.frame(x = ud[ ,1], y = ud[ ,2], z = aux[i, ]), 100, 100)$xyz.est$z
      if(p0 > 0) z[z < 0] <- 0
      image(matrix(z, 100, 100), axes=F, col = my.colors(40), zlim = range(aux))
      legend("topleft", legend = i, adj = c(1.3, 0.1), cex = 1.2,
             pch = NA, lty = 0, bty = "n", text.col = "darkgrey", text.font = 2)
      box(col = "darkgrey")
    }

    par(op)

  }


  if(method == "movie") {

    d <- u <- seq(0, m - 1, length = m)
    ud <- expand.grid(u, d)
    op <- par(mar = c(0, 0, 0, 0), oma = c(0.5, 0.5, 0.5, 0.5))
    if(p0 > 0 & sample.moments(RF[1:nfields, ], raw = F, central = F, coef = T)$coefficients[2] > 1) aux  <-  sqrt(RF[1:nfields, ])
    else aux <- RF[1:nfields, ]
    animation::ani.options(loop=0)
    animation::saveGIF({
      par(mar = rep(0, 4))
      for(i in 1:nfields){
        z <- mba.surf(data.frame(x = ud[ ,1], y = ud[ ,2], z = aux[i, ]), 100, 100)$xyz.est$z
        if(p0 > 0) z[z < 0] <- 0
        image(matrix(z, 100, 100), axes=F, col = my.colors(40), zlim = range(sqrt(RF[1:nfields,])))
        legend("topleft", legend = i, adj = c(1.3, 0.1), cex = 1.2,
               pch = NA, lty = 0, bty = "n", text.col = "darkgrey", text.font = 2)
        box(col = "darkgrey")
      }
    }, movie.name = "movieRF.gif",
    interval = 0.5,  ani.width = 480, ani.height = 480)
    par(op)

  }


}

# actf.R
# Merged: actf, actfInv, acti, actpnts, actpntsB6, fitactf
# Group C — style pass applied; C++ fast path added to actpnts()

#' ACTF auto-correlation transformation function
#'
#' Provides transformation for continuous distributions, based on two parameters.
#'
#' @param rhox marginal correlation value
#' @param b 1st parameter
#' @param c 2nd parameter
#'
#' @return numeric vector of transformed Gaussian correlations
#' @seealso \code{\link{actpnts}}, \code{\link{fitactf}}
#' @export
#' @keywords internal
#'
#' @examples
#' actf(.4, 1, 0)
#'
actf <- function(rhox, b, c) {
  ((1 + b * rhox) ^ (1 - c) - 1) / ((1 + b) ^ (1 - c) - 1)
}

#' ACTF auto-correlation transformation function for discrete distributions
#'
#' Provides transformation for discrete distributions, based on two parameters.
#'
#' @inheritParams actf
#'
#' @return numeric vector of transformed Gaussian correlations
#' @seealso \code{\link{actpnts}}, \code{\link{fitactf}}
#' @export
#' @keywords internal
#'
#' @examples
#' actfdiscrete(.4, .2, 1)
#'
actfdiscrete <- function(rhox, b, c) {
  1 - (1 - rhox ^ b) ^ c
}

#' Inverse ACTF
#'
#' Inverse of the autocorrelation transformation function.
#'
#' @inheritParams actf
#' @param rhoz Gaussian correlation value
#'
#' @return numeric vector of back-transformed marginal correlations
#' @seealso \code{\link{actf}}
#' @keywords internal
#' @export
#'
#' @examples
#' actfInv(.4, 1, 0)
#'
actfInv <- function(rhoz, b, c) {
  ((rhoz * ((1 + b) ^ (1 - c) - 1) + 1) ^ (1 / (1 - c)) - 1) / b
}

#' ACTI — autocorrelation transformation integral function
#'
#' Expression supplied to the double integral inside \code{\link{actpnts}}.
#'
#' @param x x-plane value
#' @param y y-plane value
#' @param dist distribution name
#' @param distarg a list of distribution arguments
#' @param rhoz Gaussian correlation
#' @param p0 probability of zero values
#'
#' @return numeric scalar — integrand value at (x, y)
#' @seealso \code{\link{actpnts}}
#' @export
#' @import stats
#' @keywords internal
#'
#' @examples
#' acti(1, -1, "norm", list(mean = 0, sd = 1), .3, 0)
#'
acti <- function(x, y, dist, distarg, rhoz, p0) {

  do.call(what = paste0("q", dist),
          args = c(list(p = (erfc(-x / sqrt(2)) / 2 - p0) / (1 - p0)),
                   distarg)) *
    do.call(what = paste0("q", dist),
            args = c(list(p = (erfc(-y / sqrt(2)) / 2 - p0) / (1 - p0)),
                     distarg)) *
    exp((x ^ 2 + y ^ 2 - 2 * x * y * rhoz) /
          (2 * (-1 + rhoz ^ 2))) / (2 * pi * sqrt(1 - rhoz ^ 2))
}

#' AutoCorrelation Transformed Points
#'
#' Evaluates the (rho_x, rho_z) mapping between the target marginal
#' autocorrelation and the underlying Gaussian autocorrelation, using a
#' double numerical integral.
#'
#' @details
#' When the package is compiled with Rcpp support (i.e., \code{actpnts_cpp}
#' is available), the double integral is evaluated in C++ via the Cubature
#' algorithm, which is substantially faster than the nested base-R
#' \code{integrate()} fallback. The C++ path supports the following
#' distributions natively: \code{ggamma}, \code{paretoII}, \code{burrXII},
#' \code{burrIII}, \code{gev}, \code{norm}, \code{beta}, \code{gamma},
#' \code{exp}, \code{weibull}, \code{lnorm}, \code{unif}. Any other
#' distribution falls back to the R quantile function automatically, so
#' correctness is always preserved.
#'
#' @param margdist target marginal distribution
#' @param margarg list of marginal distribution arguments
#' @param p0 probability zero
#' @inheritParams moments
#'
#' @return A data frame with columns \code{rhoz} (Gaussian correlations) and
#'   \code{rhox} (corresponding target marginal correlations).
#'
#' @seealso \code{\link{fitactf}}, \code{\link{acti}}, \code{\link{generateTS}}
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic modelling
#'   of hydroclimatic processes: Preserving marginal distributions, correlation
#'   structures, and intermittency. Advances in Water Resources, 115, 234-252,
#'   \doi{10.1016/j.advwatres.2018.02.013}
#'
#' @export
#' @import stats ggplot2
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## Pareto type II marginal
#' x <- actpnts(margdist = "paretoII",
#'              margarg  = list(scale = 1, shape = .3),
#'              p0 = 0)
#' x
#'
actpnts <- function(margdist, margarg, p0 = 0, distbounds = c(-Inf, Inf)) {

  rhoz_vals <- c(seq(from = 0.1, to = 0.9, by = 0.1), 0.95)

  .min <- ifelse(test = p0 == 0,
                 yes  = -7.5,
                 no   = -sqrt(2) * inv.erfc(2 * p0))
  .max <- 7.5

  m <- moments(dist      = margdist,
               distarg   = margarg,
               raw       = FALSE,
               central   = TRUE,
               coef      = FALSE,
               distbounds = distbounds,
               p0        = p0,
               order     = 1:2)

  mu1 <- m$mu["mu1"]
  mu2 <- m$mu["mu2"]

  ## --- Fast C++ path (requires compiled actpnts_cpp) ---
  cpp_available <- exists("actpnts_cpp",
                          envir  = asNamespace("CoSMoS"),
                          mode   = "function",
                          inherits = FALSE)

  if (cpp_available) {

    ## Convert margarg list to named numeric vector for C++
    margarg_vec <- unlist(margarg)

    rho <- actpnts_cpp(
      margdist  = margdist,
      margarg   = margarg_vec,
      p0        = p0,
      lower     = .min,
      upper     = .max,
      mu1       = mu1,
      mu2       = mu2,
      rhoz_vals = rhoz_vals
    )

    return(structure(.Data = rho))
  }

  ## --- Pure R fallback (original implementation) ---
  rho <- data.frame(rhoz = rhoz_vals, rhox = 0)

  for (i in seq_len(nrow(rho))) {

    temp <- integrate(
      f = function(y) {
        sapply(y, function(y) {
          integrate(
            f           = function(x) {
              acti(x        = x,
                   y        = y,
                   rhoz     = rho[i, "rhoz"],
                   p0       = p0,
                   dist     = margdist,
                   distarg  = margarg)
            },
            lower        = .min,
            upper        = .max,
            subdivisions = 1.0e4,
            rel.tol      = 1.0e-5
          )$value
        })
      },
      lower        = .min,
      upper        = .max,
      subdivisions = 1.0e4,
      rel.tol      = 1.0e-5
    )$value

    rho[i, "rhox"] <- (temp - mu1 ^ 2) / mu2
  }

  structure(.Data = rho)
}

#' AutoCorrelation Transformed Points for Bardossy dependence structure
#'
#' Evaluates the (rho_x, rho_z) mapping using Monte Carlo integration for the
#' Bardossy copula dependence structure.
#'
#' @param margdist target marginal distribution
#' @param margarg list of marginal distribution arguments
#' @param m mean of the parent Gaussian processes controlling asymmetry
#' @param p0 probability of zero values
#'
#' @return A data frame with columns \code{rhoz} and \code{rhox}.
#'
#' @seealso \code{\link{actpnts}}, \code{\link{generateMTSFast}}
#' @references Bardossy, A. (2006). Copula-based geostatistical models for
#'   groundwater quality parameters. Water Resources Research, 42(11),
#'   \doi{10.1029/2005WR004754}
#'
#' @export
#' @import stats mvtnorm ggplot2
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' x <- actpntsB6(margdist = "paretoII",
#'                margarg  = list(scale = 1, shape = .3),
#'                m = 1,
#'                p0 = 0)
#' x
#'
actpntsB6 <- function(margdist, margarg, m, p0 = 0) {

  rho <- data.frame(rhoz = c(0.05, seq(from = 0.1, to = 0.9, by = 0.1), 0.95),
                    rhox = 0)

  set.seed(666)

  for (i in seq_len(nrow(rho))) {

    rhoz     <- rho[i, "rhoz"]
    mc.size  <- 2e5
    rnd      <- rmvnorm(mc.size,
                        mean  = rep(m, 2),
                        sigma = matrix(c(1, rhoz, rhoz, 1), 2, 2))

    u <- apply(rnd ^ 2, 2,
               function(x) jitter(rank(x) / (mc.size + 1),
                                  amount = 1 / (mc.size + 1)))

    x        <- do.call(what = paste0("q", margdist),
                        args = c(list(p = (u - p0) / (1 - p0)), margarg))
    x[u <= p0] <- 0

    rho[i, "rhox"] <- (mean(x[, 1] * x[, 2]) -
                         mean(x[, 1]) * mean(x[, 2])) /
      (sd(x[, 1]) * sd(x[, 2]))
  }

  structure(.Data = rho)
}

#' Fit the AutoCorrelation Transformation Function
#'
#' Fits the ACTF to the estimated (rho_x, rho_z) points using \code{nls}.
#'
#' @param actpnts estimated ACT points (output of \code{\link{actpnts}})
#' @param discrete logical — is the marginal distribution discrete?
#'
#' @return An object of class \code{"acti"} with components:
#'   \describe{
#'     \item{actfcoef}{fitted ACTF coefficients \code{b} and \code{c}}
#'     \item{actfpoints}{the input ACT points data frame}
#'   }
#'
#' @seealso \code{\link{actpnts}}, \code{\link{actf}}
#' @export
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' p   <- actpnts(margdist = "paretoII",
#'                margarg  = list(scale = 1, shape = .3),
#'                p0 = 0)
#' fit <- fitactf(p)
#' plot(fit)
#'
fitactf <- function(actpnts, discrete = FALSE) {

  suppressWarnings(
    fit <- nls(
      rhoz ~ do.call(
        what = ifelse(discrete, "actfdiscrete", "actf"),
        args = list(rhox, b, c)
      ),
      data      = list(rhoz = actpnts$rhoz, rhox = actpnts$rhox),
      start     = c(b = 1, c = 0),
      lower     = c(b = .001, c = 0),
      algorithm = "port",
      control   = nls.control(maxiter = 50000, warnOnly = TRUE)
    )
  )

  structure(
    .Data = list(actfcoef   = coefficients(fit),
                 actfpoints = actpnts),
    class = "acti"
  )
}

test_that("test auto-correlation transformation integral and fit", {
  expect_equal(acti(1, 1, "norm", list(), .2, 0), 0.0705,
    tolerance = 0.05
  )

  dist <- "paretoII"
  distarg <- list(scale = 1, shape = .3)

  pnts <- actpnts(
    margdist = dist,
    margarg = distarg,
    p0 = 0
  )

  expect_equal(unlist(pnts[c(1, dim(pnts)[1]), ]), c(rhoz = c(0.10, 0.95), rhox = c(0.0549, 0.9134)),
    tolerance = 0.05
  )

  expect_equal(fitactf(pnts)$`actfcoef`, c(b = 1.6154, c = 1.2813),
    tolerance = 0.05
  )

  expect_error(pnts <- actpnts(
    margdist = "gammma", ## misspelled dist
    margarg = distarg,
    p0 = 0
  ))
})

## getArg.R
## Internal helpers to introspect argument names of distribution and ACS
## functions by name.  Both functions look up targets from the CoSMoS
## namespace directly so the target functions do not need to be exported.


#' Get argument names of a distribution function
#'
#' @description
#' Returns the non-standard argument names of the quantile function
#' \code{q<dist>} (i.e. everything except \code{p}, \code{lower.tail},
#' and \code{log.p}).
#'
#' @param dist  distribution name (character), e.g. \code{"paretoII"}.
#'
#' @return A character vector of argument names, or \code{NULL} (with a
#'   message) if the distribution is not defined.
#'
#' @keywords internal

getDistArg <- function(dist) {

  fn_name <- paste0("q", dist)
  ns      <- asNamespace("CoSMoS")

  if (!existsFunction(fn_name) &&
      !exists(fn_name, envir = ns, inherits = FALSE)) {
    message("Distribution is not defined")
    return(NULL)
  }

  ## prefer the package-namespace copy; fall back to search path
  fn <- if (exists(fn_name, envir = ns, inherits = FALSE)) {
    get(fn_name, envir = ns, inherits = FALSE)
  } else {
    get(fn_name, inherits = TRUE)
  }

  x <- unlist(strsplit(deparse(args(fn))[1L], ","))

  gsub(" |= [0-9]+", "",
       x[grep("function|lower\\.tail|log\\.p = FALSE|/",
               x, invert = TRUE)])
}


#' Get argument names of an ACS function
#'
#' @description
#' Returns the non-\code{t} argument names of the ACS function
#' \code{acf<id>} (i.e. the parameter names only).
#'
#' @param id  ACS id (character), e.g. \code{"weibull"}.
#'
#' @return A character vector of parameter names, or \code{NULL} (with a
#'   message) if the ACS is not defined.
#'
#' @keywords internal

getACSArg <- function(id) {

  fn_name <- paste0("acf", id)
  ns      <- asNamespace("CoSMoS")

  if (!exists(fn_name, envir = ns, inherits = FALSE)) {
    message("ACS is not defined")
    return(NULL)
  }

  fn <- get(fn_name, envir = ns, inherits = FALSE)
  x  <- unlist(strsplit(deparse(args(fn))[1L], ","))

  gsub(" |= [0-9]+|\\)", "",
       x[grep("function", x, invert = TRUE)])
}

#' Anisotropy transformation
#'
#' Provides parametric functions that describe different types of planar deformation fields, including affine (rotation and stretching), and swirl-like deformation. For more details see Papalexiou et al.(2021) and references therein.
#'
#' @param id anisotropy type id (\code{affine}, \code{swirl}, and \code{wave})
#' @param ... additional arguments (vector of coordinates and parameters of the anisotropy transformations)
#'
#' @name anisotropyT
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S. M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond,
#' Water Resources Research, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' ## get the anisotropy field
#' at1 <- anisotropyT("affine",
#'                  spacepoints = coord,
#'                  phi1 = 0.5,
#'                  phi2 = 2,
#'                  phi12 = 0,
#'                  theta = -pi/3)
#' at2 <- anisotropyT("swirl",
#'                  spacepoints = coord,
#'                  x0 = floor(m / 2),
#'                  y0 = floor(m / 2),
#'                  b = 10,
#'                  alpha = 1.5 * pi)
#' at3 <- anisotropyT("wave",
#'                  spacepoints = coord,
#'                  phi1 = 0.5,
#'                  phi2 = 2,
#'                  beta = 3,
#'                  theta = 0)
#'
#' ## visualize anisotropy field
#' aux = data.frame(lon = at2[ ,1], lat = at2[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
#' ggplot(aux, aes(x = lon, y = lat)) +
#' geom_path(aes(group = id1)) +
#' geom_path(aes(group = id2)) +
#' geom_point(col = 2) +
#' theme_light()
#'
anisotropyT <- function(id, ...) {

    .args <- list(...)
    do.call(paste0("anisotropyT", id), args = .args)
}


#' Anisotropy transformation
#'
#' Provides parametric functions that describe different types of planar deformation fields, including affine (rotation and stretching), and swirl-like deformation. For more details see Papalexiou et al.(2021) and references therein.
#'
#' @param id anisotropy type id
#' @param arglist list of additional arguments (vector of coordinates and parameters of the anisotropy transformations)
#'
#' @name anisotropyT2
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @keywords internal
#'
anisotropyT2 <- function(id, arglist) {

  do.call(paste0("anisotropyT", id), args = arglist)
}


#' Affine anisotropy transformation
#'
#' Affine anisotropy transformation.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param phi1 stretching parameter along the x axis
#' @param phi2 stretching parameter along the y axis
#' @param phi12 shear effect
#' @param theta rotation angle
#'
#' @name anisotropyTaffine
#'
#' @import ggplot2
#' @export
#'
#' @references Allard, D., Senoussi, R., Porcu, E. (2016). Anisotropy Models for
#' Spatial Data. Mathematical Geosciences, 48(3), 305-328, \doi{10.1007/s11004-015-9594-x}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' at <- anisotropyTaffine(spacepoints = coord,
#'                        phi1 = 0.5,
#'                        phi2 = 2,
#'                        phi12 = 0,
#'                        theta = -pi/3)
#'
#' ## visualize transformed coordinate system
#' aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
#' ggplot(aux, aes(x = lon, y = lat)) +
#' geom_path(aes(group = id1)) +
#' geom_path(aes(group = id2)) +
#' geom_point(col = 2) +
#' theme_light()
#'
anisotropyTaffine <- function(spacepoints, phi1, phi2, phi12, theta) {

  The <- matrix(c(cos(theta), sin(theta), -sin(theta), cos(theta)), ncol = 2)
  Phi <- matrix(c(phi1, phi12, phi12, phi2), ncol = 2)
  t(Phi %*%  The %*% t(spacepoints))
}

#' Swirl anisotropy transformation
#'
#' Swirl anisotropy transformation.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param x0 x coordinate of the center of the swirl deformation
#' @param y0 y coordinate of the center of the swirl deformation
#' @param b scaling parameter controlling the swirl deformation
#' @param alpha rotation angle
#'
#' @name anisotropyTswirl
#'
#' @import ggplot2
#' @export
#'
#' @references Ligas, M., Banas, M., Szafarczyk, A. (2019). A method for local
#' approximation of a planar deformation field. Reports on Geodesy and
#' Geoinformatics, 108(1), 1-8, \doi{10.2478/rgg-2019-0007}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' at <- anisotropyTswirl(spacepoints = coord,
#'                       x0 = floor(m / 2),
#'                       y0 = floor(m / 2),
#'                       b = 10,
#'                       alpha = 1.5 * pi)
#'
#' ## visualize transformed coordinate system
#' aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
#' ggplot(aux, aes(x = lon, y = lat)) +
#' geom_path(aes(group = id1)) +
#' geom_path(aes(group = id2)) +
#' geom_point(col = 2) +
#' theme_light()
#'
anisotropyTswirl <- function(spacepoints, x0, y0, b, alpha) {

  x <- spacepoints[,1]
  y <- spacepoints[,2]
  r <- sqrt((x - x0)^2 + (y - y0)^2)
  The <- alpha * exp(-(r / b)^2)
  xani <- (x - x0) * cos(The) - (y - y0) * sin(The) + x0
  yani <- (x - x0) * sin(The) + (y - y0) * cos(The) + y0
  cbind(xani, yani)
}

#' Wave anisotropy transformation
#'
#' Wave anisotropy transformation.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param phi1 stretching parameter along the x axis
#' @param phi2 stretching parameter along the y axis
#' @param beta amplitude of sinusoidal wave
#' @param theta rotation angle
#'
#' @name anisotropyTwave
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' at <- anisotropyTwave(spacepoints = coord,
#'                      phi1 = 0.5,
#'                      phi2 = 2,
#'                      beta = 3,
#'                      theta = 0)
#'
#' ## visualize transformed coordinate system
#' aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
#' ggplot(aux, aes(x = lon, y = lat)) +
#' geom_path(aes(group = id1)) +
#' geom_path(aes(group = id2)) +
#' geom_point(col = 2) +
#' theme_light()
#'
anisotropyTwave <- function(spacepoints, phi1, phi2, beta, theta) {

  x <- spacepoints[,1]
  y <- spacepoints[,2]
  xani <- phi1 * x * cos(theta) - y * sin(theta)
  yani <- x * sin(theta) + phi1 * y * cos(theta) + beta * sin(x)
  cbind(xani, yani)
}

#' Daily streamflow data data
#'
#' Station details
#' * Name: Nassawango Creek near Snow Hill, Worcester County, Maryland, Hydrologic Unit 02080111
#' * Network Id: , USGS 01485500
#' * Latitude/Longitude: 38°13'44.1", 75°28'17.2"
#' * Elevation: 11.49 ft above North American Vertical Datum of 1988.
#' * Measurement unit: cubic feet per second
#'
#' more details can be found \href{https://waterdata.usgs.gov/monitoring-location/USGS-01485500/}{here}.
#'
#' @format A data.table with 23315 rows and 2 variables:
#' \describe{
#'   \item{date}{POSIXct format date/time}
#'   \item{value}{daily avarage values}
#' }
#' @source The United States Geological Survey (USGS) National Water Information System (NWIS)
"disch"

# --- Error norms (internal) ---------------------------------------------------

#' Ratio mean squared error
#'
#' @param x numeric vector of observed values
#' @param y numeric vector of reference values
#'
#' @return scalar numeric
#'
#' @seealso \code{\link{MSE}}
#' @keywords internal
rMSE <- function(x, y) {
  sum((x / y - 1)^2) / length(y)
}

#' Mean squared error
#'
#' @param x numeric vector of observed values
#' @param y numeric vector of reference values
#'
#' @return scalar numeric
#'
#' @seealso \code{\link{rMSE}}
#' @keywords internal
MSE <- function(x, y) {
  sum((x - y)^2) / length(y)
}


# --- Empirical CDF (internal) -------------------------------------------------

#' Empirical cumulative distribution function
#'
#' Computes the ECDF using the Weibull plotting position.
#'
#' @param x numeric vector of values
#'
#' @return a \code{data.table} with columns \code{p} (plotting positions) and
#'   \code{value} (sorted unique values), keyed on \code{value}
#'
#' @seealso \code{\link{fitDist}}
#' @keywords internal
ECDF <- function(x) {

  st <- sort(x)

  aux <- data.table(p     = rank(x = st, ties.method = "first") / (length(st) + 1),
                    value = st,
                    key   = "value")

  J <- value <- NULL

  structure(.Data = aux[J(unique(value)), mult = "last"])
}


# --- Fitting norm (internal) --------------------------------------------------

#' Norm minimised during distribution fitting
#'
#' Computes one of four fitting norms used as the objective function in
#' \code{\link{fitDist}}.
#'
#' @param par numeric vector of distribution parameter values
#' @param val numeric vector of empirical data
#' @param dist character; name of the distribution to fit (e.g. \code{"norm"},
#'   \code{"ggamma"})
#' @param norm character; norm identifier — one of \code{"N1"}, \code{"N2"},
#'   \code{"N3"}, \code{"N4"}
#' @param n.points integer; number of ECDF points used in the norm computation
#'
#' @return scalar numeric; the norm value (returns \code{10e9} if non-finite)
#'
#' @seealso \code{\link{fitDist}}
#' @keywords internal
N <- function(par, val, dist, norm, n.points) {

  edf <- ECDF(val)
  aux <- nrow(edf)

  if (aux != n.points) {
    if (aux < n.points) n.points <- aux
    edf <- edf[seq(from = 1, to = aux, length.out = n.points), ]
  }

  a   <- getDistArg(dist)
  arg <- as.list(setNames(par, a))

  if (norm %in% c("N1", "N2")) {
    FN <- edf$value
    F  <- do.call(paste0("q", dist), args = c(list(edf$p), arg))
    out <- if (norm == "N1") rMSE(F, FN) else MSE(F, FN)
  }

  if (norm %in% c("N3", "N4")) {
    Xi <- edf$p
    Xu <- do.call(paste0("p", dist), args = c(list(edf$value), arg))
    out <- if (norm == "N3") rMSE(Xu, Xi) else MSE(Xu, Xi)
  }

  if (!is.finite(out)) out <- 10e9

  out
}


# --- ACS fitting objective (internal) -----------------------------------------

#' Auxiliary objective function for \code{fitACS}
#'
#' Computes the mean squared error between a theoretical autocorrelation
#' structure and the empirical ACS; passed as \code{fn} to \code{optim} inside
#' \code{\link{fitACS}}.
#'
#' @param par numeric vector of ACS parameter values
#' @param id character; ACS identifier (e.g. \code{"weibull"})
#' @param eACS numeric vector of empirical ACS values
#' @param error character; error criterion — currently only \code{"MSE"}
#'
#' @return scalar numeric; the MSE between theoretical and empirical ACS
#'
#' @seealso \code{\link{fitACS}}
#' @keywords internal
optimACS <- function(par, id, eACS, error = "MSE") {

  arg <- as.list(setNames(par, getACSArg(id)))
  ACS <- do.call(acs, c(list(id = id, t = 0:(length(eACS) - 1)), arg))
  do.call(error, list(x = ACS, y = eACS))
}


# --- Public fitting functions -------------------------------------------------

#' Distribution fitting
#'
#' Fits a parametric distribution to data using the Nelder-Mead simplex
#' algorithm to minimise one of four fitting norms.
#'
#' @param data numeric vector of values to fit
#' @param dist character; distribution name (e.g. \code{"norm"}, \code{"ggamma"})
#' @param n.points integer; number of ECDF points used in norm computation
#' @param norm character; norm identifier — one of \code{"N1"} (ratio RMSE on
#'   quantiles), \code{"N2"} (MSE on quantiles), \code{"N3"} (ratio RMSE on
#'   probabilities), \code{"N4"} (MSE on probabilities)
#' @param constrain logical; if \code{TRUE}, constrains \code{shape2} parameters
#'   to \code{(0, 0.48)} to enforce finite upper tails
#' @param opts list of \code{nloptr} minimisation options
#'
#' @return An object of class \code{"fitDist"}: a named list of fitted
#'   distribution parameters with attributes \code{dist}, \code{edf} (empirical
#'   CDF), and \code{nfo} (full \code{nloptr} output).
#'
#' @seealso \code{\link{fitACS}}, \code{\link{plot.fitDist}}
#'
#' @export
#' @import nloptr
#'
#' @examples
#'
#' x <- fitDist(rnorm(1000), "norm", 30, "N1", FALSE)
#' x
#'
fitDist <- function(data,
                    dist,
                    n.points,
                    norm,
                    constrain,
                    opts = list("algorithm" = "NLOPT_LN_NELDERMEAD",
                                "xtol_rel"  = 1.0e-8,
                                "maxeval"   = 1.0e4)) {

  a   <- getDistArg(dist)
  lwr <- NULL

  if (constrain) {
    if (any(a == "shape2")) {
      start <- c(rep(1, length(a) - 1), .1)
      upr   <- c(rep(Inf, length(a) - 1), .48)
    } else {
      start <- rep(1, length(a))
      upr   <- rep(Inf, length(a))
    }
  } else {
    if (any(a == "shape2")) {
      start <- c(rep(1, length(a) - 1), .1)
      upr   <- rep(Inf, length(a))
    } else {
      start <- rep(1, length(a))
      upr   <- rep(Inf, length(a))
    }
  }

  fit <- nloptr(x0     = start,
                eval_f = N,
                lb     = lwr,
                ub     = upr,
                val    = data,
                dist   = dist,
                n.points = n.points,
                norm   = norm,
                opts   = opts)

  out <- as.list(setNames(fit$solution, a))

  structure(.Data = out,
            dist  = dist,
            edf   = ECDF(data),
            nfo   = fit,
            class = "fitDist")
}


#' Autocorrelation structure fitting
#'
#' Fits a parametric autocorrelation structure (ACS) to empirical ACF values
#' using Nelder-Mead optimisation with MSE criterion.
#'
#' @param acf numeric vector of autocorrelation function values from lag 0
#' @param ID character; ACS identifier (e.g. \code{"weibull"}, \code{"paretoII"})
#' @param start numeric vector of starting parameter values; if \code{NULL},
#'   all parameters start at 1
#' @param lag integer; number of lags to use; if \code{NULL}, lags up to the
#'   first value \eqn{\le 0.01} are used (or all lags if none drops below 0.01)
#'
#' @return An object of class \code{"fitACS"}: a named list of fitted ACS
#'   parameters with attributes \code{ID} and \code{eACS} (empirical ACS used
#'   for fitting).
#'
#' @seealso \code{\link{fitDist}}, \code{\link{plot.fitACS}}, \code{\link{acs}}
#'
#' @export
#' @import stats
#'
#' @examples
#'
#' x <- arima.sim(model = list(ar = 0.8), n = 1000)
#'
#' acsfit <- fitACS(acf(x, plot = FALSE)$acf, "weibull", c(1, 1))
#'
fitACS <- function(acf, ID, start = NULL, lag = NULL) {

  if (is.null(lag)) {
    if (length(which(acf <= .01)) == 0) {
      eACS <- acf
    } else {
      eACS <- acf[1:which(acf <= .01)[1]]
    }
  } else {
    eACS <- acf[1:(lag + 1)]
  }

  if (is.null(start)) start <- rep(1, length(getACSArg(ID)))

  par <- optim(par = start,
               fn  = optimACS,
               id  = ID,
               eACS = eACS,
               error = "MSE")$par

  out <- as.list(setNames(par, getACSArg(ID)))

  structure(.Data = out,
            ID   = ID,
            eACS = eACS,
            class = "fitACS")
}

test_that("test distribution fitting", {
  expect_equal(
    {
      x <- fitDist(rnorm(10000),
        dist = "norm",
        n.points = 40,
        norm = "N2",
        constrain = FALSE
      )
      attributes(x) <- NULL
      x
    },
    list(0, 1),
    tolerance = .5
  )
})

#' CoSMoS: Complete Stochastic Modelling Solution
#'
#' CoSMoS is an R package that makes time series generation with desired
#' properties easy. Just choose the characteristics of the time series you
#' want to generate, and it will do the rest.
#'
#' The generated time series preserve any probability distribution and any
#' linear autocorrelation structure. Users can generate as many and as long
#' time series from processes such as precipitation, wind, temperature,
#' relative humidity etc. It is based on a framework that unified, extended,
#' and improved a modelling strategy that generates time series by transforming
#' "parent" Gaussian time series having specific characteristics
#' (Papalexiou, 2018).
#'
#' @author \strong{Coded by:} Filip Strnad \email{strnadf@fzp.czu.cz} and
#'   Francesco Serinaldi \email{francesco.serinaldi@ncl.ac.uk}
#' @author \strong{Conceptual design by:} Simon Michael Papalexiou
#'   \email{sm.papalexiou@usask.ca}
#' @author \strong{Tested and documented by:} Yannis Markonis
#'   \email{markonis@fzp.czu.cz}
#' @author \strong{Maintained by:} Kevin Shook
#'   \email{kevin.shook@usask.ca}
#'
#' @section Funding:
#' The package was partly funded by the Global Institute for Water Security
#' (GIWS; \href{https://water.usask.ca/}{https://water.usask.ca/}) and the
#' Global Water Futures (GWF; \href{https://gwf.usask.ca/}{https://gwf.usask.ca/})
#' program.
#'
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic modelling
#'   of hydroclimatic processes: Preserving marginal distributions, correlation
#'   structures, and intermittency. Advances in Water Resources 115, 234-252,
#'   \doi{10.1016/j.advwatres.2018.02.013}
#' @references Papalexiou, S.M., Markonis, Y., Lombardo, F., AghaKouchak, A.,
#'   Foufoula-Georgiou, E. (2018). Precise Temporal Disaggregation Preserving
#'   Marginals and Correlations (DiPMaC) for Stationary and Nonstationary
#'   Processes. Water Resources Research, 54(10), 7435-7458,
#'   \doi{10.1029/2018WR022726}
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#'   Preserving Marginal Distributions, Correlations, and Intermittency, With
#'   Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#'   e2019WR026331, \doi{10.1029/2019WR026331}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#'   Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#'   Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @useDynLib CoSMoS, .registration = TRUE
#' @importFrom Rcpp sourceCpp
"_PACKAGE"

#' Advection fields
#'
#' Provides parametric functions that describe different types of advection fields.
#'
#' @param id advection type id (\code{uniform}, \code{rotation}, \code{spiral}, \code{spiralCE}, \code{radial}, and \code{hyperbolic})
#' @param ... other arguments (vector of coordinates and parameters of advection field functions)
#'
#' @name advectionF
#'
#' @import ggplot2 ggquiver
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(ggquiver)
#' library(ggplot2)
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' ## get the advection field
#' af <- advectionF("spiral",
#'                  spacepoints = coord,
#'                  x0 = floor(m / 2),
#'                  y0 = floor(m / 2),
#'                  a = 3,
#'                  b = 2,
#'                  rotation = 1)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionF <- function(id, ...) {

  .args <- list(...)
  do.call(paste0("advectionF", id), args = .args)
}


#' Advection fields
#'
#' Provides parametric functions that describe different types of advection fields.
#'
#' @param id advection type id
#' @param arglist list of additional arguments (vector of coordinates and parameters of advection field functions)
#'
#' @name advectionF2
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @keywords internal
#'
advectionF2 <- function(id, arglist) {

  do.call(paste0("advectionF", id), args = arglist)
}


#' Uniform advection field
#'
#' Provides an advection field with constant orthogonal (u and v) components at each grid point. This mimics rigid translation in a given direction according to the components u and v of the velocity vector.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param u velocity component along the x axis
#' @param v velocity component along the y axis
#'
#' @name advectionFuniform
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(ggquiver)
#' library(ggplot2)
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFuniform(spacepoints = coord,
#'                        u = 2,
#'                        v = 6)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFuniform <- function(spacepoints, u, v) {

  d <- nrow(spacepoints)
  data.frame(u = rep(u, d), v = rep(v, d))
}


#' Rotational advection field
#'
#' Provides an advection field corresponding to rotation around a specified center.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param x0 x coordinate of the center of rotation
#' @param y0 y coordinate of the center of rotation
#' @param a parameter controlling the x component of rotational velocity
#' @param b parameter controlling the y component of rotational velocity
#'
#' @details # Note
#' \itemize{
#' \item if a > 0, b > 0: clockwise rotation around (x0, y0)
#' \item if a < 0, b < 0: counter-clockwise rotation around (x0, y0)
#' }
#'
#' @name advectionFrotation
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(ggquiver)
#' library(ggplot2)
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFrotation(spacepoints = coord,
#'                         x0 = floor(m / 2),
#'                         y0 = floor(m / 2),
#'                         a = 3,
#'                         b = 2)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFrotation <- function(spacepoints, x0, y0, a, b) {

  x <- spacepoints[,1] - x0
  y <- spacepoints[,2] - y0
  u <-    a * y
  v <- (-b * x)
  data.frame(u = u, v = v)
}


#' Spiraling advection field
#'
#' Provides an advection field corresponding to a spiral motion to/from a specified reference point (sink).
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param x0 x coordinate of reference point (sink)
#' @param y0 y coordinate of reference point (sink)
#' @param a parameter controlling the x component of rotational velocity
#' @param b parameter controlling the y component of rotational velocity
#' @param rotation parameter controlling the rotational direction. The following combinations hold: \cr
#' \itemize{
#' \item if a > 0, b > 0, and direction = 1: spiraling CLOCKWISE TO (x0, y0)
#' \item if a < 0, b < 0, and direction = 1: spiraling COUNTER-CLOCKWISE FROM (x0, y0)
#' \item if a > 0, b > 0, and direction = 2: spiraling COUNTER-CLOCKWISE TO (x0, y0)
#' \item if a < 0, b < 0, and direction = 2: spiraling CLOCKWISE FROM (x0, y0)
#' }
#'
#' @name advectionFspiral
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(ggquiver)
#' library(ggplot2)
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFspiral(spacepoints = coord,
#'                         x0 = floor(m / 2),
#'                         y0 = floor(m / 2),
#'                         a = 3,
#'                         b = 2,
#'                         rotation = 1)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFspiral <- function(spacepoints, x0, y0, a, b, rotation = 1) {

  x <- spacepoints[,1] - x0
  y <- spacepoints[,2] - y0
  if (rotation == 1) {
    u <-  ( b * y - a * x)
    v <-  (-a * x - b * y)
  }
  if (rotation == 2) {
    u <-  (-a * x - b * y)
    v <-  ( a * x - b * y)
  }
  data.frame(u = u, v = v)
}



#' Spiraling advection field satisfying continuity equation
#'
#' Provides an advection field corresponding to a spiral motion to/from a specified reference point (sink) satisfying continuity equation (from \href{https://github.com/johannesgerer/jburkardt-f/tree/2a3860f296af9ab0a07c6dc805b496bc75ad3935/spiral_data}{Git Mirror of John Burkardt's collection of FORTRAN 90 Software}).
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param a parameter controlling the intensity of rotational velocity (a > 0 clokwise; a < 0 conter-clockwise)
#' @param C parameter ranging in (0, 2*pi)
#'
#' @name advectionFspiralCE
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#' library(ggquiver)
#' library(ggplot2)
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFspiralCE(spacepoints = coord,
#'                         a = 5,
#'                         C = 1)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFspiralCE <- function(spacepoints, a, C) {

  xaux <- spacepoints[,1]
  yaux <- spacepoints[,2]
  x <- (xaux - min(xaux)) / diff(range(xaux))
  y <- (yaux - min(yaux)) / diff(range(yaux))
  dPPdx <- (sin(C * pi * x) * (C * pi) * (1 - x)^2 - (1 - cos(C * pi * x)) *
              (2 * (1 - x))) * ((1 - cos(C * pi * y)) * (1 - y)^2)
  dPPdy <- ((1 - cos(C * pi * x)) * (1 - x)^2) * (sin(C * pi * y) * (C * pi) *
                                                    (1 - y)^2 - (1 - cos(C * pi * y)) * (2 * (1 - y)))
  u <- -a * dPPdy
  v <- a * dPPdx

  data.frame(u = u, v = v)
}


#' Radial advection field
#'
#' Provides an advection field corresponding to radial motion from or towards a specified reference point.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param x0 x coordinate of the center of radial motion
#' @param y0 y coordinate of the center of radial motion
#' @param a parameter controlling the x component of radial velocity
#' @param b parameter controlling the y component of radial velocity
#'
#' @details # Note
#' \itemize{
#' \item if a > 0, b > 0: divergence from (x0, y0) (source point effect)
#' \item if a < 0, b < 0: convergence to (x0, y0) (sink effect)
#' }
#'
#' @name advectionFradial
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#' library(ggquiver)
#' library(ggplot2)
#'
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFradial(spacepoints = coord,
#'                         x0 = floor(m / 2),
#'                         y0 = floor(m / 2),
#'                         a = 3,
#'                         b = 2)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFradial <- function(spacepoints, x0, y0, a, b) {

  x <- spacepoints[,1] - x0
  y <- spacepoints[,2] - y0
  u <- a * x
  v <- b * y
  data.frame(u = u, v = v)
}



#' Hyperbolic advection field
#'
#' Provides an advection field with hyperbolic trajectories.
#'
#' @param spacepoints vector of coordinates (2 x d), where d is the number of locations/grid points
#' @param x0 x coordinate of the center of hyperbola
#' @param y0 y coordinate of the center of hyperbola
#' @param a parameter controlling the x component of rotational velocity
#' @param b parameter controlling the y component of rotational velocity
#'
#' @details # Note
#' \itemize{
#' \item if a > 0, b > 0: toward bottom-left and top-right corner
#' \item if a < 0, b < 0: toward top-left and bottom-right corner
#' }
#'
#' @name advectionFhyperbolic
#'
#' @import ggplot2
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#' library(ggquiver)
#' library(ggplot2)
#' ## specify coordinates
#' m = 25
#' aux <- seq(0, m - 1, length = m)
#' coord <- expand.grid(aux, aux)
#'
#' af <- advectionFhyperbolic(spacepoints = coord,
#'                            x0 = floor(m / 2),
#'                            y0 = floor(m / 2),
#'                            a = 3,
#'                            b = 2)
#'
#' ## visualize advection field
#' dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
#' ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
#' geom_quiver() +
#' theme_light()
#'
advectionFhyperbolic <- function(spacepoints, x0, y0, a, b) {

  x <- spacepoints[,1] - x0
  y <- spacepoints[,2] - y0
  u <- a * y
  v <- b * x
  data.frame(u = u, v = v)
}

#' SpatioTemporal Correlation Structure
#'
#' Provides a parametric function that describes the values of the linear
#' spatiotemporal autocorrelation up to desired lags.
#' For more details on the parametric spatiotemporal correlation structures see
#' section 2.3 and 2.4 in Papalexiou and Serinaldi (2020).
#'
#' @param id spatiotemporal correlation structure ID
#' @param ... additional arguments (t as time lag, s as spatial lag (distance),  and stcs parameters)
#'
#' @name stcs
#'
#' @import plot3D
#'
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(plot3D)
#'
#' ## specify grid of spatial and temporal lags
#' d <- 31
#' st <- expand.grid(0:(d-1),
#'                   0:(d-1))
#'
#' ## get the STCS
#' wc <- stcs("clayton",
#'            t = st[, 1],
#'            s = st[, 2],
#'            scfid = "weibull",
#'            tcfid = "weibull",
#'            copulaarg = 2,
#'            scfarg = list(scale = 20,
#'                          shape = 0.7),
#'            tcfarg = list(scale = 1.1,
#'                          shape = 0.8))
#'
#' g14 <- stcs("gneiting14",
#'             t = st[, 1],
#'             s = st[, 2],
#'             a = 1/50,
#'             c = 1/10,
#'             alpha = 1,
#'             beta = 1,
#'             gamma = 0.5,
#'             tau = 1)
#'
#' g16 <- stcs("gneiting16",
#'             t = st[, 1],
#'             s = st[, 2],
#'             a = 1/50,
#'             c = 1/10,
#'             alpha = 1,
#'             beta = 1,
#'             nu = 0.5,
#'             tau = 1)
#'
#' ## note: for nu = 0.5 stcfgneiting16 is equivalent to
#' ## stcfgneiting14 with gamma = 0.5
#'
#' ## visualize the STCS
#'
#' wc.m <- matrix(wc,
#'                nrow = d)
#'
#' persp3D(z = wc.m, x = 1: nrow(wc.m), y = 1:ncol(wc.m),
#'         expand = 1, main = "", scale = TRUE, facets = TRUE,
#'         xlab="Time lag", ylab = "Distance", zlab = "STCF",
#'         colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
#'         resfac = 5,  col= gg2.col(100))
#'
#' g14.m <- matrix(g14,
#'                 nrow = d)
#'
#' persp3D(z = g14.m, x = 1: nrow(wc.m), y = 1:ncol(wc.m),
#'         expand = 1, main = "", scale = TRUE, facets = TRUE,
#'         xlab="Time lag", ylab = "Distance", zlab = "STCF",
#'         colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
#'         resfac = 5,  col= gg2.col(100))
#'
stcs <- function(id, ...) {

    .args <- list(...)
    do.call(paste0("stcf", id), args = .args)
}

#' SpatioTemporal Correlation Structure
#'
#' Provides a parametric function that describes the values of the linear
#' spatiotemporal autocorrelation up to desired lags.
#' For more details on the parametric spatiotemporal correlation structures see
#' section 2.3 and 2.4 in Papalexiou and Serinaldi (2020).
#'
#' @param id spatiotemporal correlation structure ID
#' @param arglist list of additional arguments (t as time lag, s as spatial lag (distance),  and stcs parameters)
#'
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @keywords internal
#'
stcs2 <- function (id, arglist) {

    do.call(paste0("stcf", id), args = arglist)
}



#' Clayton SpatioTemporal Correlation Structure
#'
#' Provides spatiotemporal correlation structure function based on Clayton copula.
#' For more details on the parametric spatiotemporal correlation structures
#' see section 2.3 and 2.4 in Papalexiou and Serinaldi (2020).
#'
#' @param t time lag
#' @param s spatial lag (distance)
#' @param scfid ID of the spatial (marginal) correlation structure (e.g. weibull)
#' @param tcfid ID of the temporal (marginal) correlation structure (e.g. weibull)
#' @param copulaarg parameter of the Clayton copula linking the marginal correlation structures
#' @param scfarg parameters of spatial (marginal) correlation structure
#' @param tcfarg parameters of temporal (marginal) correlation structure
#'
#' @name stcfclayton
#'
#' @import plot3D
#'
#' @export
#'
#' @references Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#'
#' library(plot3D)
#'
#' ## specify grid of spatial and temporal lags
#' d <- 31
#' st <- expand.grid(0:(d - 1),
#'                   0:(d - 1))
#'
#' ## get the STCS
#' wc <- stcfclayton(t = st[, 1],
#'                   s = st[, 2],
#'                   scfid = "weibull",
#'                   tcfid = "weibull",
#'                   copulaarg = 2,
#'                   scfarg = list(scale = 20,
#'                                 shape = 0.7),
#'                   tcfarg = list(scale = 1.1,
#'                                 shape = 0.8))
#'
#' ## visualize the STCS
#' wc.m <- matrix(wc,
#'                nrow = d)
#'
#' persp3D(z = wc.m, x = 1: nrow(wc.m), y = 1:ncol(wc.m),
#'         expand = 1, main = "", scale = TRUE, facets = TRUE,
#'         xlab="Time lag", ylab = "Distance", zlab = "STCF",
#'         colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
#'         resfac = 5,  col= gg2.col(100))
#'
stcfclayton <- function (t, s, scfid, tcfid, copulaarg, scfarg, tcfarg) {

    if (copulaarg < 0) {
        return(NaN)
    }
    else {
        scfarg$t <- s
        tcfarg$t <- t
        scf <- do.call(paste0("acf", scfid), args = scfarg)
        tcf <- do.call(paste0("acf", tcfid), args = tcfarg)
        if(copulaarg > 0) stcf <- (tcf^(-copulaarg) + scf^(-copulaarg) - 1 )^(-1/copulaarg)
        if(copulaarg == 0) stcf <- tcf * scf
        return(stcf)
    }
}

#' Gneiting-14 SpatioTemporal Correlation Structure
#'
#' Provides spatiotemporal correlation structure function proposed by
#' Gneiting (2002; Eq.14 at p. 593).
#'
#' @param t time lag
#' @param s spatial lag (distance)
#' @param a nonnegative scaling parameter of time
#' @param c nonnegative scaling parameter of space
#' @param alpha smoothness parameter of time. Valid range: \eqn{(0,1]}
#' @param gamma smoothness parameter of space. Valid range: \eqn{(0,1]}
#' @param beta space-time interaction parameter. Valid range: \eqn{[0,1]}
#' @param tau space-time interaction parameter. Valid range: \eqn{\ge 1} (for 2-dimensional fields)
#'
#' @name stcfgneiting14
#'
#' @import plot3D
#'
#' @export
#'
#' @references Gneiting, T. (2002). Nonseparable, Stationary Covariance Functions
#' for Space-Time Data, Journal of the American Statistical Association,
#' 97:458, 590-600, \doi{10.1198/016214502760047113}
#'
#' @examples
#'
#' library(plot3D)
#'
#' ## specify grid of spatial and temporal lags
#' d <- 31
#' st <- expand.grid(0:(d - 1),
#'                   0:(d - 1))
#'
#' ## get the STCS
#' g14 <- stcfgneiting14(t = st[, 1],
#'                       s = st[, 2],
#'                       a = 1/50,
#'                       c = 1/10,
#'                       alpha = 1,
#'                       beta = 1,
#'                       gamma = 0.5,
#'                       tau = 1)
#'
#' ## visualize the STCS
#'
#' g14.m <- matrix(g14,
#'                 nrow = d)
#'
#' persp3D(z = g14.m, x = 1: nrow(g14.m), y = 1:ncol(g14.m),
#'         expand = 1, main = "", scale = TRUE, facets = TRUE,
#'         xlab="Time lag", ylab = "Distance", zlab = "STCF",
#'         colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
#'         resfac = 5,  col= gg2.col(100))
#'
stcfgneiting14 <- function (t, s, a, c, alpha, beta, gamma, tau) {

    if((a < 0) | (c < 0) | (alpha <= 0) | (alpha > 1) |
       (beta < 0) | (beta > 1) | (gamma <= 0) | (gamma > 1) |
       (tau < 1)) {

        return(NaN)
    } else {

        stcf <- 1 / ((a * abs(t)^(2 * alpha) + 1)^tau) *
            exp( - (c * abs(s)^(2 * gamma)) / ((a * abs(t)^(2 * alpha) + 1)^(beta * gamma)) )

        return(stcf)
    }
}

#' Gneiting-16 SpatioTemporal Correlation Structure
#'
#' Provides spatiotemporal correlation structure function proposed by
#' Gneiting (2002; Eq.16 at p. 594).
#'
#' @param t time lag
#' @param s spatial lag (distance)
#' @param a nonnegative scaling parameter of time
#' @param c nonnegative scaling parameter of space
#' @param alpha smoothness parameter of time. Valid range: \eqn{(0,1]}
#' @param nu smoothness parameter of space. Valid range: \eqn{>0}
#' @param beta space-time interaction parameter. Valid range: \eqn{[0,1]}
#' @param tau space-time interaction parameter. Valid range: \eqn{\ge 1} (for 2-dimensional fields)
#'
#' @name stcfgneiting16
#'
#' @import plot3D
#'
#' @export
#'
#' @references Gneiting, T. (2002). Nonseparable, Stationary Covariance Functions
#' for Space-Time Data, Journal of the American Statistical Association,
#' 97:458, 590-600, \doi{10.1198/016214502760047113}
#'
#' @examples
#'
#' library(plot3D)
#'
#' ## specify grid of spatial and temporal lags
#' d <- 31
#' st <- expand.grid(0:(d - 1),
#'                   0:(d - 1))
#'
#' ## get the STCS
#' g16 <- stcfgneiting16(t = st[, 1],
#'                       s = st[, 2],
#'                       a = 1/50,
#'                       c = 1/10,
#'                       alpha = 1,
#'                       beta = 1,
#'                       nu = 0.5, tau = 1)
#'
#' ## visualize the STCS
#'
#' g16.m <- matrix(g16,
#'                 nrow = d)
#'
#' persp3D(z = g16.m, x = 1: nrow(g16.m), y = 1:ncol(g16.m),
#'         expand = 1, main = "", scale = TRUE, facets = TRUE,
#'         xlab="Time lag", ylab = "Distance", zlab = "STCF",
#'         colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
#'         resfac = 5,  col= gg2.col(100))
#'
stcfgneiting16 <- function (t, s, a, c, alpha, beta, nu, tau) {

    if((a < 0) | (c < 0) | (alpha <= 0) | (alpha > 1) |
       (beta < 0) | (beta > 1) | (nu <= 0) | (tau < 1)) {
        return(NaN)
    }
    else {
        stcf <- 1 / (2^(nu - 1)  * gamma(nu) * (a * abs(t)^(2 * alpha) + 1)^tau) *
            ((c * abs(s)) / ((a * abs(t)^(2 * alpha) + 1)^(beta / 2) ))^nu *
            besselK((c * abs(s)) / ((a * abs(t)^(2 * alpha) + 1)^(beta / 2) ), nu)
        return(stcf)
    }
}

# Internal utility functions for CoSMoS
# These functions are not exported and not intended for direct use.

#' Compute spectral values for DHM circulant-embedding simulation
#'
#' Computes the one-sided spectral values \eqn{S_j} from an autocovariance
#' function (ACVF) vector using the Dietrich-Newsam-Haskard circulant-embedding
#' approach.
#'
#' @param acvf numeric vector of autocovariance values starting at lag 0
#'
#' @return numeric vector of \eqn{S_j} values (length = \code{length(acvf)})
#'
#' @keywords internal
DHMgenSj <- function(acvf) {
  MM <- length(acvf) - 1
  Sj <- Re(fft(c(acvf, rev(acvf[2:MM]))))[1:(MM + 1)]
  Sj <- abs(Sj)
  if (!all(Sj >= 0))
    stop("some of the S_j's are negative")
  Sj
}

#' Simulate one realisation using DHM circulant-embedding
#'
#' Simulates one realisation of a stationary Gaussian process from precomputed
#' spectral values \eqn{S_j} via the Dietrich-Newsam-Haskard algorithm.
#'
#' @param Sj numeric vector of spectral values from \code{DHMgenSj}
#' @param rn optional vector of standard normal random numbers of length
#'   \code{2 * length(Sj) - 2}; generated internally if \code{NULL}
#'
#' @return numeric vector of length \code{length(Sj) - 1}
#'
#' @keywords internal
DHMgenSim <- function(Sj, rn = NULL) {
  if (is.null(rn))
    rn <- rnorm(2 * length(Sj) - 2)
  M  <- 2 * length(Sj) - 2
  N  <- M / 2
  Yj <- 0:N
  Yj[1]     <- sqrt(Sj[1]) * rn[1]
  Yj[N + 1] <- sqrt(Sj[N + 1]) * rn[M]
  js         <- 2:N
  Yj[js]    <- sqrt(0.5 * Sj[js]) *
    complex(real      = rn[2 * (1:(N - 1))],
            imaginary = rn[2 * (1:(N - 1)) + 1])
  Re(fft(c(Yj, Conj(rev(Yj[js])))))[1:N] / sqrt(M)
}


#' L-moments
#'
#' Computes the first four L-moments of a sample using probability-weighted
#' moments.
#'
#' @param x numeric vector of values
#'
#' @return named numeric vector of length 4: \code{l1}, \code{l2},
#'   \code{l3} (L-skewness ratio), \code{l4} (L-kurtosis ratio)
#'
#' @seealso \code{\link{reportTS}}
#' @keywords internal
lmom <- function(x) {

  x  <- sort(x)
  n  <- length(x)
  nn <- rep(n - 1, n)
  pp <- seq(0, n - 1)

  p1 <- pp / nn
  p2 <- p1 * (pp - 1) / (nn - 1)
  p3 <- p2 * (pp - 2) / (nn - 2)

  b0 <- sum(x) / n
  b1 <- sum(p1 * x) / n
  b2 <- sum(p2 * x) / n
  b3 <- sum(p3 * x) / n

  l1 <- b0
  l2 <- 2 * b1 - b0
  l3 <- 2 * (3 * b2 - b0) / (2 * b1 - b0) - 3
  l4 <- 5 * (2 * (2 * b3 - 3 * b2) + b0) / (2 * b1 - b0) + 6

  unlist(mget(paste0("l", 1:4)))
}


#' Complementary error function
#'
#' @param x numeric vector
#'
#' @return numeric vector
#'
#' @rdname erfc
#' @keywords internal
erfc <- function(x) {
  2 * pnorm(-sqrt(2) * x)
}

#' Inverse complementary error function
#'
#' @rdname erfc
#' @keywords internal
inv.erfc <- function(x) {
  x[x < 0 | x > 2] <- NA
  -qnorm(x / 2) / sqrt(2)
}

test_that("test sample l moments", {
  expect_equal(
    lmom(rnorm(10000,
      mean = 2,
      sd = 1
    )),
    c(
      l1 = 2.01,
      l2 = 0.55,
      l3 = 0.00,
      l4 = 0.12
    ),
    tolerance = .2
  )
})

#' Check generated time series
#'
#' Compares sample statistics of generated time series against theoretically
#' expected values.
#'
#' @param TS a \code{cosmosts} object from \code{\link{generateTS}}, or a list
#'   of numeric vectors, or a single numeric vector
#' @inheritParams moments
#'
#' @return An object of class \code{c("checkTS", "matrix")} with rows
#'   \code{"expected"} and one row per simulated series, and columns for
#'   \code{mean}, \code{sd}, \code{skew}, \code{p0}, \code{acf_t1},
#'   \code{acf_t2}, \code{acf_t3}. Attributes \code{margdist}, \code{margarg},
#'   and \code{p0} are attached for use by \code{\link{plot.checkTS}}.
#'
#' @seealso \code{\link{generateTS}}, \code{\link{plot.checkTS}},
#'   \code{\link{moments}}
#'
#' @export
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' x <- generateTS(margdist = "burrXII",
#'                 margarg = list(scale = 1,
#'                                shape1 = .75,
#'                                shape2 = .25),
#'                 acsvalue = acs(id = "weibull",
#'                                t = 0:30,
#'                                scale = 10,
#'                                shape = .75),
#'                 n = 1000, p = 30, p0 = .5, TSn = 5)
#'
#' checkTS(x)
#'
checkTS <- function(TS, distbounds = c(-Inf, Inf)) {

  if (!is.list(TS)) TS <- list(TS)

  att      <- attributes(TS[[1]])
  margdist <- att$margdist
  margarg  <- att$margarg
  p0       <- att$p0
  acsvalue <- att$acsvalue

  ac <- sapply(TS, function(x) acf(x, plot = FALSE)$acf)

  out <- data.frame(
    mean = c(popmean(margdist, margarg, distbounds = distbounds, p0 = p0),
             sapply(TS, mean)),
    sd   = c(popsd(margdist, margarg, distbounds = distbounds, p0 = p0),
             sapply(TS, sd)),
    skew = c(popskew(margdist, margarg, distbounds = distbounds, p0 = p0),
             sapply(TS, function(x) sample.moments(x, raw = FALSE,
                                                   central = FALSE,
                                                   coef = TRUE)$coefficients[2])),
    p0     = c(p0, sapply(TS, function(x) length(which(x == 0)) / length(x))),
    acf_t1 = c(acsvalue[2], ac[2, ]),
    acf_t2 = c(acsvalue[3], ac[3, ]),
    acf_t3 = c(acsvalue[4], ac[4, ])
  )

  row.names(out) <- c("expected", paste0("simulation", seq_along(TS)))

  structure(.Data    = as.matrix(round(out, 2)),
            class    = c("checkTS", "matrix"),
            margdist = margdist,
            margarg  = margarg,
            p0       = p0)
}

# --- Internal seasonal helpers ------------------------------------------------

#' Stratify a time series by season
#'
#' Splits a two-column (date, value) time series into a list of seasonal
#' subsets, plus a parallel list of nonzero-only subsets.
#'
#' @param TS data frame or data table with columns \code{date} and \code{value}
#' @param season character; name of a date-component function
#'   (e.g. \code{"month"}, \code{"week"})
#'
#' @return a list of two elements: \code{[[1]]} named list of full seasonal
#'   subsets; \code{[[2]]} named list of nonzero seasonal subsets
#'
#' @seealso \code{\link{analyzeTS}}
#' @keywords internal
stratifySeasonData <- function(TS, season) {

  TS    <- as.data.frame(TS)
  strat <- split(TS, do.call(season, list(x = TS[, "date"])))
  names(strat) <- paste("data", seq_along(strat), sep = "_")

  nz <- lapply(strat, function(x) x[x[, "value"] > 0, ])
  names(nz) <- paste("data_nz", seq_along(nz), sep = "_")

  structure(.Data = list(strat, nz))
}


#' Calculate seasonal autocorrelation function
#'
#' Computes the empirical ACF for each season of a time series.
#'
#' @param TS data frame or data table with columns \code{date} and \code{value}
#' @param season character; name of a date-component function
#'   (e.g. \code{"month"})
#' @param lag.max integer; maximum lag for the ACF (default 50)
#'
#' @return a named list (one element per season) of numeric ACF vectors
#'   starting at lag 0
#'
#' @seealso \code{\link{analyzeTS}}, \code{\link{fitACS}}
#' @keywords internal
seasonalACF <- function(TS, season, lag.max = 50) {

  TS <- as.data.table(TS)
  TS[, season := do.call(season, list(x = date))]

  lag0 <- NULL
  TS[, lag0 := .I]
  for (i in 1:lag.max) TS[, paste0("lag", i) := lag0 - i]

  out <- lapply(unique(TS[, season]), function(j) {

    index <- as.data.frame(
      TS[season == j, .SD, .SDcols = grep("lag", names(TS))]
    )

    as <- sapply(1:lag.max, function(i) {
      xi <- index[index[paste0("lag", i)] > 0, "lag0"]
      yi <- index[index[paste0("lag", i)] > 0, paste0("lag", i)]
      x  <- unlist(TS[xi, "value"])
      y  <- unlist(TS[yi, "value"])
      cor(x, y, use = "p")
    })

    c(1, as)
  })

  names(out) <- paste(season, unique(TS[, season]))
  out
}


#' Seasonal AR model
#'
#' Generates a seasonal Gaussian AR process over a vector of dates, one season
#' at a time, using precomputed ACS-derived AR coefficients.
#'
#' @param x POSIXct vector of dates for which to generate the process
#' @param ACS named list of ACS vectors (one per season, starting at lag 0)
#' @param season character; name of a date-component function (default
#'   \code{"month"})
#'
#' @return a \code{data.table} with columns \code{date}, \code{gauss}
#'   (Gaussian process values), and \code{season} (season index)
#'
#' @seealso \code{\link{simulateTS}}, \code{\link{analyzeTS}}
#' @keywords internal
#' @import data.table
seasonalAR <- function(x, ACS, season = "month") {

  time <- data.table(time = x)
  y <- s <- n <- . <- id <- value <- NULL

  time[, y := year(time)]
  time[, s := do.call(season, .(time))]
  time[, n := .N, by = .(y, s)]

  d     <- as.data.frame(unique(time[, -1]))
  alpha <- lapply(ACS, YW)

  esd <- vapply(seq_along(alpha), function(i) {
    sqrt(1 - sum(alpha[[i]] * ACS[[i]][2:length(ACS[[i]])]))
  }, numeric(1))

  ## initialise with a short AR(1) burn-in
  init_season <- which(gsub(paste(season, ""), "", names(ACS)) == d[1, "s"])
  init_val    <- AR1(max(sapply(alpha, length)),
                     ACS[[init_season]][2])

  ## pre-allocate output list then rbind once
  chunks <- vector("list", nrow(d))

  for (j in seq_len(nrow(d))) {

    s_j  <- d[j, "s"]
    ss   <- which(gsub(paste(season, ""), "", names(alpha)) == s_j)
    p    <- length(alpha[[ss]])
    n_j  <- d[j, "n"]

    if (j == 1L) {
      val <- c(init_val, numeric(n_j))
    } else {
      ## carry over last p values from previous chunk
      prev <- unlist(lapply(chunks[seq_len(j - 1)], function(ch) ch$value))
      val  <- c(tail(prev, p), numeric(n_j))
    }

    gn    <- rnorm(n_j + p, mean = 0, sd = esd[s_j])
    a_rev <- rev(alpha[[ss]])

    for (i in (p + 1):(n_j + p)) {
      val[i] <- sum(val[(i - p):(i - 1)] * a_rev) + gn[i]
    }

    chunks[[j]] <- data.table(value = val[(p + 1):(n_j + p)], id = s_j)
  }

  out <- rbindlist(chunks)

  data.table(date   = x,
             gauss  = out[, value],
             season = out[, id])
}


# --- Public analysis pipeline -------------------------------------------------

#' Analyse, report, and simulate seasonal time series
#'
#' \code{analyzeTS} automatically performs seasonal analysis, fits distributions
#' and correlation structures. \code{reportTS} visualises the fitted
#' distributions and correlation structures, or returns a table of fitted
#' parameters and descriptive statistics. \code{simulateTS} takes the result of
#' \code{analyzeTS} and generates synthetic realisations.
#'
#' @details
#' In practice, we typically want to simulate a natural process from observed
#' data. \code{analyzeTS} fits a marginal distribution and autocorrelation
#' structure for each season; \code{reportTS} lets you inspect the fit;
#' \code{simulateTS} generates synthetic time series with the same seasonal
#' statistical properties.
#'
#' Recommended distributions by variable type:
#' * *precipitation / streamflow*: \code{ggamma}, \code{burrXII}, \code{burrIII}
#' * *relative humidity*: \code{beta}
#' * *temperature*: \code{norm}
#'
#' @param TS data frame or data table with columns \code{date} and \code{value}
#' @param season character; name of a date-component function
#'   (e.g. \code{"month"}, \code{"week"})
#' @param acsID character; ACS identifier passed to \code{\link{fitACS}}
#' @param lag.max integer; maximum lag for the empirical ACF
#' @param aTS an \code{analyzeTS} result object
#' @param method character; report type — \code{"dist"} for distribution fits,
#'   \code{"acs"} for ACS fits, \code{"stat"} for descriptive statistics table
#' @param from POSIXct; start of simulation period (defaults to start of
#'   observed series)
#' @param to POSIXct; end of simulation period (defaults to end of observed
#'   series)
#' @inheritParams N
#' @inheritParams fitDist
#'
#' @return
#' * \code{analyzeTS}: a list with elements \code{data}, \code{dfits},
#'   \code{afits}, and attributes \code{season}, \code{dist}, \code{acsID},
#'   \code{date}
#' * \code{reportTS}: a \code{ggplot} object (\code{"dist"} or \code{"acs"}
#'   method) or a \code{data.frame} (\code{"stat"} method)
#' * \code{simulateTS}: a \code{data.table} with columns \code{date} and
#'   \code{value}
#'
#' @seealso \code{\link{fitDist}}, \code{\link{fitACS}}, \code{\link{generateTS}}
#'
#' @rdname analyzeTS
#' @export
#' @import ggplot2 data.table
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## Load data included in the package
#' data("precip")
#' \donttest{
#' ## Fit seasonal ACSs and distributions to the data
#' a <- analyzeTS(precip)
#'
#' reportTS(a, "dist")  ## seasonal distribution fits
#' reportTS(a, "acs")   ## seasonal ACS fits
#' reportTS(a, "stat")  ## descriptive statistics
#'
#' ## Simulate a time series of the same length
#' sim <- simulateTS(a)
#'
#' precip[, id := "observed"]
#' sim[, id := "simulated"]
#' dta <- rbind(precip, sim)
#'
#' ggplot(dta) +
#'   geom_line(aes(x = date, y = value)) +
#'   facet_wrap(~id, ncol = 1) +
#'   theme_classic()
#'
#' ## Simulate a time series of different length
#' sim <- simulateTS(a,
#'                   from = as.POSIXct("1978-12-01 00:00:00"),
#'                   to   = as.POSIXct("2008-12-01 00:00:00"))
#' }
#' \dontshow{
#' precip <- precip[between(date,
#'   as.POSIXct("1990-1-01", format("%Y-%m-%d"), tz = "America/Regina"),
#'   as.POSIXct("1990-1-5",  format("%Y-%m-%d"), tz = "America/Regina"))]
#' a <- analyzeTS(precip, opts = list("algorithm" = "NLOPT_LN_NELDERMEAD",
#'                                    "maxeval" = 10))
#' }
#'
analyzeTS <- function(TS, season = "month", dist = "ggamma", acsID = "weibull",
                      norm = "N1", n.points = 30, lag.max = 30,
                      constrain = FALSE, opts = NULL) {

  if (is.null(opts)) opts <- formals(fitDist)$opts

  ea <- seasonalACF(TS, season = season, lag.max = lag.max)
  a  <- lapply(ea, function(x) fitACS(x, acsID))

  x  <- stratifySeasonData(TS, season)
  f  <- lapply(x[[2]], function(x) {
    fitDist(x$value, dist,
            norm      = norm,
            n.points  = n.points,
            constrain = constrain,
            opts      = opts)
  })

  structure(.Data = list(data = x, dfits = f, afits = a),
            season = season,
            dist   = dist,
            acsID  = acsID,
            date   = TS[, "date"])
}


#' @rdname analyzeTS
#' @export
reportTS <- function(aTS, method = "dist") {

  dist   <- attr(aTS, "dist")
  acsID  <- attr(aTS, "acsID")
  season <- attr(aTS, "season")

  if (method == "stat") {

    nz <- aTS$data[[2]]

    dp  <- as.data.frame(round(t(sapply(aTS$dfits, function(x) do.call(rbind, x))), 3))
    names(dp) <- getDistArg(dist)

    ap  <- as.data.frame(round(t(sapply(aTS$afits, function(x) do.call(rbind, x))), 3))
    names(ap) <- getACSArg(acsID)

    laux <- t(sapply(nz, function(x) lmom(x$value)))

    l <- round(data.frame(l.var  = laux[, 1] / laux[, 2],
                           l.skew = laux[, 3],
                           l.kurt = laux[, 4]), 2)

    s <- t(round(sapply(nz, function(x) {
      c(mean = mean(x$value, na.rm = TRUE),
        sd   = sd(x$value,   na.rm = TRUE),
        min  = min(x$value,  na.rm = TRUE),
        q    = quantile(x$value, na.rm = TRUE, probs = .05),
        q    = quantile(x$value, na.rm = TRUE, probs = .25),
        q    = quantile(x$value, na.rm = TRUE, probs = .5),
        q    = quantile(x$value, na.rm = TRUE, probs = .75),
        q    = quantile(x$value, na.rm = TRUE, probs = .95),
        max  = max(x$value,  na.rm = TRUE),
        skew = sample.moments(x$value, raw = FALSE, central = FALSE,
                              coef = TRUE)$coefficients[2])
    }), 2))

    err <- round(sapply(aTS$dfits, function(x) attr(x, "nfo")$objective), 4)

    p0 <- 1 - round(
      sapply(nz,                  nrow) /
      sapply(aTS$data[[1]], nrow), 2)

    a <- round(as.data.frame(
      t(sapply(aTS$afits, function(x) attr(x, "eACS")[2:5]))), 2)
    names(a) <- paste0("acs.l.", 2:5)

    stat <- cbind(dist = dist, dp,
                  error = err,
                  acsID = acsID, ap,
                  s, l, p0, a)
    rownames(stat) <- paste(season, seq_len(nrow(stat)), sep = "_")

    out <- stat
  }

  if (method == "dist") {

    f <- aTS$dfits

    CDF <- lapply(f, function(x) {
      edf  <- attr(x, "edf")
      dist <- attr(x, "dist")
      mm   <- do.call(paste0("p", dist), c(list(q = range(edf$value)), x))
      p    <- exp(seq(ifelse(!is.finite(log(mm[1])), .00001, log(mm[1])),
                      log(mm[2]), length.out = 10000))
      data.frame(p     = p,
                 value = do.call(paste0("q", dist), c(list(p = p), x)))
    })
    names(CDF) <- paste(season, seq_along(CDF), sep = "_")

    sea <- NULL
    cdf <- rbindlist(CDF, idcol = "sea")
    cdf[, sea := factor(sea, levels = paste(season, seq_along(CDF), sep = "_"))]

    EDF <- lapply(f, function(x) attr(x, "edf"))
    names(EDF) <- paste(season, seq_along(EDF), sep = "_")

    edf <- rbindlist(EDF, idcol = "sea")
    edf[, sea := factor(sea, levels = paste(season, seq_along(EDF), sep = "_"))]

    out <- ggplot() +
      geom_point(data = edf,
                 aes(x = edf$value, y = log(1 - edf$p), shape = factor(2)),
                 colour = "royalblue4", alpha = .5) +
      scale_shape_manual(values = 19, label = "Empirical", name = NULL) +
      geom_line(data = cdf,
                aes(x = cdf$value, y = log(1 - cdf$p), linetype = factor(1)),
                colour = "red4", lwd = .5, alpha = .75) +
      scale_linetype_manual(values = 1, label = "Fitted", name = NULL) +
      scale_y_continuous(
        breaks = seq(-10, 0, length.out = 5),
        labels = format(exp(seq(log(.0001), log(1), length.out = 5)),
                        scientific = TRUE)) +
      labs(x = "Nonzero values", y = "Exceedance probability",
           title = "Probability distribution fit") +
      theme_gray() +
      facet_wrap(~sea, scales = "free", nrow = 4) +
      theme(legend.position  = "bottom",
            strip.background = element_rect(fill = "grey5"),
            strip.text       = element_text(colour = "grey95"))
  }

  if (method == "acs") {

    a <- aTS$afits

    ACS <- lapply(a, function(x) {
      eACS <- attr(x, "eACS")
      lag  <- 0:(length(eACS) - 1)
      id   <- attr(x, "ID")
      ACS  <- do.call(acs, c(list(id = id, t = lag), x))
      data.frame(lag = lag, ACS = ACS, eACS = eACS)
    })
    names(ACS) <- paste(season, seq_along(ACS), sep = "_")

    sea <- NULL
    acs_dt <- rbindlist(ACS, idcol = "sea")
    acs_dt[, sea := factor(sea, levels = paste(season, seq_along(ACS), sep = "_"))]

    out <- ggplot(acs_dt) +
      geom_point(aes(x = as.factor(acs_dt$lag), y = acs_dt$eACS,
                     shape = factor(2)),
                 colour = "royalblue4", alpha = .5) +
      scale_shape_manual(values = 19, label = "Empirical", name = NULL) +
      geom_line(aes(x = acs_dt$lag + 1, y = acs_dt$ACS, linetype = factor(1)),
                colour = "red4", lwd = .5, alpha = .75) +
      scale_linetype_manual(values = 1, label = "Fitted", name = NULL) +
      labs(x = bquote(lag ~ tau), y = "Autocorrelation",
           title = "Autocorrelation structure fit") +
      theme_grey() +
      facet_wrap(~sea, scales = "free", nrow = 4) +
      theme(legend.position  = "bottom",
            strip.background = element_rect(fill = "grey5"),
            strip.text       = element_text(colour = "grey95"))
  }

  out
}


#' @rdname analyzeTS
#' @export
simulateTS <- function(aTS, from = NULL, to = NULL) {

  dist   <- attr(aTS, "dist")
  acsID  <- attr(aTS, "acsID")
  season <- attr(aTS, "season")
  date   <- data.table(attr(aTS, "date"))

  x <- aTS$data
  f <- aTS$dfits
  a <- aTS$afits

  distbounds <- if (dist == "beta") c(0, 1) else c(-Inf, Inf)

  ## compute p0 directly from stratified data — avoids running reportTS
  p0_vec <- 1 - round(
    sapply(x[[2]], nrow) /
    sapply(x[[1]], nrow), 2)

  ACS <- vector("list", length(x[[1]]))
  for (i in seq_along(x[[1]])) {
    p  <- actpnts(margdist   = attr(f[[i]], "dist"),
                  margarg    = f[[i]],
                  p0         = p0_vec[i],
                  distbounds = distbounds)
    fp <- fitactf(p)

    lag  <- 0:(length(attr(a[[i]], "eACS")) - 1)
    id   <- attr(a[[i]], "ID")
    as_i <- do.call(acs, c(list(id = id, t = lag), a[[i]]))
    ACS[[i]] <- actf(as_i, fp$actfcoef[1], fp$actfcoef[2])
  }
  names(ACS) <- names(a)

  p0 <- uval <- gauss <- value <- . <- season_id <- rn <- NULL

  if (is.null(from)) from <- date[1, date]
  if (is.null(to))   to   <- date[.N, date]

  by <- difftime(date[2, date], date[1, date])

  gausian <- seasonalAR(x = seq(from = from, to = to, by = by),
                        ACS = ACS)

  setkey(gausian, season)

  para <- as.data.table(
    t(sapply(f, function(x) as.matrix(do.call(cbind, x))))
  )
  names(para) <- names(f[[1]])
  para[, season := as.numeric(gsub("data_nz_", "", rownames(para)))]
  para[, p0     := p0_vec]
  setkey(para, season)

  aux <- merge(gausian, para, all.x = TRUE)
  aux <- aux[order(date)]
  aux[, uval := (pnorm(q = gauss) - p0) / (1 - p0)]
  aux[uval < 0, uval := 0]

  d <- getDistArg(dist)

  for (i in para[, season]) {
    trans.para <- para[season == i, !c("p0", "season")]
    aux[season == i,
        value := do.call(paste0("q", dist),
                         args = c(list(p = uval), as.list(trans.para)))]
  }

  aux[, .(date, value)]
}

# --- plot.fitDist -------------------------------------------------------------

#' Plot method for \code{fitDist} objects
#'
#' Displays the empirical CDF against the fitted theoretical CDF on a
#' log-exceedance-probability scale.
#'
#' @param x a \code{fitDist} object returned by \code{\link{fitDist}}
#' @param ... currently unused
#'
#' @return a \code{ggplot} object (invisibly returned; also printed)
#'
#' @seealso \code{\link{fitDist}}
#'
#' @export
#' @import ggplot2
#' @method plot fitDist
#'
#' @examples
#'
#' x <- fitDist(rnorm(1000), "norm", 30, "N1", FALSE)
#' plot(x)
#'
plot.fitDist <- function(x, ...) {

  edf  <- attr(x, "edf")
  dist <- attr(x, "dist")
  nfo  <- attr(x, "nfo")

  mm <- do.call(paste0("p", dist),
                c(list(q = range(edf$value)), x))
  p  <- exp(seq(log(mm[1]), log(mm[2]), length.out = 10000))

  cdf <- data.frame(p     = p,
                    value = do.call(paste0("q", dist), c(list(p = p), x)))

  ggplot() +
    geom_line(data = cdf,
              aes(x = cdf$value, y = log(1 - cdf$p)),
              colour = "grey25", lwd = 1, alpha = .75) +
    geom_point(data = edf,
               aes(x = edf$value, y = log(1 - edf$p)),
               colour = "red4", alpha = .5) +
    labs(x     = "Nonzero values",
         y     = "Exceedance probability",
         title = paste("Fitting norm (error) value =", round(nfo$objective, 5))) +
    scale_y_continuous(
      breaks = seq(-10, 0, length.out = 5),
      labels = format(exp(seq(log(.0001), log(1), length.out = 5)),
                      scientific = TRUE)) +
    theme_grey()
}


# --- plot.fitACS --------------------------------------------------------------

#' Plot method for \code{fitACS} objects
#'
#' Displays the empirical ACF alongside the fitted theoretical autocorrelation
#' structure.
#'
#' @param x a \code{fitACS} object returned by \code{\link{fitACS}}
#' @param ... currently unused
#'
#' @return a \code{ggplot} object (invisibly returned; also printed)
#'
#' @seealso \code{\link{fitACS}}
#'
#' @export
#' @import ggplot2
#' @method plot fitACS
#'
#' @examples
#'
#' x <- arima.sim(model = list(ar = 0.8), n = 1000)
#' acsfit <- fitACS(acf(x, plot = FALSE)$acf, "weibull", c(1, 1))
#' plot(acsfit)
#'
plot.fitACS <- function(x, ...) {

  eACS <- attr(x, "eACS")
  lag  <- 0:(length(eACS) - 1)
  id   <- attr(x, "ID")

  ACS <- do.call(acs, c(list(id = id, t = lag), x))

  df <- data.frame(lag  = lag,
                   ACS  = ACS,
                   eACS = eACS)

  ggplot(df) +
    geom_line(aes(x = lag, y = ACS),
              colour = "grey25", lwd = .75, alpha = .75) +
    geom_point(aes(x = lag, y = eACS),
               colour = "red4", alpha = .5) +
    labs(x     = bquote(lag ~ tau),
         y     = "Autocorrelation",
         title = "") +
    theme_grey()
}


# --- plot.acti ----------------------------------------------------------------

#' Plot method for \code{acti} objects
#'
#' Visualises the autocorrelation transformation function (ACTF) fitted by
#' \code{\link{fitactf}}.
#'
#' @param x an \code{acti} object returned by \code{\link{fitactf}}
#' @param ... optional arguments; \code{main} sets the plot title
#'
#' @return a \code{ggplot} object (invisibly returned; also printed)
#'
#' @seealso \code{\link{fitactf}}, \code{\link{actpnts}}
#'
#' @export
#' @import ggplot2 graphics utils methods
#' @method plot acti
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' p   <- actpnts(margdist = "paretoII",
#'                margarg  = list(scale = 1, shape = .3),
#'                p0 = 0)
#' fit <- fitactf(p)
#'
#' plot(fit)
#' plot(fit, main = "Pareto type II\nautocorrelation transformation")
#'
plot.acti <- function(x, ...) {

  args <- list(...)
  main <- if (!is.null(args[["main"]])) args[["main"]] else ""

  temp <- seq(0, 1, .001)
  dta  <- data.frame(x = temp,
                     y = actf(temp, x$actfcoef[1], x$actfcoef[2]))

  ggplot() +
    geom_line(aes(x = dta$x, y = dta$y),
              colour = "steelblue4", lwd = 1.5) +
    geom_point(aes(x = x$actfpoints$rhox, y = x$actfpoints$rhoz),
               colour = "grey35", size = 3.5) +
    geom_abline(lty = 5) +
    scale_x_continuous(limits = c(0, 1), expand = c(0.01, 0),
                       breaks = seq(0, 1, .2)) +
    scale_y_continuous(limits = c(0, 1), expand = c(0.01, 0),
                       breaks = seq(0, 1, .2)) +
    labs(x     = bquote(Autocorrelation ~ rho[x]),
         y     = bquote(Gaussian ~ rho[z]),
         title = main) +
    theme_gray() +
    theme(legend.position  = "bottom",
          strip.background = element_rect(fill = "grey5"),
          strip.text       = element_text(colour = "grey95"),
          axis.text        = element_text(size = 10),
          axis.title       = element_text(size = 15, face = "bold"))
}


# --- plot.cosmosts ------------------------------------------------------------

#' Plot method for \code{cosmosts} objects
#'
#' Visualises time series generated by \code{\link{generateTS}} as bar charts,
#' one panel per series.
#'
#' @param x a \code{cosmosts} object returned by \code{\link{generateTS}}
#' @param ... currently unused
#'
#' @return a \code{ggplot} object (invisibly returned; also printed)
#'
#' @seealso \code{\link{generateTS}}, \code{\link{regenerateTS}}
#'
#' @export
#' @import ggplot2 data.table
#' @method plot cosmosts
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ts <- generateTS(margdist = "ggamma",
#'                  margarg  = list(scale = 1, shape1 = .8, shape2 = .8),
#'                  acsvalue = acs(id = "paretoII", t = 0:30,
#'                                 scale = 1, shape = .75),
#'                  n = 1000, p = 30, TSn = 2)
#' plot(ts)
#'
plot.cosmosts <- function(x, ...) {

  dta <- data.table(n = seq_along(x[[1]]),
                    do.call(cbind, x))

  names(dta)[2:(length(x) + 1)] <- paste("timeseries", seq_along(x), sep = "_")

  m.dta <- melt(data = as.data.table(dta), id.vars = "n")

  ggplot(data    = m.dta,
         mapping = aes(x = m.dta$n, y = m.dta$value)) +
    geom_col() +
    labs(x = "", y = "value") +
    facet_wrap(~variable, ncol = 1) +
    theme_gray() +
    theme(legend.position  = "bottom",
          strip.background = element_rect(fill = "grey5"),
          strip.text       = element_text(colour = "grey95"))
}


# --- quickTSPlot --------------------------------------------------------------

#' Quick visualisation of basic time series properties
#'
#' Returns a composite figure showing the time series, empirical density
#' function, and empirical autocorrelation function.
#'
#' @param TS numeric vector (or \code{data.frame}/\code{data.table}) of time
#'   series values
#' @param ci numeric; confidence level for the zero-autocorrelation band
#'   (default \code{0.95})
#'
#' @return a \code{ggdraw} object (printed as a side effect)
#'
#' @seealso \code{\link{generateTS}}, \code{\link{plot.cosmosts}}
#'
#' @name quickTSPlot
#'
#' @export
#'
#' @examples
#' ggamma_sim <- rggamma(n = 1000, scale = 1, shape1 = 1, shape2 = .5)
#' quickTSPlot(ggamma_sim)
#'
quickTSPlot <- function(TS, ci = 0.95) {

  value <- Lag <- ACF <- NULL

  n   <- if (inherits(TS, c("data.frame", "data.table"))) nrow(TS) else length(TS)
  dta <- data.frame(time = seq_len(n), value = TS)

  p1 <- ggplot(dta, aes(x = time, y = value)) +
    geom_line() +
    theme_light() +
    labs(x = "Time", y = "Value", title = "Time series plot")

  p2 <- ggplot(dta[dta[, 2] != 0, ], aes(x = value, after_stat(density))) +
    geom_histogram() +
    theme_light() +
    labs(x = "Nonzero values", y = "Density", title = "Empirical density function")

  acf0 <- acf(TS, plot = FALSE)
  acf1 <- data.frame(Lag = acf0$lag, ACF = acf0$acf)
  clim <- qnorm((1 + ci) / 2) / sqrt(acf0$n.used)

  p3 <- ggplot(data = acf1, aes(x = Lag, y = ACF)) +
    geom_hline(aes(yintercept =  clim), linetype = 2) +
    geom_hline(aes(yintercept = -clim), linetype = 2) +
    geom_hline(aes(yintercept = 0)) +
    geom_segment(mapping = aes(xend = Lag, yend = 0)) +
    labs(title = "Autocorrelation function") +
    theme_light()

  p1 / (p2 | p3)

}


# --- plot.checkTS -------------------------------------------------------------

#' Plot method for \code{checkTS} objects
#'
#' Displays boxplots of simulated statistics against theoretical expected values
#' for each statistic tracked by \code{\link{checkTS}}.
#'
#' @param x a \code{checkTS} object returned by \code{\link{checkTS}}
#' @param ... currently unused
#'
#' @return a \code{ggplot} object (invisibly returned; also printed)
#'
#' @seealso \code{\link{checkTS}}
#'
#' @export
#' @import ggplot2 data.table
#' @method plot checkTS
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' x <- generateTS(margdist = "burrXII",
#'                 margarg = list(scale = 1,
#'                                shape1 = .75,
#'                                shape2 = .15),
#'                 acsvalue = acs(id = "weibull",
#'                                t = 0:30,
#'                                scale = 10,
#'                                shape = .75),
#'                 n = 1000, p = 30, p0 = .25, TSn = 100)
#'
#' chck <- checkTS(x)
#' plot(chck)
#'
plot.checkTS <- function(x, ...) {

  att      <- attributes(x)
  margdist <- att$margdist
  margarg  <- att$margarg
  p0       <- att$p0

  variable <- value <- NULL
  dta   <- melt(as.data.table(x, keep.rownames = TRUE), id.vars = "rn")
  dta.e <- dta[dta$rn == "expected", ]
  dta.s <- dta[dta$rn != "expected", ]

  ggplot() +
    geom_boxplot(data = dta.s,
                 aes(x = variable, y = value, group = variable)) +
    geom_point(data = dta.e,
               aes(x = variable, y = value, group = variable),
               size = 2, colour = "red1") +
    facet_wrap("variable", scales = "free", nrow = 1) +
    labs(x        = "",
         y        = "",
         title    = paste("Marginal =", margdist),
         subtitle = paste(
           paste(names(margarg), "=", margarg, collapse = "; "),
           paste("\np0 =", p0),
           collapse = " ")) +
    theme_gray() +
    theme(legend.position  = "bottom",
          strip.background = element_rect(fill = "grey5"),
          strip.text       = element_text(colour = "grey95"),
          axis.title.x     = element_blank(),
          axis.text.x      = element_blank(),
          axis.ticks.x     = element_blank())
}

#' Hourly station precipitation data
#'
#' Station details
#' * Name: Philadelphia International Airport
#' * Network ID: COOP:366889
#' * Latitude/Longitude:	39.87327°, -75.22678°
#' * Elevation: 3m
#'
#' more details can be found \href{https://www.ncdc.noaa.gov/cdo-web/datasets/PRECIP_HLY/stations/COOP:366889/detail}{here}.
#'
#' @format A data.table with 79633 rows and 2 variables:
#' \describe{
#'   \item{date}{POSIXct format date/time}
#'   \item{value}{precipitation totals}
#' }
#' @source The National Oceanic and Atmospheric Administration (NOAA)
"precip"

#' Population statistics
#'
#' Computes theoretical descriptive statistics of a distribution.
#'
#' @inheritParams moments
#' @param stat character; statistic to compute — one of \code{"mean"},
#'   \code{"sd"}, \code{"var"}, \code{"cvar"}, \code{"skew"}, \code{"kurt"}
#'
#' @return scalar numeric
#'
#' @seealso \code{\link{moments}}, \code{\link{checkTS}}
#'
#' @name PopulationStat
#'
#' @export
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' populationstat("mean", "norm", list(mean = 2, sd = 1))
#' populationstat("sd",   "norm", list(mean = 2, sd = 1))
#' populationstat("var",  "norm", list(mean = 2, sd = 1))
#' populationstat("cvar", "norm", list(mean = 2, sd = 1))
#' populationstat("skew", "norm", list(mean = 2, sd = 1))
#' populationstat("kurt", "norm", list(mean = 2, sd = 1))
#'
populationstat <- function(stat = "mean", dist, distarg,
                           p0 = 0, distbounds = c(-Inf, Inf)) {
  do.call(paste0("pop", stat),
          args = list(dist = dist, distarg = distarg,
                      distbounds = distbounds, p0 = p0))
}

#' @rdname PopulationStat
#' @keywords internal
popmean <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, central = TRUE, raw = FALSE, coef = FALSE)
  unname(x[["mu"]][1])
}

#' @rdname PopulationStat
#' @keywords internal
popsd <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, central = TRUE, raw = FALSE, coef = FALSE)
  unname(sqrt(x[["mu"]][2]))
}

#' @rdname PopulationStat
#' @keywords internal
popvar <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, central = TRUE, raw = FALSE, coef = FALSE)
  unname(x[["mu"]][2])
}

#' @rdname PopulationStat
#' @keywords internal
popcvar <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, coef = TRUE, raw = FALSE, central = FALSE)
  unname(x[["coefficients"]][1])
}

#' @rdname PopulationStat
#' @keywords internal
popskew <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, coef = TRUE, raw = FALSE, central = FALSE)
  unname(x[["coefficients"]][2])
}

#' @rdname PopulationStat
#' @keywords internal
popkurt <- function(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf)) {
  x <- moments(dist = dist, distarg = distarg, distbounds = distbounds,
               p0 = p0, coef = TRUE, raw = FALSE, central = FALSE)
  unname(x[["coefficients"]][3])
}

#' VAR model parameters to simulate correlated parent Gaussian random vectors and fields
#'
#' Compute VAR model parameters to simulate parent Gaussian random vectors with specified spatiotemporal correlation structure using the method described by Biller and Nelson (2003).
#'
#' @param spacepoints it can be a numeric integer, which is interpreted as the side length m of the square field (m x m), or a matrix (d x 2) of coordinates (e.g. longitude and latitude) of d spatial locations (e.g. d gauge stations)
#' @param p order of VAR(p) model
#' @param margdist target marginal distribution of the field
#' @param margarg  list of marginal distribution arguments. Please consult the documentation of the selected marginal distribution indicated in the argument \code{margdist} for the list of required parameters
#' @param p0 probability zero
#' @param distbounds distribution bounds (default set to \code{c(-Inf, Inf)})
#' @param stcsid spatiotemporal correlation structure ID
#' @param stcsarg  list of spatiotemporal correlation structure arguments. Please consult the documentation of the selected spatiotemporal correlation structure indicated in the argument \code{stcsid} for the list of required parameters
#' @param scalefactor factor specifying the distance between the centers of two pixels (default set to 1)
#' @param anisotropyid spatial anisotropy ID (\code{affine} by default, \code{swirl} or \code{wave})
#' @param anisotropyarg list of arguments characterizing the spatial anisotropy according to the syntax of the function \code{\link{anisotropyT}}. Isotropic fields by default
#' @param advectionid advection field ID (\code{uniform} by default, \code{rotation}, \code{spiral}, \code{spiralCE}, \code{radial}, or \code{hyperbolic})
#' @param advectionarg list of arguments characterizing the advection field according to the syntax of \code{advectionF}. No advection by default
#' @param dsid dependence structure ID (\code{gauss} by default, \code{student}, \code{bardossy}, and \code{bardossyF})
#' @param dsarg argument characterizing the dependence structure: \code{NULL} for \code{gauss} by default, number of degrees of freedom for \code{student} or parameter \code{m} in \code{(-Inf, Inf)} for \code{bardossy} (see Note section for more details)

#' @name fitVAR
#'
#' @importFrom Matrix nearPD
#' @importFrom pracma distmat
#'
#' @export
#'
#' @details
#' The fitting algorithm has \eqn{O(m*m)^3} complexity for a \eqn{(m*m)} field
#' or equivalently \eqn{O(d^3)} complexity for a \eqn{d}-dimensional vector.
#' Very large values of \eqn{(m*m)} (or \eqn{d}) and high order AR correlation
#' structures can be unpractical on standard machines.
#'
#' Here, we give indicative CPU times for some settings, referring to a
#' Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
#' 4-core, 8 logical processors, and 32GB RAM. \cr:
#' CPU time:\cr
#' d = 100 or m = 10, p = 1: ~  0.4s \cr
#' d = 900 or m = 30, p = 1: ~  6.0s \cr
#' d = 900 or m = 30, p = 5: ~ 47.0s \cr
#' d = 2500 or m = 50, p = 1: ~100.0s
#'
#' @note
#' While all the advection types can be applied to isotropic random fields,
#' anisotropic random fields require more care. We suggest combining \code{affine}
#' anysotropy with \code{uniform} advection, and \code{swirl} anisotropy
#' with \code{rotation} or \code{spiral} advection with the same rotation center..\cr
#' Concerning the Bardossy model, the increase of the parameter \code{m}
#' leads to a more and more symmetrical copula, and if \code{m} tends to \code{Inf},
#' then the copula converges to the Gaussian copula. The \code{bardossy} model is
#' characterized by lower tail dependence weaker than the upper tail dependence, while the
#' flipped Bárdossy dependence structure, denoted as  \code{bardossyF}, has lower tail
#' dependence stronger than the upper tail dependence.
#' See Bárdossy (2006) for more details about the properties and
#' parametrization of the multivariate Bardossy distribution
#'
#' @references Bárdossy, A. (2006), Copula-based geostatistical models for groundwater
#' quality parameters, Water Resour. Res., 42, W11416, \doi{10.1029/2005WR004754}
#' @references Biller, B., Nelson, B.L. (2003). Modeling and generating multivariate
#' time-series input processes using a vector autoregressive technique.
#' ACM Trans. Model. Comput. Simul. 13(3), 211-237, \doi{10.1145/937332.937333}
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic modelling of
#' hydroclimatic processes: Preserving marginal distributions, correlation structures,
#' and intermittency. Advances in Water Resources, 115, 234-252,
#' \doi{10.1016/j.advwatres.2018.02.013}
#' @references Papalexiou, S.M.,  Serinaldi, F. (2020). Random Fields Simplified:
#' Preserving Marginal Distributions, Correlations, and Intermittency,
#' With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
#' e2019WR026331, \doi{10.1029/2019WR026331}
#' @references Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
#' Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
#' Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
#'
#' @examples
#' ## for multivariate simulation
#' coord <- cbind(runif(4)*30, runif(4)*30)
#'
#' fit <- fitVAR(
#'   spacepoints = coord,
#'   p = 1,
#'   margdist ='burrXII',
#'   margarg = list(scale = 3,
#'                  shape1 = .9,
#'                  shape2 = .2),
#'   p0 = 0.8,
#'   stcsid = "clayton",
#'   stcsarg = list(scfid = "weibull",
#'                  tcfid = "weibull",
#'                  copulaarg = 2,
#'                  scfarg = list(scale = 20,
#'                                shape = 0.7),
#'                  tcfarg = list(scale = 1.1,
#'                                shape = 0.8))
#' )
#'
#' dim(fit$alpha)
#' dim(fit$res.cov)
#'
#' fit$m
#' fit$margarg
#' fit$margdist
#'
#' ## for random fields simulation
#' fit <- fitVAR(
#'   spacepoints = 10,
#'   p = 1,
#'   margdist ='burrXII',
#'   margarg = list(scale = 3, shape1 = .9, shape2 = .2),
#'   p0 = 0.8,
#'   stcsid = "clayton",
#'   stcsarg = list(scfid = "weibull", tcfid = "weibull",
#'                  copulaarg = 2,
#'                  scfarg = list(scale = 20, shape = 0.7),
#'                  tcfarg = list(scale = 1.1, shape = 0.8))
#' )
#'
#' dim(fit$alpha)
#' dim(fit$res.cov)
#'
#' fit$m
#' fit$margarg
#' fit$margdist
#'
fitVAR <- function(spacepoints, p, margdist, margarg, p0, distbounds = c(-Inf, Inf),
                   stcsid, stcsarg, scalefactor = 1, anisotropyid = "affine",
                   anisotropyarg = list(phi1 = 1, phi2 = 1, phi12 = 0, theta = 0),
                   advectionid = "uniform", advectionarg = list(u = 0, v = 0),
                   dsid = "gauss", dsarg = NULL) {

  if (is.numeric(spacepoints) && length(spacepoints) == 1) {
    mm <- spacepoints^2
    d <- seq(0, spacepoints - 1, length = spacepoints) * scalefactor
    dd0 <- expand.grid(d, d)
  } else {
    if (is.matrix(spacepoints) || is.array(spacepoints)) {
      mm <- nrow(spacepoints)
      dd0 <- spacepoints
    } else stop("spacepoints should be an integer or a matrix (d x 2)")
  }

  anisotropyarg$spacepoints <- dd0
  dd <- anisotropyT2(anisotropyid, anisotropyarg)
  dis <- dist(dd)

  advectionarg$spacepoints <- dd0
  uv <- advectionF2(advectionid, advectionarg)

  if (anisotropyid == "swirl") {
    advectionarg$spacepoints <- cbind(anisotropyarg$x0, anisotropyarg$y0)
    uv1 <- advectionF2(advectionid, advectionarg)
    x0 =  anisotropyarg$x0
    y0 =  anisotropyarg$y0
  }

  if (dsid %in% c("gauss", "student")) {
    pnts <- actpnts(margdist = margdist, margarg = margarg, p0 = p0, distbounds = distbounds)
  }
  if (dsid %in% c("bardossy", "bardossyF")) {
    pnts <- actpntsB6(margdist = margdist, margarg = margarg, m = dsarg, p0 = p0)
  }
  actfpara <- fitactf(pnts)

  sigma.i <- array(NA,c(mm, mm, p+1))
  for (i in 1:(p + 1)){
    stcsarg1 <- stcsarg
    stcsarg1$t <- i - 1
    anisotropyarg$spacepoints <- dd0 + uv * stcsarg1$t
    if (anisotropyid == "swirl") {
      anisotropyarg$x0 =  x0 + uv1[ ,1] * stcsarg1$t
      anisotropyarg$y0 =  y0 + uv1[ ,2] * stcsarg1$t
    }
    dd1 <- anisotropyT2(anisotropyid, anisotropyarg)
    stcsarg1$s <- distmat(dd,dd1)
    sigma.i[ , ,i] <- actf( stcs2(stcsid, stcsarg1), b = actfpara$actfcoef[1], c = actfpara$actfcoef[2])
  }

  sigma.Z <- matrix(NA,p * mm,p * mm)
  sigma <- matrix(NA, mm, p * mm)
  for (i in 1:p) {
    for (j in 1:p) {
      if (j >= i) sigma.Z[seq(1 + mm * (i - 1), mm * i), seq(1 + mm * (j - 1), mm * j)] =   sigma.i[,,j-i+1]
      else        sigma.Z[seq(1 + mm * (i - 1), mm * i), seq(1 + mm * (j - 1), mm * j)] = t(sigma.i[,,i-j+1])
    }
    sigma[ , seq(1 + mm * (i - 1), mm * i)] = sigma.i[ , , i+1]
  }

  alpha <- t(solve(sigma.Z, t(sigma)))

  sums <- matrix(0, mm, mm)
  for (i in 1:p) {
    sums <- sums  +  ( alpha[ , seq(1 + mm * (i - 1), mm * i)] %*% t(sigma.i[,,i+1]) )
  }
  sigma.u <- nearPD( sigma.i[ , , 1]  - sums )$mat

  d <- seq(0, max(dis), length = mm)
  u <- 0:200
  ud <- expand.grid(u, d)
  stcsarg$t <- ud[ ,1]
  stcsarg$s <- ud[ ,2]
  fout <- (matrix(stcs2(stcsid, stcsarg) , nrow = length(u)))

  list(alpha = alpha, res.cov = sigma.u, spacepoints = spacepoints, mm = mm, p = p, margdist = margdist, margarg = margarg, p0 = p0,
              stcs = fout, s.lags = d, t.lags=u, grid.lags = dis, distbounds = distbounds, dsid = dsid, dsarg = dsarg)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Fast actpnts via nested C++ 1D integration
#'
#' Internal C++ implementation. Do not call directly.
#'
#' @param margdist character distribution name
#' @param margarg  named numeric vector of distribution parameters
#' @param p0       probability zero
#' @param lower    integration lower bound
#' @param upper    integration upper bound
#' @param mu1      first raw moment
#' @param mu2      second central moment
#' @param rhoz_vals numeric vector of Gaussian correlation values
#' @param max_eval  max evaluations per 1D integral (default 1e4)
#' @param abs_tol   absolute tolerance (default 1e-5)
#' @param rel_tol   relative tolerance (default 1e-5)
#'
#' @return data.frame with columns rhoz and rhox
#' @keywords internal
actpnts_cpp <- function(margdist, margarg, p0, lower, upper, mu1, mu2, rhoz_vals, max_eval = 10000L, abs_tol = 1e-5, rel_tol = 1e-5) {
    .Call(`_CoSMoS_actpnts_cpp`, margdist, margarg, p0, lower, upper, mu1, mu2, rhoz_vals, max_eval, abs_tol, rel_tol)
}

#' Generate time series
#'
#' Generates time series with given properties. Provide (1) the target marginal
#' distribution and its parameters, (2) the target autocorrelation structure or
#' individual autocorrelation values up to a desired lag, and (3) the probability
#' zero if you wish to simulate an intermittent process.
#'
#' @details
#'
#' A step-by-step guide:
#' * First define the target marginal (\code{margdist}), that is, the probability
#' distribution of the generated data. For example set \code{margdist = 'ggamma'}
#' for the Generalised Gamma distribution, \code{margdist = 'burrXII'} for Burr
#' type XII etc. For a full list of supported distributions see the help
#' \href{https://CRAN.R-project.org/package=CoSMoS/vignettes/vignette.html}{vignette}.
#' In general, the package supports all built-in distribution functions of R and
#' of other packages.
#'
#' * Define the parameters (\code{margarg}) of the selected distribution.
#' For example the Generalised Gamma has one scale and two shape parameters,
#' e.g. \code{margarg = list(scale = 2, shape1 = 0.9, shape2 = 0.8)}.
#' See the help vignette for details on each distribution's parameters.
#'
#' * If you wish your time series to be intermittent (e.g. precipitation), define
#' the probability zero. For example \code{p0 = 0.9} produces 90\% zero values.
#'
#' * Define your linear autocorrelations.
#'   + Supply specific lag autocorrelations starting from lag 0 up to a desired
#'     lag, e.g. \code{acsvalue = c(1, 0.9, 0.8, 0.7)}; this preserves lag-1,
#'     lag-2 and lag-3 autocorrelations equal to 0.9, 0.8 and 0.7.
#'
#'   + Alternatively, use a parametric autocorrelation structure (see section 3.2
#'     in Papalexiou (2018)). Supported structures: \code{weibull}, \code{paretoII},
#'     \code{fgn} and \code{burrXII}. See also \code{\link[CoSMoS]{acs}}.
#'
#' * Define the AR model order \code{p}. For example if you aim to preserve the
#' first 10 lag autocorrelations then set \code{p = 10}. Set \code{p = NULL} and
#' the model will choose \code{p} to preserve the whole autocorrelation structure.
#'
#' * Set the time series length, e.g. \code{n = 1000}, and the number of time
#' series to generate, e.g. \code{TSn = 10}.
#'
#' @inheritParams actpnts
#' @inheritParams ARp
#' @param TSn number of time series to generate
#'
#' @return An object of class \code{'cosmosts'}: a list of \code{TSn} numeric
#'   vectors, each of length \code{n}, with per-series attributes recording the
#'   fitted model parameters.
#'
#' @seealso \code{\link{regenerateTS}}, \code{\link{ARp}}, \code{\link{actpnts}}
#'
#' @import data.table ggplot2
#' @export
#'
#' @references Papalexiou, S.M. (2018). Unified theory for stochastic modelling of
#' hydroclimatic processes: Preserving marginal distributions, correlation structures,
#' and intermittency. Advances in Water Resources, 115, 234-252,
#' \doi{10.1016/j.advwatres.2018.02.013}
#'
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## Case 1:
#' ## Generate 3 time series of length 1000 following the Generalised Gamma
#' ## distribution with scale = 1, shape1 = 0.8, shape2 = 0.8 and ParetoII
#' ## autocorrelation structure with scale = 1 and shape = 0.75.
#' x <- generateTS(margdist = "ggamma",
#'                 margarg = list(scale = 1,
#'                                shape1 = .8,
#'                                shape2 = .8),
#'                 acsvalue = acs(id = "paretoII",
#'                                t = 0:30,
#'                                scale = 1,
#'                                shape = .75),
#'                 n = 1000,
#'                 p = 30,
#'                 TSn = 3)
#'
#' ## see the results
#' plot(x)
#'
#' \donttest{
#'
#' ## Case 2:
#' ## Same as Case 1 but intermittent with probability zero equal to 90%.
#' y <- generateTS(margdist = "ggamma",
#'                 margarg = list(scale = 1,
#'                                shape1 = .8,
#'                                shape2 = .8),
#'                 acsvalue = acs(id = "paretoII",
#'                                t = 0:30,
#'                                scale = 1,
#'                                shape = .75),
#'                 p0 = .9,
#'                 n = 1000,
#'                 p = 30,
#'                 TSn = 3)
#'
#' ## see the results
#' plot(y)
#'
#' ## Case 3:
#' ## Generate a time series of length 1000 following the Beta distribution
#' ## (e.g. relative humidity in [0, 1]) with shape1 = 0.6, shape2 = 0.8
#' ## and ParetoII autocorrelation structure.
#' z <- generateTS(margdist = "beta",
#'                 margarg = list(shape1 = .6,
#'                                shape2 = .8),
#'                 distbounds = c(0, 1),
#'                 acsvalue = acs(id = "paretoII",
#'                                t = 0:30,
#'                                scale = 1,
#'                                shape = .75),
#'                 n = 1000,
#'                 p = 20)
#'
#' ## see the results
#' plot(z)
#'
#' ## Case 4:
#' ## Same as Case 3 but providing specific autocorrelation values for the
#' ## first three lags (lag 1 to 3 equal to 0.9, 0.8, 0.7).
#' z <- generateTS(margdist = "beta",
#'                 margarg = list(shape1 = .6,
#'                                shape2 = .8),
#'                 distbounds = c(0, 1),
#'                 acsvalue = c(1, .9, .8, .7),
#'                 n = 1000,
#'                 p = NULL)
#'
#' ## see the results
#' plot(z)
#'
#' }
#'
generateTS <- function(n, margdist, margarg, p = NULL, p0 = 0, TSn = 1,
                       distbounds = c(-Inf, Inf), acsvalue = NULL) {

  pnts <- actpnts(margdist = margdist,
                  margarg = margarg,
                  p0 = p0,
                  distbounds = distbounds)

  par <- fitactf(pnts)

  out <- lapply(1:TSn, function(i) {
    ARp(margdist = margdist,
        margarg = margarg,
        n = n,
        p = p,
        p0 = p0,
        actfpara = par,
        acsvalue = acsvalue)
  })

  structure(.Data = out, class = "cosmosts")
}

#' Bulk time series generation
#'
#' Generates additional time series using parameters already fitted by
#' \code{\link{generateTS}}, avoiding recomputation of the ACTF.
#'
#' @details
#' After calling \code{\link{generateTS}}, use \code{regenerateTS} to generate
#' more time series with the same fitted parameters. This is faster than
#' re-running \code{\link{generateTS}} because the ACTF fitting step is skipped.
#'
#' @param ts a \code{cosmosts} object returned by \code{\link{generateTS}}
#' @param TSn number of time series to generate
#'
#' @return An object of class \code{'cosmosts'}: a list of \code{TSn} numeric
#'   vectors of the same length as those in \code{ts}.
#'
#' @seealso \code{\link{generateTS}}, \code{\link{ARp}}
#'
#' @export
#'
#' @examples
#'
#' library(CoSMoS)
#'
#' ## Fit once
#' x <- generateTS(margdist = "burrXII",
#'                 margarg = list(scale = 1,
#'                                shape1 = .75,
#'                                shape2 = .25),
#'                 acsvalue = acs(id = "weibull",
#'                                t = 0:30,
#'                                scale = 10,
#'                                shape = .75),
#'                 n = 1000, p = 30, p0 = .5, TSn = 3)
#'
#' ## Generate more realisations with the same parameters
#' r <- regenerateTS(x)
#'
#' plot(r)
#'
regenerateTS <- function(ts, TSn = 1) {

  X <- ts

  margdist <- attr(X[[1]], "margdist")
  margarg  <- attr(X[[1]], "margarg")

  n   <- length(X[[1]])
  p   <- length(attr(X[[1]], "ar_coef"))
  a   <- attr(X[[1]], "ar_coef")
  p0  <- attr(X[[1]], "p0")
  esd <- attr(X[[1]], "noise_sd")

  out <- lapply(1:TSn, function(i) {

    gn  <- c(rep(0, (2 * p)), rnorm(n, mean = 0, sd = esd))
    val <- c(AR1((2 * p), alpha = a[1]), rep(0, n))

    for (j in (p + 1):(n + (2 * p))) {
      val[j] <- sum(rev(a) * val[(j - p):(j - 1)]) + gn[j]
    }

    uval <- (pnorm(val[-1:-(2 * p)]) - p0) / (1 - p0)
    uval[uval < 0] <- 0

    do.call(paste0("q", margdist), args = c(list(p = uval), margarg))
  })

  structure(.Data = out, class = "cosmosts")
}

#' Empirical cumulative distribution function
#'
#' Calculates the ECDF using the Weibull plotting position formula.
#'
#' @param x Numeric vector of values.
#'
#' @return A \code{data.table} with columns \code{p} (plotting positions)
#'   and \code{value} (sorted unique values).
#'
#' @keywords internal
#'
ECDF <- function(x) {

  st <- sort(x)

  aux <- data.table(
    p     = rank(x = st, ties.method = "first") / (length(st) + 1),
    value = st,
    key   = "value"
  )

  J <- value <- NULL  # data.table global variable declaration

  aux[J(unique(value)), mult = "last"]
}

---
title: "CoSMoS R | Complete Stochastic Modelling Solution"
author: "Simon Michael Papalexiou, Francesco Serinaldi, Filip Strnad, Yannis Markonis, Kevin Shook"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_document:
    self_contained: true
    toc: true
    toc_float: true
    number_sections: true
    theme: flatly
    highlight: tango
    fig_caption: true
vignette: >
  %\VignetteIndexEntry{CoSMoS R | Complete Stochastic Modelling Solution}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: inline
---

***

```{r setup, include=FALSE}
knitr::opts_chunk$set(eval    = TRUE,
                      echo    = TRUE,
                      fig.width  = 7,
                      fig.retina = 1,
                      warning = FALSE,
                      message = FALSE)
library(CoSMoS)
library(plot3D)
library(patchwork)
ggplot2::theme_set(ggplot2::theme_light())
```
`CoSMoS` was conceived back in 2009 (see note in section 5) and was officially released on CRAN in April 2019. Since then `CoSMoS` has become one of the leading and most widely downloaded R packages for stochastic simulation of non-Gaussian time series. Designed with the end user in mind, **`CoSMoS` makes univariate, multivariate, or random field simulations easy**. You can reliably generate time series or fields of rainfall, streamflow, wind speed, relative humidity, or any other environmental variable in seconds. Just choose the probability distribution and correlation properties of the time series you want to generate, and it will do the rest.

<br>
~_Unified Theory of Stochastic Modelling_ | [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013)

_"Hydroclimatic processes come in all “shapes and sizes”. They are characterized by different spatiotemporal correlation structures and probability distributions that can be continuous, mixed-type, discrete or even binary. Simulating such processes by reproducing precisely their marginal distribution and linear correlation structure, including features like intermittency, can greatly improve hydrological analysis and design. ... Here, a single framework is proposed that unifies, extends, and improves a general-purpose modelling strategy, based on the assumption that any process can emerge by transforming a specific “parent”" Gaussian process."_

~_Precise Temporal Disaggregation_ | [Papalexiou et al. (2018)](https://doi.org/10.1029/2018WR022726)

_"Hydroclimatic variables such as precipitation and temperature are often measured or simulated by climate models at coarser spatiotemporal scales than those needed for operational purposes. … Here we introduce a novel disaggregation method, named Disaggregation Preserving Marginals and Correlations (DiPMaC), that is able to disaggregate a coarse-scale time series to any finer scale, while reproducing the probability distribution and the linear correlation structure of the fine-scale process. DiPMaC is also generalized for arbitrary nonstationary scenarios to reproduce time varying marginals."_

~_Random Fields Simplified_ | [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)

_"Nature manifests itself in space and time. The spatiotemporal complexity of processes such as precipitation, temperature, and wind, does not allow purely deterministic modeling. Spatiotemporal random fields have a long history in modeling such processes, and yet a single unified framework offering the flexibility to simulate processes that may differ profoundly does not exist. Here we introduce a blueprint to efficiently simulate spatiotemporal random fields that preserve any marginal distribution, any valid spatiotemporal correlation structure, and intermittency."_

~_Advancing Space-Time Simulation_ | [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466)


_"...we advance random field simulation by introducing the concepts of general velocity fields and general anisotropy transformations. To illustrate the potential of CoSMoS, we simulate random fields with complex patterns and motion mimicking rainfall storms moving across an area, spiraling fields resembling weather cyclones, fields converging to (or diverging from) a point, and colliding air masses."_

***
# Step by step guide to timeseries simulation {#section_1}

In most cases we can accurately simulate a process by reproducing its marginal distribution and its autocorrelation structure. The first is typically represented by a parametric probability distribution; the second either by a parametric correlation structure, or by specific correlation coefficient values. `CoSMoS` is the first software that enables the generation of time series while preserving any marginal distribution, correlation structure, and intermittency. Intermittency can be defined by the probability of zero values, and is essential in the simulation of precipitation at fine scales (e.g., daily, hourly) but also for any other intermittent process, such as flows of ephemeral streams, wind speed, etc.

Accurate generation of time series requires knowledge of the statistical properties of the process under study. Processes such as precipitation, streamflow, temperature, wind, or evaporation may have very different properties; that is, they can be described by different probability distributions, correlation structures, and may or may not be intermittent. 

There are four main steps required to generate a time series:
 1. Choosing a probability distribution
 2. Choosing a correlation structure, 
 3. Adding intermittency, and
 4. Validating the results.

The four steps are explained below.

## Choosing a Probability Distribution

The choice of a probability distribution to describe the variable you wish to simulate is crucial. The probability distribution expresses how frequently specific magnitudes of the variable occur in the data set. Air **temperature** is typically described by bell-shaped distributions like the [Normal](https://en.wikipedia.org/wiki/Normal_distribution). **Relative humidity** needs distributions defined in the interval (0,1) such as the [Beta](https://en.wikipedia.org/wiki/Beta_distribution) or the [Kumaraswamy](https://en.wikipedia.org/wiki/Kumaraswamy_distribution). **Streamflow** typically has heavy tails and power-type distributions might be needed to reproduce the behavior of the extremes. **Wind speed** typically can be modelled by positively skewed distributions such as the [Weibull](https://en.wikipedia.org/wiki/Weibull_distribution). **Precipitation** at fine scales has to be modelled by mixed-type (or else zero inflated) distributions, that is, having a probability value to describe the occurrence of zeros, and a continuous distribution, such the [Generalized Gamma](https://en.wikipedia.org/wiki/Generalized_gamma_distribution) or the [Burr type XII](https://en.wikipedia.org/wiki/Burr_distribution), to describe the frequency of nonzero values. 

`CoSMoS` supports **all the standard distribution** functions of `R` and from other packages (tens of distributions) but also comes with some built-in distributions such as Generalized Gamma, the [Pareto type II](https://en.wikipedia.org/wiki/Pareto_distribution), and the Burr type III and XII distributions. More details on the `CoSMoS` distributions and their parameters can be found in [Section 4](#section_4). 

Distributions may have differing numbers and types of parameters (e.g. location, scale, and shape parameters). For example, the Generalized Gamma distribution has one scale and two shape parameters.

`CoSMoS` has built-in functions to fit distributions to data and assess their goodness-of-fit based on graphs; but you can also use any other package for distribution fitting. 

## Choosing a Correlation Structure

The correlation structure expresses how much the random variables depend upon each other. Processes at fine temporal scales are typically more correlated than those at coarse scales. CoSMoS offers two options to introduce correlations: (1) by defining specific lag autocorrelations as a list starting from lag 0 up to a desired limit, or (2) by using a pre-defined parametric autocorrelation structure.

For example, you can generate a time series with lag-1 autocorrelation $\rho_1 = 0.8$ and Generalized Gamma marginal distribution by:

```{r, warning=FALSE, message=FALSE}
## (i) specifying the sample size
no <- 1000
## (ii) defining the type of marginal distribution and its parameters
marginaldist <- "ggamma"
param <- list(scale = 1, shape1 = .8, shape2 = .8)
## (iii) defining the desired autocorrelation
acf.my <- c(1, 0.8)
## (iv) simulating
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf.my)
## and (v) visually checking the generated time series
quickTSPlot(ggamma_sim[[1]])
```

You can specify also as many autocorrelation values as you wish. In this example, autocorrelation values are specified up to lag 4:

```{r, warning=FALSE, message=FALSE}
acf <- c(1, 0.6, 0.5, 0.4, 0.3) # up to lag-4
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
quickTSPlot(ggamma_sim[[1]])
```
A better approach is to use an autocorrelation structure expressed by a parametric function. `CoSMoS` has built in autocorrelation structures than can be used by the `acs()` function. You can use the following ACS structures:

*	Fractional Gaussian Noise; this is the well-known one-parameter fGn ACS.
*	Weibull; a flexible two parameter of exponential form.
*	Pareto II: a two parameter power-type ACS.
*	Burr XII: a three parameter power-type ACS. 

For most practical applications a two-parameter ACS is sufficient; we suggest either the Weibull or the Pareto II. 


The `acs()` function produces autocorrelation values based on any of the four structures and any desired lag. Thus, instead of setting each autocorrelation coefficient explicitly (which might be problematic from a technical viewpoint), we can generate series preserving any one of the four ACSs. For example, you can easily define an ACS and visualize its values up to any lag. Here, all four ACSs are defined and visualized up to lag 10. You can see how changing the ACS function changes how the correlation coefficients decay to zero.

```{r, warning=FALSE, message=FALSE}
## specify lag
lags <- 0:10

## get the ACS
f <- acs(id = "fgn", t = lags, H = .75)
b <- acs(id = "burrXII", t = lags, scale = 1, shape1 = .6, shape2 = .4)
w <- acs(id = "weibull", t = lags, scale = 2, shape = 0.8)
p <- acs(id = "paretoII", t = lags, scale = 3, shape = 0.3)

## visualize the ACS
dta <- data.table(lags, f, b, w, p)
m.dta <- melt(data = dta, id.vars = "lags")

ggplot(data = m.dta, mapping = aes(x = lags, y = value, group = variable, colour = variable)) +
  geom_point(size = 2.5) +
  geom_line(lwd = 1) +
  scale_color_manual(
    values = c("steelblue4", "red4", "green4", "darkorange"),
    labels = c("FGN", "Burr XII", "Weibull", "Pareto II"), name = ""
  ) +
  labs(x = bquote(lag ~ tau), y = "ACS") +
  scale_x_continuous(breaks = lags)
```

We can re-generate the previously generated series which used explicitly defined correlations up to lag 4, now with a two parameter Pareto II ACF up to lag 30, which improves the modelling parsimony. More details about the parametric autocorrelation structures can be found in section 3.2 in [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013).

```{r, warning=FALSE, message=FALSE}
acf <- acs(id = "paretoII", t = 0:30, scale = 1, shape = .75)
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
dta <- data.frame(time = 1:no, value = ggamma_sim[[1]])

quickTSPlot(dta$value)
```

Apart from the four autocorrelation functions provided by `acs()`, we can also create our own. Note that it is important to ensure that the function is positive definite, otherwise the autoregressive model cannot be fitted. This example shows the generation of a time series with the same Generalized Gamma marginal distribution as above, but with a user-defined exponential ACS function up to lag 30. Note that although the time series density is very similar to that in the previous plot, the texture of the time series is very different, due to the changed ACS function.

```{r, warning=FALSE, message=FALSE}
my_acf <- exp(seq(0, -2, -0.1))
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = my_acf)
quickTSPlot(ggamma_sim[[1]])
```

## Adding Intermittency (if required)

`CoSMoS` easily simulates intermittent processes such as precipitation. The only extra step needed is to define the probability of zero events. For example, say you wish to generate 5 mutually independent timeseries having the previously defined Generalized Gamma distribution with the Pareto II ACS and probability zero p0 = 0.9. The generated data thus will have 90% of zero values (i.e. dry days). 

```{r, fig.height = 7, fig.cap = "Five simulated intermittent Generalized Gamma time series (p0 = 0.9).", warning=FALSE, message=FALSE}
prob_zero <- .9
## the argument `TSn = 5` enables the simulation of 5 timeseries.
ggamma_sim <- generateTS(
  n = no, margdist = marginaldist, margarg = param, acsvalue = acf,
  p0 = prob_zero, TSn = 5
)
plot(x = ggamma_sim, main = "") 
```

## Validating the Results

You can readily check some basic statistics of the generated time series using the `checkTS()` function, which return the first three moments, probability zero and the first three autocorrelation coefficients.

```{r, warning=FALSE, message=FALSE}
checkTS(ggamma_sim)
```

***

# Multivariate and Random Fields Simulation  {#section_2}

The above methods can readily be extended to multidimensional cases, thus enabling the simulation of spatiotemporally correlated random vectors (i.e. correlated timeseries at multiple locations) and random fields (i.e. gridded data), as discussed in detail by [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331). This requires the definition of suitable spatiotemporal correlation structures (see [Porcu et al. (2020)](https://doi.org/10.1002/wics.1512) for a thorough review of this topic).

## Spatiotemporal Correlation Structures

CoSMoS allows the definition of spatiotemporal correlation functions using the function `stcs()`, which the spatiotemporal counterpart of the purely temporal `acs()`. Two classes of spatiotemporal correlation functions are provided:

*	Clayton [(Papalexiou and Serinaldi 2020)](https://doi.org/10.1029/2019WR026331)
*	Gneiting [(Gneiting 2002)](https://doi.org/10.1198/016214502760047113)  

The `stcs()` function can produce values of the linear spatiotemporal correlation for any desired time lag and spatial distance using these two correlation function classes, which comprise a variety of structures covering most cases of practical interest. This example shows the Clayton-Weibull spatiotemporal correlation structure:

```{r, fig.height = 7, fig.cap = "Clayton-Weibull spatiotemporal correlation structure.", warning=FALSE, message=FALSE}
d <- 51
st <- expand.grid(0:(d - 1), 0:(d - 1))

## get the STCS
wc <- stcfclayton(
  t = st[, 1], s = st[, 2], scfid = "weibull", tcfid = "weibull", copulaarg = 2,
  scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 5.1, shape = 0.8)
)

## visualize the STCS
wc.m <- matrix(data = wc, nrow = d)
j <- tail(which(wc.m[1, ] > 0.15), 1)
i <- tail(which(wc.m[, 1] > 0.15), 1)
wc.m <- wc.m[1:i, 1:j]

persp3D(
  z = wc.m, x = 1:nrow(wc.m), y = 1:ncol(wc.m),
  expand = 1, main = "", scale = TRUE, facets = TRUE,
  xlab = "Time lag", ylab = "Distance", zlab = "STCF", colkey = list(side = 4, length = 0.5),
  phi = 20, theta = 120, resfac = 10, col = gg2.col(100)
)
```

## Multivariate Time Series Simulation 

Similar to the one-dimensional case (i.e. using the function `generateTS()`), `CoSMoS` provides the functions `generateMTS()` and `generateMTSFast()` to simulate multiple timeseries with specified (identical) marginals and spatiotemporal correlation structures (STCs). Gaussian Vector AutoRegressive (VAR) models are used to reproduce the parent-Gaussian STFC (see [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013) and [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)). The VAR parameters see e.g. [(Biller and Nelson 2003)](https://doi.org/10.1145/937332.937333) corresponding to the desired spatiotemporal correlation function of the target process, are produced by the function `fitVAR()`. 

In this example, a set of five random locations is defined that could represent precipitation in five different places. A VAR is then fitted using  

*	a 4^th^ order process (the spatiotemporal correlations will be preserved up to temporal lag 4),  
* the Burr XII marginal distribution,  
* a probability zero `p0 = 0.8`,  
* a Clayton-Weibull spatiotemporal correlation structure (see [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)), using Weibull functions for both the temporal and spatial correlations.  
  
From the five locations, and the VAR, a set of five timeseries is generated. When the series are plotted, the spatio-temporal correlations among the series can be seen.

```{r, fig.height = 5, fig.cap = "Five spatiotemporally correlated time series simulated with `generateMTS()`.", warning=FALSE, message=FALSE}
## set a sequence of hypothetical coordinates
d <- 5
coord <- cbind(runif(d) * 30, runif(d) * 30)

## compute VAR model parameters
fit <- fitVAR(
  spacepoints = coord,
  p = 4,
  margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 25, shape = 0.7),
    tcfarg = list(scale = 3.1, shape = 0.8)
  )
)

## generate correlated timeseries
sim <- generateMTS(n = 500, STmodel = fit)

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[, 1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) +
  geom_line() +
  facet_grid(rows = vars(variable), scales = "free_y")
```

The function `generateMTSFast()` generates multiple time series with approximately separable STCS [(Serinaldi and Kilsby 2018)](https://doi.org/10.1029/2018WR023055), using an algorithm which is computationally less expensive than that in `generateMTS()`. This allows the simulation of a larger number of cross-correlated and serially dependent timeseries, at the cost of using separable STCSs and accepting some lack of accuracy in the exact reproduction of some terms of the required STCS [(Serinaldi and Kilsby 2018)](https://doi.org/10.1029/2018WR023055). 

This example uses `generateMTSFast()` with similar parameters to the previous example using `generateMTS()`.

```{r, fig.height = 5, fig.cap = "Five spatiotemporally correlated time series simulated with `generateMTSFast()`.", warning=FALSE, message=FALSE}
## set a sequence of hypothetical coordinates
d <- 5
coord <- cbind(runif(d) * 30, runif(d) * 30)

## fit and generate correlated timeseries
sim <- generateMTSFast(
  n = 500,
  spacepoints = coord,
  p0 = 0.7,
  margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull",
    scfarg = list(scale = 25, shape = 0.7),
    tcfarg = list(scale = 3.1, shape = 0.8)
  )
)

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[, 1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) +
  geom_line() +
  facet_grid(rows = vars(variable), scales = "free_y")
```

## Random Fields Simulation 

### Isotropic Random Fields

`CoSMoS` simulates spatially and temporally correlated isotropic random fields with the functions `generateRF()` and `generateRFFast()`, which are analogs of `generateMTS()` and `generateMTSFast()`, with the same syntax. The only difference is the specification of the spatial points, which is an integer denoting the side length \eqn{m} of a square grid \eqn{(mxm)}. As was mentioned in the previous section, the algorithm in `generateRFFast()` is computationally less expensive than that of `generateRF()`, enabling the simulation of random fields over a greater number of grid points (see [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013), [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331), and [Serinaldi and Kilsby (2018)](https://doi.org/10.1029/2018WR023055) for more details).

The example below shows the use of both `generateRF` and `generateRFFast`. On the test machine, `generateRF` took approximately 16 times longer than `generateRFFast` for this configuration (`fitVAR` requires roughly 15 s; the subsequent `checkRF` diagnostics roughly 20 s).

```{r, warning=FALSE, message=FALSE, error=TRUE}
## fit VAR model parameters for a 20x20 grid
fit <- fitVAR(spacepoints = 20, p = 3, margdist = "burrXII",
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton",
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2,
              scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 1.1, shape = 0.8)))

## generate isotropic random fields with nonseparable correlation structure
sim1 <- generateRF(n = 1000, STmodel = fit)

## fast simulation of isotropic random fields with separable correlation structure
sim2 <- generateRFFast(n = 1000, spacepoints = 20, p0 = 0.7, margdist = "paretoII",
                       margarg = list(scale = 1, shape = .3),
                       stcsarg = list(scfid = "weibull", tcfid = "weibull",
                       scfarg = list(scale = 20, shape = 0.7),
                       tcfarg = list(scale = 1.1, shape = 0.8)))
```

### Validating the Simulated Random Fields

The properties of the generated random fields can be checked with `checkRF()`. If the parameter `method = "stat"` (the default), the function returns the first three moments, and the probability zero at 20 grid points. If the parameter `method = "statplot"`, `checkRF()` returns a multi-panel plot comparing the theoretical and empirical spatial correlation functions at 0, 1, and 2 time lags, the contour plot of the target STCS, and the empirical and target CDFs. If the parameter `method = "fields"`, the function returns level plots of the selected number of simulated random fields (`nfields`) to visualize the temporal persistence of spatial patterns. If the parameter `method = "movie"`, the fields are written to an animated GIF called "movieRF.gif" in the current working directory.

This example returns the diagnostic plots and fields for the first simulation, which used `generateRF`.

```{r, fig.height = 6.5, fig.cap = "Validation of simulated random fields: empirical moments (top), correlation functions and CDF (middle), spatial snapshots (bottom).", warning=FALSE, message=FALSE}
## check random fields
checkRF(RF = sim1, nfields = 9*9, method = "stat")
checkRF(RF = sim1, nfields = 9*9, method = "statplot")
checkRF(RF = sim1, nfields = 9*9, method = "field")
```

### Anisotropic Random Fields

The functions `generateRF()` and `generateRFFast()` allows for the simulation of spatially and temporally correlated anisotropic random fields by specifying the arguments `anisotropyid` and `anisotropyarg` when calculating the model parameters via `fitVAR()`. Three types of anisotropic effects are supported, namely: affine (rotation and stretching), and swirl-like, and "wavy" deformation (see the documentation of the function `anisotropyT`, [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466), and references therein for more details).    

The example below shows the simulation of swirl-like random fields mimicking the shape of atmospheric cyclones. First, the transformed coordinate system is visualised:

```{r, fig.width = 4, fig.height = 4, fig.cap = "Swirl-transformed coordinate system (b = 10, α = 1.8π).", warning=FALSE, message=FALSE}
## specify a grid of coordinates
m <- 30
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

## transform the coordinate system
at <- anisotropyTswirl(
  spacepoints = coord, x0 = floor(m / 2), y0 = floor(m / 2),
  b = 10, alpha = 1.8 * pi
)

## visualize transformed coordinate system
aux <- data.frame(lon = at[, 1], lat = at[, 2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
ggplot(aux, aes(x = lon, y = lat)) +
  geom_path(aes(group = id1)) +
  geom_path(aes(group = id2)) +
  geom_point(col = 2)
```  

Then, a set of swirl-like random fields with the desired spatiotemporal correlation structure is generated and visualised:

```{r, fig.height = 7, fig.cap = "Simulated swirl-like (cyclone-shaped) random fields over a 30×30 grid.", warning=FALSE, message=FALSE}
## compute VAR model parameters
fit <- fitVAR(
  spacepoints = m, p = 1, margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.1, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 2, shape = 0.7), tcfarg = list(scale = 20.1, shape = 0.8)
  ),
  anisotropyid = "swirl",
  anisotropyarg = list(x0 = floor(m / 2), y0 = floor(m / 2), b = 10, alpha = 1.8 * pi)
)

## generate
set.seed(9)
sim3 <- generateRF(n = 25, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 5 * 5, method = "field")
```

### Lagrangian Random Fields

`CoSMoS` implements advances in space-time simulation presented in [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466) that enable the simulation of Lagrangian random fields, i.e. random fields characterized by mass transport (advection). The motion is described by velocity vectors with specified orthogonal components _u_ and _v_ at each grid point. Motion vectors can be specified by setting the arguments `advectionid` and `advectionarg` of `fitVAR` function. Several types of advection fields are supported, but users can create their own following the structure of the existing functions  (see the documentation of the function `advectionF` for more details).

The example below refers to the generation of random fields with affine anisotropy and uniform advection, mimicking rainfall storms moving north-easterly.

```{r, fig.height = 7, fig.cap = "Simulated Lagrangian random fields with affine anisotropy and uniform north-easterly advection.", warning=FALSE, message=FALSE}
## compute VAR model parameters
fit <- fitVAR(
  spacepoints = 30, p = 1, margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  anisotropyid = "affine",
  anisotropyarg = list(phi1 = 2., phi2 = 0.5, phi12 = 0, theta = -pi / 3),
  advectionid = "uniform", advectionarg = list(u = 2.5, v = 1.5)
)

## generate
set.seed(10) # 1 # 5
sim3 <- generateRF(n = 81, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 9 * 9, method = "field")
```

### Random Fields with non-Gaussian Dependence Structures 

`CoSMoS` also enables the simulation of (Lagrangian anisotropic) random fields with dependence structures (copulas) different from Gaussian. Namely, we consider the Student copula [(Demarta and McNeil, 2007)](https://doi.org/10.1111/j.1751-5823.2005.tb00254.x) and Bárdossy copula [(Bárdossy, 2006)](https://doi.org/10.1029/2005WR004754). The former is characterized by upper and lower tail dependence stronger than those of a Gaussian dependence structure, while the latter allows for asymmetric tail dependence, in particular, upper or lower tail dependence weaker than that of a Gaussian copula. Both models are parametrized by dependence parameters that are identical or directly related to the correlation values characterizing a parent multidimensional Gaussian distribution. Therefore, these models preserve pairwise correlations at multiple locations, and their simulation still relies on the algorithms available for Gaussian processes. We report below an example of correlated samples at two hypothetical locations characterized by standard Gaussian marginal distributions, identical linear correlation, but different copulas (Gaussian, Student, and Bárdossy). 
 
```{r, fig.height = 3.5, fig.cap = "Bivariate samples (n = 5000) from Gaussian, Student (df = 4), and flipped Bárdossy copulas with identical linear correlation.", warning=FALSE, message=FALSE}
## set the hypothetical coordinates of two locations
coord <- matrix(rep(c(0, 1.65), times = 2), ncol = 2)

## compute VAR model parameters
## Gaussian dependence structure
fit1 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  )
)
## Student dependence structure
fit2 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  dsid = "student", dsarg = 4
)
## Bardossy dependence structure
fit3 <- fitVAR(
  spacepoints = coord, p = 1, margdist = "norm",
  margarg = list(mean = 0, sd = 1), p0 = 0.0, stcsid = "clayton",
  stcsarg = list(
    scfid = "weibull", tcfid = "weibull", copulaarg = 2,
    scfarg = list(scale = 20, shape = 0.7),
    tcfarg = list(scale = 30.1, shape = 0.8)
  ),
  dsid = "bardossyF", dsarg = 0.5
)
## generate
set.seed(10)
sim1 <- data.frame(generateRF(n = 5000, STmodel = fit1))
sim2 <- data.frame(generateRF(n = 5000, STmodel = fit2))
sim3 <- data.frame(generateRF(n = 5000, STmodel = fit3))

## visual check
sim1 <- cbind(sim1, id = "1 - Gauss copula")
sim2 <- cbind(sim2, id = "2 - Student copula")
sim3 <- cbind(sim3, id = "3 - flipped Bárdossy copula")
dta <- rbind(sim1, sim2, sim3)

ggplot(data = dta) +
  geom_point(mapping = aes(x = X1, y = X2), alpha = 0.07) +
  facet_wrap(facets = ~id, nrow = 1)

## linear correlations
pc <- round(c(
  cor.G = cor(sim1[, 1:2])[1, 2], cor.S = cor(sim2[, 1:2])[1, 2],
  cor.B = cor(sim3[, 1:2])[1, 2]
), 3)
pc
```

***

# Stochastic Simulation of Real-World Processes {#section_3}
In practice, it is often desired to simulate a natural process based on the properties of an observed time series. This section shows how this can easily be done using `CoSMoS` functions.

## Hourly Precipitation
Precipitation can be easily simulated in `CoSMoS` from observed data. This is an example of a time series of observed hourly precipitation (units are hundredths of an inch).

```{r, fig.cap = "Observed hourly precipitation record.", warning=FALSE, message=FALSE}
data("precip")
quickTSPlot(precip$value)
```

To generate a synthetic time series with similar characteristics to the observed precipitation, you have to determine (1) the marginal distribution, (2) the autocorrelation structure, and (3) the probability zero for each individual month. This can be done by trying to fit various distributions and autocorrelation structures with `analyzeTS()` and then checking the goodness-of-fit with `reportTS()`, which can return the plots of the distribution (`method = "dist"`), ACS (`method = "acs"`) or basic statistics (`method = "stat"`). The `analyzeTS()` call is computationally intensive (approximately 75 s on a modern laptop); results below are loaded from pre-computed objects.

```{r, eval=FALSE, warning=FALSE, message=FALSE}
precip_ggamma <- analyzeTS(TS = precip, season = "month", dist = "ggamma",
                           acsID = "weibull", lag.max = 12)
```

```{r, include=FALSE}
precip_ggamma <- readRDS("figures/precip_ggamma.rds")
```

```{r, fig.height = 9, fig.cap = "Fitted Generalized Gamma distribution (top) and Weibull ACS (bottom) against monthly precipitation data.", warning=FALSE, message=FALSE}
reportTS(aTS = precip_ggamma, method = "dist") 
reportTS(aTS = precip_ggamma, method = "acs") 
```

In this example, the Generalized Gamma distribution and the Weibull autocorrelation structure describe well the observed time series. Other combinations might not give such good results. For example, the Pareto II distribution and the Fractional Gaussian Noise ACS do not give good monthly fits:

```{r, eval=FALSE, warning=FALSE, message=FALSE}
precip_pareto <- analyzeTS(TS = precip, season = "month", dist = "paretoII", acsID = "fgn", lag.max = 12)
```

```{r, include=FALSE}
precip_pareto <- readRDS("figures/precip_pareto.rds")
```

```{r, fig.height = 9, fig.cap = "Fitted Pareto II distribution (top) and FGN ACS (bottom) against monthly precipitation data: note the poor fit relative to the Generalized Gamma / Weibull combination.", warning=FALSE, message=FALSE}
reportTS(aTS = precip_pareto, method = "dist") 
reportTS(aTS = precip_pareto, method = "acs") 
```

Once you choose the marginal distribution and the autocorrelation structure you can produce as many and as long synthetic series you wish. The generated series will reproduce the distribution, correlation structure and probability zero (intermittency), using the `simulateTS()` function. This example uses the Generalized Gamma distribution and the Weibull autocorrelation structure, which were fitted above.

```{r, fig.cap = "Observed (top) and simulated (bottom) hourly precipitation.", warning=FALSE, message=FALSE}
sim_precip <- simulateTS(
  aTS = precip_ggamma, from = as.POSIXct(x = "1978-12-01 00:00:00"),
  to = as.POSIXct(x = "2008-12-01 00:00:00")
)
dta <- precip
dta[, id := "observed"]
sim_precip[, id := "simulated"]
dta <- rbind(dta, sim_precip)

ggplot(data = dta) +
  geom_line(mapping = aes(x = date, y = value)) +
  facet_wrap(facets = ~id, ncol = 1)
```

## Daily streamflow

We can also apply the same methods to daily streamflows. For this example we have used the streamflow records of the Nassawango Creek near Snow Hill (Maryland, US; USGS Id 01485500) from the US Geological Survey (USGS) repository [(https://waterdata.usgs.gov/)](https://waterdata.usgs.gov/). In this case, the fitted distribution was log-normal, and the ACS was Pareto II. The `analyzeTS()` call requires approximately 240 s on a modern laptop; results are loaded from pre-computed objects.

```{r, eval=FALSE, warning=FALSE, message=FALSE}
data("disch")
str <- analyzeTS(TS = disch, dist = "lnorm", norm = "N2", acsID = "paretoII",
                 lag.max = 20, constrain = TRUE, season = "month")
```

```{r, include=FALSE}
data("disch")
str <- readRDS("figures/str.rds")
```

```{r, fig.height = 9, fig.cap = "Fitted log-normal distribution and Pareto II ACS against monthly streamflow data for Nassawango Creek.", warning=FALSE, message=FALSE}
reportTS(aTS = str) 
```

Given the fitted values, streamflows can be simulated as:

```{r, fig.height = 3.5, fig.cap = "Observed (top) and simulated (bottom) daily streamflow for Nassawango Creek.", warning=FALSE, message=FALSE}
sim_str <- simulateTS(aTS = str)

dta <- disch
dta[, id := "observed"]
sim_str[, id := "simulated"]
dta <- rbind(dta, sim_str)

ggplot(data = dta) +
  geom_line(mapping = aes(x = date, y = value)) +
  facet_wrap(facets = ~id, ncol = 1)
```


# Recommended and Supported Distributions in `CoSMoS` {#section_4}

## Recommended Distributions for Hydroclimatic Variables
As general guideline we suggest the following distributions for different variables:

*	Precipitation: Generalized Gamma (`ggamma`), Burr type XII (`burrXII`)
*	Streamflow: Burr type III and XII (`burrIII`, `burrXII`), Generalized Gamma (`ggamma`), Lognormal (`lnorm`)
*	Relative humidity: Beta (beta), Kumaraswamy (`kumar` in [ `extraDistr`](https://cran.R-project.org/package=extraDistr)  package)
*	Wind speed: Weibull (`weibull`), Generalized Gamma (`ggamma`)
*	Temperature: Normal (`norm`)
*	Maxima of any variable: GEV (`gev`)

## Supported Distribution in `CoSMoS`

CoSMoS supports most distributions available in `R` with defined quantile, distribution, and probability density functions. Additionally, we provide four flexible distributions that we highly recommend.

| Name | Argument | Parameters |
|---|---|---|
| Burr type III | `burrIII` | `list(scale, shape1, shape2)` |
| Burr type XII | `burrXII` | `list(scale, shape1, shape2)` |
| Generalized Gamma | `ggamma` | `list(scale, shape1, shape2)` |
| Pareto II (Lomax) | `paretoII` | `list(scale, shape)` |

Full PDF, CDF, quantile, and raw moment formulae for each distribution are given below.

**Burr type III** (`burrIII`, parameters: `scale`, `shape1`, `shape2`)

$$f_{\text{BrIII}}(x)=\frac{\left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}-1} 
\left(\gamma_1^{-1} \left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}}+1\right)
^{-\gamma_1 \gamma_2-1}}{\beta } \\
F_{\text{BrIII}}(x)=
\left(\gamma_1^{-1} \left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}}+1\right)
^{-\gamma _1 \gamma _2} 
\\ Q_{\text{BrIII}}(u)=\beta  \left(\gamma _1 \left(u^{-\frac{1}{\gamma _1 \gamma _2}}-1\right)\right)^{-\gamma _2} \\
m_{\text{BrIII}}(q)=\frac{\beta ^q \gamma _1^{\gamma _2 (-q)} \Gamma \left(\left(q+\gamma _1\right) \gamma _2\right) \Gamma \left(1-q \gamma _2\right)}{\Gamma \left(\gamma _1 \gamma _2\right)}$$

**Burr type XII** (`burrXII`, parameters: `scale`, `shape1`, `shape2`)

$$f_{\text{BrXII}}(x)=\frac{\left(\frac{x}{\beta }\right)^{\gamma _1-1} \left(\gamma _2 \left(\frac{x}{\beta }\right)^{\gamma _1}+1\right)^{-\frac{1}{\gamma _1 \gamma _2}-1}}{\beta } \\
F_{\text{BrXII}}(x)=1-\left(\gamma _2 \left(\frac{x}{\beta }\right)^{\gamma _1}+1\right)^{-\frac{1}{\gamma _1 \gamma _2}} \\ 
Q_{\text{BrXII}}(u)=\beta  \left(-\frac{1-(1-u)^{-\gamma _1 \gamma _2}}{\gamma _2}\right)^{\frac{1}{\gamma _1}} \\
m_{\text{BrXII}}(q)=\frac{\beta ^q \gamma _2^{-\frac{q}{\gamma _1}-1} B\left(\frac{q+\gamma _1}{\gamma _1},\frac{1-q \gamma _2}{\gamma _1 \gamma _2}\right)}{\gamma _1}$$

**Generalized Gamma** (`ggamma`, parameters: `scale`, `shape1`, `shape2`)

$$f_{\mathcal{G}\mathcal{G}}(x)=\frac{\gamma _2 e^{-\left(\frac{x}{\beta }\right)^{\gamma _2}} \left(\frac{x}{\beta }\right)^{\gamma _1-1}}{\beta  \Gamma \left(\frac{\gamma _1}{\gamma _2}\right)} \\
F_{\mathcal{G}\mathcal{G}}(x)=Q\left(\frac{\gamma _1}{\gamma _2},0,\left(\frac{x}{\beta }\right)^{\gamma _2}\right) \\ 
Q_{\mathcal{G}\mathcal{G}}(u)=\beta  Q^{-1}\left(\frac{\gamma _1}{\gamma _2},0,u\right)^{\frac{1}{\gamma _2}} \\ 
m_{\mathcal{G}\mathcal{G}}(q)=\frac{\beta ^q \Gamma \left(\frac{q}{\gamma _2}+\frac{\gamma _1}{\gamma _2}\right)}{\Gamma \left(\frac{\gamma _1}{\gamma _2}\right)}$$

**Pareto II** (`paretoII`, also known as Lomax, parameters: `scale`, `shape`)

$$f_{\text{PII}}(x)=\frac{\left(\frac{\gamma  x}{\beta }+1\right)^{-\frac{1}{\gamma }-1}}{\beta } \\
F_{\text{PII}}(x)=1-\left(\frac{\gamma  x}{\beta }+1\right)^{-1/\gamma } \\
Q_{\text{PII}}(u)=\frac{\beta  \left((1-u)^{-\gamma }-1\right)}{\gamma } \\
m_{\text{PII}}(q)=\frac{\Gamma (q+1) \left(\frac{\beta }{\gamma }\right)^q \Gamma \left(\frac{1}{\gamma }-q\right)}{\Gamma \left(\frac{1}{\gamma }\right)}$$


# Historical Note {#section_5}

The genesis of CoSMoS dates from 2009 when Simon Michael Papalexiou developed the core of its methods while at National Technical University of Athens (NTUA), Greece, and applied them initially for multivariate rainfall modeling for the needs of an internal project at NTUA. Simon soon invited a few "colleagues" to build a start-up company, named Oktana, to develop stochastic simulation software. The details were described in a legal document [(Papalexiou, 2010)](https://www.researchgate.net/publication/333323778_Stochastic_modelling_demystified) as a first step to create proprietary software, although Oktana was never created.

In 2015, Simon with Yannis Markonis reconsidered the idea of a start-up company participating in the Climate-KIC competition. In 2017 Simon moved to California and Yannis to Prague, and they started discussing the release of CoSMoS as open-source software. The methods were published in arXiv on July 21, 2017 [(Papalexiou, 2017)](https://arxiv.org/abs/1707.06842) and few months later in Advances in Water Resources [(Papalexiou, 2018)](https://doi.org/10.1016/j.advwatres.2018.02.013), followed by a publication on disaggregation (DiPMaC) in WRR [(Papalexiou et al., 2018)](https://doi.org/10.1029/2018WR022726). In the meanwhile, Filip Strnad joined the team in 2018 and coded the first version of CoSMoS in `R`, as the original code was developed by Simon in Mathematica.

Francesco Serinaldi joined the team in 2019 initially to co-author a paper on random fields with Simon which was published in WRR in January 2020 [(Papalexiou and Serinaldi, 2020)](https://doi.org/10.1029/2019WR026331). Francesco developed the R code of CoSMoS version 2, collaborating with Filip and Kevin, and extending its functionalities to include the methods of the previously mentioned paper, that is, enabling the simulation of random fields and multivariate time series. The last addition to our team is Kevin Shook who has become the package maintainer and offers advice on `R`.

The latest version of CoSMoS implements major advances in space time modeling introduced in [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466) that allow realistic simulations of hydro-environmental fluxes, including rainfall storms, fields resembling weather cyclones, fields converging to (or diverging from) a point, colliding air masses, and more. 

We are glad to see this ten-year Odyssey finally back on track to Ithaca. Leaving the burdens of the past behind, empowered by a great team and standing for dignity and academic integrity, we will keep striving to offer free, innovative, and reliable software for stochastic modelling.


# References {#section_6}

1. Papalexiou, S.M. (2010). _Stochastic modelling demystified_. <https://doi.org/10.13140/RG.2.2.34889.60008>
2. Papalexiou, S.M. (2017). _A unified theory for exact stochastic modelling of univariate and multivariate processes with continuous, mixed type, or discrete marginal distributions and any correlation structure_. arXiv:1707.06842. <https://arxiv.org/abs/1707.06842>
3. Papalexiou, S.M. (2018). _Unified theory for stochastic modelling of hydroclimatic processes: Preserving marginal distributions, correlation structures, and intermittency_. Advances in Water Resources, 115, 234–252. <https://doi.org/10.1016/j.advwatres.2018.02.013>
4. Papalexiou, S.M., Markonis, Y., Lombardo, F., AghaKouchak, A., Foufoula-Georgiou, E. (2018). _Precise Temporal Disaggregation Preserving Marginals and Correlations (DiPMaC) for Stationary and Nonstationary Processes_. Water Resources Research, 54(10), 7435–7458. <https://doi.org/10.1029/2018WR022726>
5. Papalexiou, S.M., Serinaldi, F. (2020). _Random Fields Simplified: Preserving Marginal Distributions, Correlations, and Intermittency, With Applications From Rainfall to Humidity_. Water Resources Research, 56(2), e2019WR026331. <https://doi.org/10.1029/2019WR026331>
6. Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). _Advancing Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond_. Water Resources Research, 57, e2020WR029466. <https://doi.org/10.1029/2020WR029466>
7. Serinaldi, F., Kilsby, C.G. (2018). _Unsurprising Surprises: The Frequency of Record-breaking and Overthreshold Hydrological Extremes Under Spatial and Temporal Dependence_. Water Resources Research, 54(9), 6460–6487. <https://doi.org/10.1029/2018WR023055>

---
title: "CoSMoS R | Complete Stochastic Modelling Solution"
author: "Simon Michael Papalexiou, Francesco Serinaldi, Filip Strnad, Yannis Markonis, Kevin Shook"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    number_sections: yes
vignette: >
  %\VignetteIndexEntry{CoSMoS R | Complete Stochastic Modelling Solution}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
  header-includes:
   - \usepackage{amsmath}
editor_options: 
  chunk_output_type: inline
---

***

```{r setup, include=FALSE}
knitr::opts_chunk$set(eval = TRUE,
                      echo = TRUE,
                      fig.width = 7, 
                      warning = FALSE,
                      message = FALSE)
library(CoSMoS)
library(plot3D)
library(data.table)
```
`CoSMoS` was conceived back in 2009 (see note in section 5) and was officially released on CRAN in April 2019. Since then `CoSMoS` has become one of the leading and most widely downloaded R packages for stochastic simulation of non-Gaussian time series. Designed with the end user in mind, **`CoSMoS` makes univariate, multivariate, or random field simulations easy**. You can reliably generate time series or fields of rainfall, streamflow, wind speed, relative humidity, or any other environmental variable in seconds. Just choose the probability distribution and correlation properties of the time series you want to generate, and it will do the rest.

<br>
~_Unified Theory of Stochastic Modelling_ | [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013)

_"Hydroclimatic processes come in all “shapes and sizes”. They are characterized by different spatiotemporal correlation structures and probability distributions that can be continuous, mixed-type, discrete or even binary. Simulating such processes by reproducing precisely their marginal distribution and linear correlation structure, including features like intermittency, can greatly improve hydrological analysis and design. ... Here, a single framework is proposed that unifies, extends, and improves a general-purpose modelling strategy, based on the assumption that any process can emerge by transforming a specific “parent”" Gaussian process."_

~_Precise Temporal Disaggregation_ | [Papalexiou et al. (2018)](https://doi.org/10.1029/2018WR022726)

_"Hydroclimatic variables such as precipitation and temperature are often measured or simulated by climate models at coarser spatiotemporal scales than those needed for operational purposes. … Here we introduce a novel disaggregation method, named Disaggregation Preserving Marginals and Correlations (DiPMaC), that is able to disaggregate a coarse-scale time series to any finer scale, while reproducing the probability distribution and the linear correlation structure of the fine-scale process. DiPMaC is also generalized for arbitrary nonstationary scenarios to reproduce time varying marginals."_

~_Random Fields Simplified_ | [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)

_"Nature manifests itself in space and time. The spatiotemporal complexity of processes such as precipitation, temperature, and wind, does not allow purely deterministic modeling. Spatiotemporal random fields have a long history in modeling such processes, and yet a single unified framework offering the flexibility to simulate processes that may differ profoundly does not exist. Here we introduce a blueprint to efficiently simulate spatiotemporal random fields that preserve any marginal distribution, any valid spatiotemporal correlation structure, and intermittency."_

~_Advancing Space-Time Simulation_ | [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466)


_"...we advance random field simulation by introducing the concepts of general velocity fields and general anisotropy transformations. To illustrate the potential of CoSMoS, we simulate random fields with complex patterns and motion mimicking rainfall storms moving across an area, spiraling fields resembling weather cyclones, fields converging to (or diverging from) a point, and colliding air masses."_

***
# Step by step guide to timeseries simulation {#section_1}

NOTE: For R code chunks running longer than ~10s, we report the CPU time for a Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz, 4-core, 32GB RAM.

In most cases we can  accurately simulate a process by reproducing its marginal distribution and its autocorrelation structure. The first is typically represented by a parametric probability distribution; the second either by a parametric correlation structure, or by specific correlation coefficient values. `CoSMoS` is the first software that enables the generation of time series while preserving any marginal distribution, correlation structure, and intermittency. Intermittency can be defined by the probability of zero values, and is essential in the simulation of precipitation at fine scales (e.g., daily, hourly) but also for any other intermittent process, such as flows of ephemeral streams, wind speed, etc.

Accurate generation of time series requires knowledge of the statistical properties of the process under study. Processes such as precipitation, streamflow, temperature, wind, or evaporation may have very different properties; that is, they can be described by different probability distributions, correlation structures, and may or may not be intermittent. 

There are four main steps required to generate a time series:
 1. Choosing a probability distribution
 2. Choosing a correlation structure, 
 3. Adding intermittency, and
 4. Validating the results.

The four steps are explained below.

## Choosing a Probability Distribution

The choice of a probability distribution to describe the variable you wish to simulate is crucial. The probability distribution expresses how frequently specific magnitudes of the variable occur in the data set. Air **temperature** is typically described by bell-shaped distributions like the [Normal](https://en.wikipedia.org/wiki/Normal_distribution). **Relative humidity** needs distributions defined in the interval (0,1) such as the [Beta](https://en.wikipedia.org/wiki/Beta_distribution) or the [Kumaraswamy](https://en.wikipedia.org/wiki/Kumaraswamy_distribution). **Streamflow** typically has heavy tails and power-type distributions might be needed to reproduce the behavior of the extremes. **Wind speed** typically can be modelled by positively skewed distributions such as the [Weibull](https://en.wikipedia.org/wiki/Weibull_distribution). **Precipitation** at fine scales has to be modelled by mixed-type (or else zero inflated) distributions, that is, having a probability value to describe the occurrence of zeros, and a continuous distribution, such the [Generalized Gamma](https://en.wikipedia.org/wiki/Generalized_gamma_distribution) or the [Burr type XII](https://en.wikipedia.org/wiki/Burr_distribution), to describe the frequency of nonzero values. 

`CoSMoS` supports **all the standard distribution** functions of `R` and from other packages (tens of distributions) but also comes with some built-in distributions such as Generalized Gamma, the [Pareto type II](https://en.wikipedia.org/wiki/Pareto_distribution), and the Burr type III and XII distributions. More details on the `CoSMoS` distributions and their parameters can be found in [Section 4](#section_4). 

Distributions may have differing numbers and types of parameters (e.g. location, scale, and shape parameters). For example, the Generalized Gamma distribution has one scale and two shape parameters.

`CoSMoS` has built-in functions to fit distributions to data and assess their goodness-of-fit based on graphs; but you can also use any other package for distribution fitting. 

## Choosing a Correlation Structure

The correlation structure expresses how much the random variables depend upon each other. Processes at fine temporal scales are typically more correlated than those at coarse scales. CoSMoS offers two options to introduce correlations: (1) by defining specific lag autocorrelations as a list starting from lag 0 up to a desired limit, or (2) by using a pre-defined parametric autocorrelation structure.

For example, you can generate a time series with lag-1 autocorrelation $\rho_1 = 0.8$ and Generalized Gamma marginal distribution by:

```{r, warning=FALSE, message=FALSE}
## (i) specifying the sample size
no <- 1000
## (ii) defining the type of marginal distribution and its parameters
marginaldist <- "ggamma"
param <- list(scale = 1, shape1 = .8, shape2 = .8)
## (iii) defining the desired autocorrelation
acf.my <- c(1, 0.8)
## (iv) simulating
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf.my)
## and (v) visually checking the generated time series
quickTSPlot(ggamma_sim[[1]]) 

```

You can specify also as many autocorrelation values as you wish. In this example, autocorrelation values are specified up to lag 4:

```{r, warning=FALSE, message=FALSE}
acf <- c(1, 0.6, 0.5, 0.4, 0.3) #up to lag-4
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
quickTSPlot(ggamma_sim[[1]])
```
A better approach is to use an autocorrelation structure expressed by a parametric function. `CoSMoS` has built in autocorrelation structures than can be used by the `acs()` function. You can use the following ACS structures:

*	Fractional Gaussian Noise; this is the well-known one-parameter fGn ACS.
*	Weibull; a flexible two parameter of exponential form.
*	Pareto II: a two parameter power-type ACS.
*	Burr XII: a three parameter power-type ACS. 

For most practical applications a two-parameter ACS is sufficient; we suggest either the Weibull or the Pareto II. 


The `acs()` function produces autocorrelation values based on any of the four structures and any desired lag. Thus, instead of setting each autocorrelation coefficient explicitly (which might be problematic from a technical viewpoint), we can generate series preserving any one of the four ACSs. For example, you can easily define an ACS and visualize its values up to any lag. Here, all four ACSs are defined and visualized up to lag 10. You can see how changing the ACS function changes how the correlation coefficients decay to zero.

```{r, warning=FALSE, message=FALSE}
## specify lag
lags <- 0:10

## get the ACS
f <- acs(id = "fgn", t = lags, H = .75)
b <- acs(id = "burrXII", t = lags, scale = 1, shape1 = .6, shape2 = .4)
w <- acs(id = "weibull", t = lags, scale = 2, shape = 0.8)
p <- acs(id = "paretoII", t = lags, scale = 3, shape = 0.3)

## visualize the ACS
dta <- data.table(lags, f, b, w, p)
m.dta <- melt(data = dta, id.vars = "lags")

ggplot(data = m.dta, mapping = aes(x = lags, y = value, group = variable, colour = variable)) + 
  geom_point(size = 2.5) + geom_line(lwd = 1) +
  scale_color_manual(values = c("steelblue4", "red4", "green4", "darkorange"), 
  labels = c("FGN", "Burr XII", "Weibull", "Pareto II"), name = "") +
  labs(x = bquote(lag ~ tau), y = "ACS") + scale_x_continuous(breaks = lags) + theme_light()
```

We can re-generate the previously generated series which used explicity defined correlations up to lag 4, now with a two parameter Pareto II ACF up to lag 30, which improves the modelling parsimony. More details about the parametric autocorrelation structures can be found in section 3.2 in [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013).

```{r, warning=FALSE, message=FALSE}
acf <- acs(id = "paretoII", t = 0:30, scale = 1, shape = .75)
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf)
dta <- data.frame(time = 1:no, value = ggamma_sim[[1]])

quickTSPlot(dta$value)
```

Apart from the four autocorrelation functions provided by `acs()`, we can also create our own. Note that it is important to ensure that the function is positive definite, otherwise the autoregressive model cannot be fitted. This example shows the generation of a time series with the same Generalized Gamma marginal distribution as above, but with a user-defined exponential ACS function up to lag 30. Note that although the time series density is very similar to that in the previous plot, the texture of the time series is very different, due to the changed ACS function.

```{r, warning=FALSE, message=FALSE}
my_acf <- exp(seq(0, -2, -0.1))
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = my_acf)
quickTSPlot(ggamma_sim[[1]])

```

## Adding Intermittency (if required)

`CoSMoS` easily simulates intermittent processes such as precipitation. The only extra step needed is to define the probability of zero events. For example, say you wish to generate 5 mutually independent timeseries having the previously defined Generalized Gamma distribution with the Pareto II ACS and probability zero p0 = 0.9. The generated data thus will have 90% of zero values (i.e. dry days). 

```{r, fig.height = 7, warning=FALSE, message=FALSE}
prob_zero <- .9
## the argument `TSn = 5` enables the simulation of 5 timeseries.
ggamma_sim <- generateTS(n = no, margdist = marginaldist, margarg = param, acsvalue = acf, 
                         p0 = prob_zero, TSn = 5)
plot(x = ggamma_sim, main = "") + theme_light()
```

## Validating the Results

You can readily check some basic statistics of the generated time series using the `checkTS()` function, which return the first three moments, probability zero and the first three autocorrelation coefficients.

```{r, warning=FALSE, message=FALSE}
checkTS(ggamma_sim)
```

***

# Multivariate and Random Fields Simulation  {#section_2}

The above methods can readily be extended to multidimensional cases, thus enabling the simulation of spatiotemporally correlated random vectors (i.e. correlated timeseries at multiple locations) and random fields (i.e. gridded data), as discussed in detail by [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331). This requires the definition of suitable spatiotemporal correlation structures (see [Porcu et al. (2020)](https://doi.org/10.1002/wics.1512) for a thorough review of this topic).

## Spatiotemporal Correlation Structures

CoSMoS allows the definition of spatiotemporal correlation functions using the function `stcs()`, which the spatiotemporal counterpart of the purely temporal `acs()`. Two classes of spatiotemporal correlation functions are provided:

*	Clayton [(Papalexiou and Serinaldi 2020)](https://doi.org/10.1029/2019WR026331)
*	Gneiting [(Gneiting 2002)](https://doi.org/10.1198/016214502760047113)  

The `stcs()` function can produce values of the linear spatiotemporal correlation for any desired time lag and spatial distance using these two correlation function classes, which comprise a variety of structures covering most cases of practical interest. This example shows the Clayton-Weibull spatiotemporal correlation structure:

```{r, fig.height = 7, warning=FALSE, message=FALSE}

## specify grid of spatial and temporal lags
d <- 51
st <- expand.grid(0:(d - 1), 0:(d - 1))

## get the STCS
wc <- stcfclayton(t = st[, 1], s = st[, 2], scfid = "weibull", tcfid = "weibull", copulaarg = 2,
                  scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 5.1,shape = 0.8))

## visualize the STCS
wc.m <- matrix(data = wc, nrow = d)
j <- tail(which(wc.m[1, ] > 0.15), 1)
i <- tail(which(wc.m[, 1] > 0.15), 1)
wc.m <- wc.m[1:i, 1:j]

persp3D(z = wc.m, x = 1: nrow(wc.m), y = 1:ncol(wc.m),
expand = 1, main = "", scale = TRUE, facets = TRUE,
xlab="Time lag", ylab = "Distance", zlab = "STCF", colkey = list(side = 4, length = 0.5),
phi = 20, theta = 120, resfac = 5,  col= gg2.col(100))
```

## Multivariate Time Series Simulation 

Similar to the one-dimensional case (i.e. using the function `generateTS()`), `CoSMoS` provides the functions `generateMTS()` and `generateMTSFast()` to simulate multiple timeseries with specified (identical) marginals and spatiotemporal correlation structures (STCs). Gaussian Vector AutoRegressive (VAR) models are used to reproduce the parent-Gaussian STFC (see [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013) and [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)). The VAR parameters see e.g. [(Biller and Nelson 2003)](https://doi.org/10.1145/937332.937333) corresponding to the desired spatiotemporal correlation function of the target process, are produced by the function `fitVAR()`. 

In this example, a set of five random locations is defined that could represent precipitation in five different places. A VAR is then fitted using  

*	a 4^th^ order process (the spatiotemporal correlations will be preserved up to temporal lag 4),  
* the Burr XII marginal distribution,  
* a probability zero `p0 = 0.8`,  
* a Clayton-Weibull spatiotemporal correlation structure (see [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331)), using Weibull functions for both the temporal and spatial correlations.  
  
From the five locations, and the VAR, a set of five timeseries is generated. When the series are plotted, the spatio-temporal correlations among the series can be seen.

```{r, fig.height = 5, warning=FALSE, message=FALSE}
## set a sequence of hypothetical coordinates 
d <- 5
coord <- cbind(runif(d)*30, runif(d)*30)

## compute VAR model parameters  
fit <- fitVAR(spacepoints = coord, 
              p = 4, 
              margdist = "burrXII",
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), 
              p0 = 0.8, 
              stcsid = "clayton",
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2,
                   scfarg = list(scale = 25, shape = 0.7), 
                   tcfarg = list(scale = 3.1, shape = 0.8) ) )

## generate correlated timeseries  
sim <- generateMTS(n = 500, STmodel = fit)

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[,1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) + geom_line() +
       facet_grid(rows = vars(variable), scales = "free_y") + theme_light()
```

The function `generateMTSFast()` generates multiple time series with approximately separable STCS [(Serinaldi and Kilsby 2018)](https://doi.org/10.1029/2018WR023055), using an algorithm which is computationally less expensive than that in `generateMTS()`. This allows the simulation of a larger number of cross-correlated and serially dependent timeseries, at the cost of using separable STCSs and accepting some lack of accuracy in the exact reproduction of some terms of the required STCS [(Serinaldi and Kilsby 2018)](https://doi.org/10.1029/2018WR023055). 

This example uses `generateMTSFast()` with similar parameters to the previous example using `generateMTS()`.

```{r, fig.height = 5, warning=FALSE, message=FALSE}
## set a sequence of hypothetical coordinates
d <- 5
coord <- cbind(runif(d)*30, runif(d)*30)

## fit and generate correlated timeseries
sim <- generateMTSFast(n = 500, 
                       spacepoints = coord, 
                       p0 = 0.7, 
                       margdist ="burrXII",
                       margarg = list(scale = 3, shape1 = .9, shape2 = .2),  
                       stcsarg = list(scfid = "weibull", tcfid = "weibull",
                       scfarg = list(scale = 25, shape = 0.7),
                       tcfarg = list(scale = 3.1, shape = 0.8)) )

## visualize simulated timeseries
dta <- melt(data = data.table(time = 1:nrow(sim), sim[,1:d]), id.vars = "time")

ggplot(data = dta, mapping = aes(x = time, y = value)) + geom_line() +
       facet_grid(rows = vars(variable), scales = "free_y") + theme_light()
```

## Random Fields Simulation 

### Isotropic Random Fields

`CoSMoS` simulates spatially and temporally correlated isotropic random fields with the functions `generateRF()` and `generateRFFast()`, which are analogs of `generateMTS()` and `generateMTSFast()`, with the same syntax. The only difference is the specification of the spatial points, which is an integer denoting the side length \eqn{m} of a square grid \eqn{(mxm)}. As was mentioned in the previous section, the algorithm in `generateRFFast()` is computationally less expensive than that of `generateRF()`, enabling the simulation of random fields over a greater number of grid points (see [Papalexiou (2018)](https://doi.org/10.1016/j.advwatres.2018.02.013), [Papalexiou and Serinaldi (2020)](https://doi.org/10.1029/2019WR026331), and [Serinaldi and Kilsby (2018)](https://doi.org/10.1029/2018WR023055) for more details).

The example below shows the use of both `generateRF` and `generateRFFast`. The generation of random fields using `generateRF` took approximately 16 times as long as using `generateRFFast`.

```{r, warning=FALSE, message=FALSE}
## compute VAR model parameters
## CPU time: ~15s 
fit <- fitVAR(spacepoints = 20, p = 3, margdist ="burrXII",
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton",
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2,
              scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 1.1, shape = 0.8) ) )

## generate isotropic random fields with nonseparable correlation structure 
sim1 <- generateRF(n = 1000, STmodel = fit)

## fast simulation of isotropic random fields with separable correlation structure
sim2 <- generateRFFast(n = 1000, spacepoints = 20, p0 = 0.7, margdist ="paretoII",
                       margarg = list(scale = 1, shape = .3),
                       stcsarg = list(scfid = "weibull", tcfid = "weibull",
                       scfarg = list(scale = 20, shape = 0.7), 
                       tcfarg = list(scale = 1.1, shape = 0.8)))

```

### Validating the Simulated Random Fields

The properties of the generated random fields can be checked with `checkRF()`. If the parameter `method = "stat"` (the default), the function returns the first three moments, and the probability zero at 20 grid points. If the parameter `method = "statplot"`, `checkRF()` returns a multi-panel plot comparing the theoretical and empirical spatial correlation functions at 0, 1, and 2 time lags, the contour plot of the target STCS, and the empirical and target CDFs. If the parameter `method = "fields"`, the function returns level plots of the selected number of simulated random fields (`nfields`) to visualize the temporal persistence of spatial patterns. If the parameter `method = "movie"`, the fields are written to an animated GIF called "movieRF.gif" in the current working directory.

This example returns the stats, stat plots and fields for the first simulation, which used `generateRF`.

```{r, fig.height = 6.5, warning=FALSE, message=FALSE}
## check random fields
## CPU time: ~20s 
checkRF(RF = sim1, nfields = 9*9, method = "stat")
checkRF(RF = sim1, nfields = 9*9, method = "statplot")
checkRF(RF = sim1, nfields = 9*9, method = "field")
```

### Anisotropic Random Fields

The functions `generateRF()` and `generateRFFast()` allows for the simulation of spatially and temporally correlated anisotropic random fields by specifying the arguments `anisotropyid` and `anisotropyarg` when calculating the model parameters via `fitVAR()`. Three types of anisotropic effects are supported, namely: affine (rotation and stretching), and swirl-like, and "wavy" deformation (see the documentation of the function `anisotropyT`, [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466), and references therein for more details).    

The example below reports the simulation of swirl-like random fields mimicking for instance the shape of atmospheric cyclones. Firstly, one can visualize the  the transformed of the coordinate system   

```{r, fig.width = 4, fig.height = 4, warning=FALSE, message=FALSE}
## specify a grid of coordinates
m = 30
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

## transform the coordinate system
at <- anisotropyTswirl(spacepoints = coord, x0 = floor(m / 2), y0 = floor(m / 2), 
                       b = 10, alpha = 1.8 * pi)

## visualize transformed coordinate system
  aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
  ggplot(aux, aes(x = lon, y = lat)) + 
    geom_path(aes(group = id1)) + geom_path(aes(group = id2)) + geom_point(col = 2) + theme_light()

```  

Then, one can generate a set of swirl-like random fields with the desired spatiotemporal correlation structure and visualize them

```{r, fig.height = 7, warning=FALSE, message=FALSE}
## compute VAR model parameters
fit <- fitVAR(spacepoints = m, p = 1, margdist = 'burrXII', 
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.1, stcsid = "clayton", 
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2,
                        scfarg = list(scale = 2, shape = 0.7), tcfarg = list(scale = 20.1, shape = 0.8)),
              anisotropyid = "swirl", 
              anisotropyarg = list(x0 = floor(m / 2), y0 = floor(m / 2), b = 10, alpha = 1.8 * pi) )

## generate
set.seed(9)
sim3 <- generateRF(n=25, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 5*5, method = "field")

```

### Lagrangian Random Fields

`CoSMoS` implements advances in space-time simulation presented in [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466) that enable the simulation of Lagrangian random fields, i.e. random fields characterized by mass transport (advection). The motion is described by velocity vectors with specified orthogonal components _u_ and _v_ at each grid point. Motion vectors can be specified by setting the arguments `advectionid` and `advectionarg` of `fitVAR` function. Several types of advection fields are supported, but users can create their own following the structure of the existing functions  (see the documentation of the function `advectionF` for more details).

The example below refers to the generation of random fields with affine anisotropy and uniform advection, mimicking rainfall storms moving north-easterly.

```{r, fig.height = 7.0, warning=FALSE, message=FALSE}
## compute VAR model parameters
fit <- fitVAR(spacepoints = 30, p = 1, margdist = 'burrXII', 
              margarg = list(scale = 3, shape1 = .9, shape2 = .2), p0 = 0.8, stcsid = "clayton", 
              stcsarg = list(scfid = "weibull", tcfid = "weibull", copulaarg = 2, 
                      scfarg = list(scale = 20, shape = 0.7), tcfarg = list(scale = 30.1, shape = 0.8)),
              anisotropyid = "affine", 
              anisotropyarg = list(phi1 = 2., phi2 = 0.5, phi12 = 0, theta = -pi/3),
              advectionid = "uniform", advectionarg = list(u = 2.5, v = 1.5) )

## generate
set.seed(10) # 1 # 5
sim3 <- generateRF(n=81, STmodel = fit)

## check
checkRF(RF = sim3, nfields = 9*9, method = "field")

```

***

# Stochastic Simulation of Real-World Processes {#section_3}
In practice, it is often desired to simulate a natural process based on the properties of an observed time series. This section shows how this can easily be done using `CoSMoS` functions.

## Hourly Precipitation
Precipitation can be easily simulated in `CoSMoS` from observed data. This is an example of a time series of observed hourly precipitation (units are hundredths of an inch).

```{r, warning=FALSE, message=FALSE}
data("precip")
quickTSPlot(precip$value)
```

To generate a synthetic time series with similar characteristics to the observed precipitation, you have to determine (1) the marginal distribution, (2) the autocorrelation structure, and (3) the probability zero for each individual month. This can be done by trying to fit various distributions and autocorrelation structures with `analyzeTS()` and then checking the goodness-of-fit with `reportTS()`, which can return the plots of the distribution (`method = "dist"`), ACS (`method = "acs"`) or basic statistics (`method = "stat"`).

```{r, fig.height = 9, warning=FALSE, message=FALSE}
## CPU time: ~75s 
precip_ggamma <- analyzeTS(TS = precip, season = "month", dist = "ggamma",
                           acsID = "weibull", lag.max = 12)

reportTS(aTS = precip_ggamma, method = "dist") + theme_light()
reportTS(aTS = precip_ggamma, method = "acs") + theme_light()
reportTS(aTS = precip_ggamma, method = "stat")
```

In this example, the Generalized Gamma distribution and the Weibull autocorrelation structure describe well the observed time series. Other combinations might not give such good results. For example, the Pareto II distribution and the Fractional Gaussian Noise ACS do not give good monthly fits:

```{r, fig.height = 9, warning=FALSE, message=FALSE}
precip_pareto <- analyzeTS(TS = precip, season = "month", dist = "paretoII", acsID = "fgn", lag.max = 12)

reportTS(aTS = precip_pareto, method = "dist")+ theme_light()
reportTS(aTS = precip_pareto, method = "acs") + theme_light()
```

Once you choose the marginal distribution and the autocorrelation structure you can produce as many and as long synthetic series you wish. The generated series will reproduce the distribution, correlation structure and probability zero (intermittency), using the `simulateTS()` function. This example uses the Generalized Gamma distribution and the Weibull autocorrelation structure, which were fitted above.

```{r, warning=FALSE, message=FALSE}
sim_precip <- simulateTS(aTS = precip_ggamma, from = as.POSIXct(x = "1978-12-01 00:00:00"),
                         to = as.POSIXct(x = "2008-12-01 00:00:00"))
dta <- precip
dta[, id := "observed"]
sim_precip[, id := "simulated"]
dta <- rbind(dta, sim_precip)

ggplot(data = dta) + geom_line(mapping = aes(x = date, y = value)) + 
  facet_wrap(facets = ~id, ncol = 1) + theme_light()
```

## Daily streamflow

We can also apply the same methods to daily streamflows. For this example we have used the streamflow records of the Nassawango Creek near Snow Hill (Maryland, US; USGS Id 01485500) from the US Geological Survey (USGS) repository [(https://waterdata.usgs.gov)](https://waterdata.usgs.gov). In this case, the fitted distribution was log-normal, and the ACS was Pareto II.

```{r, fig.height = 9, warning=FALSE, message=FALSE}
## CPU time: ~240s
data("disch")

str <- analyzeTS(TS = disch, dist = "lnorm", norm = "N2", acsID = "paretoII", 
                 lag.max = 20, constrain = TRUE, season = "month")

reportTS(aTS = str) + theme_light()
reportTS(aTS = str, method = "stat")
```

Given the fitted values, streamflows can be simulated as

```{r, fig.height = 3.5, warning=FALSE, message=FALSE}
sim_str <- simulateTS(aTS = str)

dta <- disch
dta[, id := "observed"]
sim_str[, id := "simulated"]
dta <- rbind(dta, sim_str)

ggplot(data = dta) + geom_line(mapping = aes(x = date, y = value)) + 
  facet_wrap(facets = ~id, ncol = 1) + theme_light()
```


# Recommended and Supported Distributions in `CoSMoS` {#section_4}

## Recommended Distributions for Hydroclimatic Variables
As general guideline we suggest the following distributions for different variables:

*	Precipitation: Generalized Gamma (`ggamma`), Burr type XII (`burrXII`)
*	Streamflow: Burr type III and XII (`burrIII`, `burrXII`), Generalized Gamma (`ggamma`), Lognormal (`lnorm`)
*	Relative humidity: Beta (beta), Kumaraswamy (`kumar` in [ `extraDistr`](https://cran.R-project.org/package=extraDistr)  package)
*	Wind speed: Weibull (`weibull`), Generalized Gamma (`ggamma`)
*	Temperature: Normal (`norm`)
*	Maxima of any variable: GEV (`gev`)

## Supported Distribution in `CoSMoS`
CoSMoS supports most distributions available in `R` with defined quantile, distribution, and probability density functions. Additionally, we provide four flexible distributions that we highly recommend.  

**Burr type III**

distribution argument/name - `burrIII`    
parameters - `list(scale, shape1, shape2)`

$$f_{\text{BrIII}}(x)=\frac{\left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}-1} 
\left(\gamma_1^{-1} \left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}}+1\right)
^{-\gamma_1 \gamma_2-1}}{\beta } \\
F_{\text{BrIII}}(x)=
\left(\gamma_1^{-1} \left(\frac{x}{\beta }\right)^{-\frac{1}{\gamma _2}}+1\right)
^{-\gamma _1 \gamma _2} 
\\ Q_{\text{BrIII}}(u)=\beta  \left(\gamma _1 \left(u^{-\frac{1}{\gamma _1 \gamma _2}}-1\right)\right)^{-\gamma _2} \\
m_{\text{BrIII}}(q)=\frac{\beta ^q \gamma _1^{\gamma _2 (-q)} \Gamma \left(\left(q+\gamma _1\right) \gamma _2\right) \Gamma \left(1-q \gamma _2\right)}{\Gamma \left(\gamma _1 \gamma _2\right)}$$

**Burr type XII**

distribution argument/name - `burrXII`     
parameters - `list(scale, shape1, shape2)`

$$f_{\text{BrXII}}(x)=\frac{\left(\frac{x}{\beta }\right)^{\gamma _1-1} \left(\gamma _2 \left(\frac{x}{\beta }\right)^{\gamma _1}+1\right)^{-\frac{1}{\gamma _1 \gamma _2}-1}}{\beta } \\
F_{\text{BrXII}}(x)=1-\left(\gamma _2 \left(\frac{x}{\beta }\right)^{\gamma _1}+1\right)^{-\frac{1}{\gamma _1 \gamma _2}} \\ 
Q_{\text{BrXII}}(u)=\beta  \left(-\frac{1-(1-u)^{-\gamma _1 \gamma _2}}{\gamma _2}\right)^{\frac{1}{\gamma _1}} \\
m_{\text{BrXII}}(q)=\frac{\beta ^q \gamma _2^{-\frac{q}{\gamma _1}-1} B\left(\frac{q+\gamma _1}{\gamma _1},\frac{1-q \gamma _2}{\gamma _1 \gamma _2}\right)}{\gamma _1}$$

**Generalized Gamma**

distribution argument/name - `ggamma`    
parameters - `list(scale, shape1, shape2)`

$$f_{\mathcal{G}\mathcal{G}}(x)=\frac{\gamma _2 e^{-\left(\frac{x}{\beta }\right)^{\gamma _2}} \left(\frac{x}{\beta }\right)^{\gamma _1-1}}{\beta  \Gamma \left(\frac{\gamma _1}{\gamma _2}\right)} \\
F_{\mathcal{G}\mathcal{G}}(x)=Q\left(\frac{\gamma _1}{\gamma _2},0,\left(\frac{x}{\beta }\right)^{\gamma _2}\right) \\ 
Q_{\mathcal{G}\mathcal{G}}(u)=\beta  Q^{-1}\left(\frac{\gamma _1}{\gamma _2},0,u\right)^{\frac{1}{\gamma _2}} \\ 
m_{\mathcal{G}\mathcal{G}}(q)=\frac{\beta ^q \Gamma \left(\frac{q}{\gamma _2}+\frac{\gamma _1}{\gamma _2}\right)}{\Gamma \left(\frac{\gamma _1}{\gamma _2}\right)}$$

**Pareto II** (also known as Lomax)

distribution argument/name - `paretoII`     
parameters - `list(scale, shape)`

$$f_{\text{PII}}(x)=\frac{\left(\frac{\gamma  x}{\beta }+1\right)^{-\frac{1}{\gamma }-1}}{\beta } \\
F_{\text{PII}}(x)=1-\left(\frac{\gamma  x}{\beta }+1\right)^{-1/\gamma } \\
Q_{\text{PII}}(u)=\frac{\beta  \left((1-u)^{-\gamma }-1\right)}{\gamma } \\
m_{\text{PII}}(q)=\frac{\Gamma (q+1) \left(\frac{\beta }{\gamma }\right)^q \Gamma \left(\frac{1}{\gamma }-q\right)}{\Gamma \left(\frac{1}{\gamma }\right)}$$


# Historical Note {#section_5}

The genesis of CoSMoS dates from 2009 when Simon Michael Papalexiou developed the core of its methods while at National Technical University of Athens (NTUA), Greece, and applied them initially for multivariate rainfall modeling for the needs of an internal project at NTUA. Simon soon invited a few "colleagues" to build a start-up company, named Oktana, to develop stochastic simulation software. The details were described in a legal document [(Papalexiou, 2010)](http://dx.doi.org/10.13140/RG.2.2.34889.60008) as a first step to create proprietary software, although Oktana was never created.

In 2015, Simon with Yannis Markonis reconsidered the idea of a start-up company participating in the Climate-KIC competition. In 2017 Simon moved to California and Yannis to Prague, and they started discussing the release of CoSMoS as open-source software. The methods were published in arXiv on July 21, 2017 [(Papalexiou, 2017)](https://arxiv.org/abs/1707.06842) and few months later in Advances in Water Resources [(Papalexiou, 2018)](https://doi.org/10.1016/j.advwatres.2018.02.013), followed by a publication on disaggregation (DiPMaC) in WRR [(Papalexiou et al., 2018)](https://doi.org/10.1029/2018WR022726). In the meanwhile, Filip Strnad joined the team in 2018 and coded the first version of CoSMoS in `R`, as the original code was developed by Simon in Mathematica.

Francesco Serinaldi joined the team in 2019 initially to co-author a paper on random fields with Simon which was published in WRR in January 2020 [(Papalexiou and Serinaldi, 2020)](https://doi.org/10.1029/2019WR026331). Francesco developed the R code of CoSMoS version 2, collaborating with Filip and Kevin, and extending its functionalities to include the methods of the previously mentioned paper, that is, enabling the simulation of random fields and multivariate time series. The last addition to our team is Kevin Shook who has become the package maintainer and offers advice on `R`.

The latest version of CoSMoS implements major advances in space time modeling introduced in [Papalexiou, Serinaldi and Porcu (2021)](https://doi.org/10.1029/2020WR029466) that allow realistic simulations of hydro-environmental fluxes, including rainfall storms, fields resembling weather cyclones, fields converging to (or diverging from) a point, colliding air masses, and more. 

We are serene to see this ten-year Odyssey finally back on track to Ithaca. Leaving the burdens of the past behind, empowered by a great team and standing for dignity and academic integrity, we will keep striving to offer free, innovative, and reliable software for stochastic modelling. As a closing remark, we thank our imitators, after all, as Oscar Wilde said: “Imitation is the sincerest form of flattery that mediocrity can pay to greatness”.


# References {#section_6}

1. Papalexiou, S.M. (2018). _Unified theory for stochastic modelling of hydroclimatic processes: Preserving marginal distributions, correlation structures, and intermittency_. Advances in Water Resources, 115, 234-252, [link](https://doi.org/10.1016/j.advwatres.2018.02.013)
2.	Papalexiou, S.M., Markonis, Y., Lombardo, F., AghaKouchak, A., Foufoula-Georgiou, E. (2018). _Precise Temporal Disaggregation Preserving Marginals and Correlations (DiPMaC) for Stationary and Nonstationary Processes_. Water Resources Research, 54(10), 7435-7458, [link](https://doi.org/10.1029/2018WR022726)
3.	Papalexiou, S.M., Serinaldi, F. (2020). _Random Fields Simplified: Preserving Marginal Distributions, Correlations, and Intermittency, With Applications From Rainfall to Humidity_. Water Resources Research, 56(2), e2019WR026331, [link](https://doi.org/10.1029/2019WR026331)
4. Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). _Advancing Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond_. Water Resources Research, 57, e2020WR029466, [link](https://doi.org/10.1029/2020WR029466)
5.	Serinaldi, F., Kilsby, C.G. (2018). _Unsurprising Surprises: The Frequency of Record-breaking and Overthreshold Hydrological Extremes Under Spatial and Temporal Dependence_. Water Resources Research, 54(9), 6460-6487, [link](https://doi.org/10.1029/2018WR023055)


* Papalexiou, S.M. (2010). _Stochastic modelling demystified_. 10.13140/RG.2.2.34889.60008.[link](http://dx.doi.org/10.13140/RG.2.2.34889.60008)
* Papalexiou, S.M. (2017). _A unified theory for exact stochastic modelling of univariate and multivariate processes with continuous, mixed type, or discrete marginal distributions and any correlation structure_. ArXiv:1707.06842 [Math, Stat]. [link](https://arxiv.org/abs/1707.06842)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.acti}
\alias{plot.acti}
\title{Plot method for \code{acti} objects}
\usage{
\method{plot}{acti}(x, ...)
}
\arguments{
\item{x}{an \code{acti} object returned by \code{\link{fitactf}}}

\item{...}{optional arguments; \code{main} sets the plot title}
}
\value{
a \code{ggplot} object (invisibly returned; also printed)
}
\description{
Visualises the autocorrelation transformation function (ACTF) fitted by
\code{\link{fitactf}}.
}
\examples{

library(CoSMoS)

p   <- actpnts(margdist = "paretoII",
               margarg  = list(scale = 1, shape = .3),
               p0 = 0)
fit <- fitactf(p)

plot(fit)
plot(fit, main = "Pareto type II\nautocorrelation transformation")

}
\seealso{
\code{\link{fitactf}}, \code{\link{actpnts}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CoSMoS.R
\docType{package}
\name{CoSMoS-package}
\alias{CoSMoS}
\alias{CoSMoS-package}
\title{CoSMoS: Complete Stochastic Modelling Solution}
\description{
CoSMoS is an R package that makes time series generation with desired
properties easy. Just choose the characteristics of the time series you
want to generate, and it will do the rest.
}
\details{
The generated time series preserve any probability distribution and any
linear autocorrelation structure. Users can generate as many and as long
time series from processes such as precipitation, wind, temperature,
relative humidity etc. It is based on a framework that unified, extended,
and improved a modelling strategy that generates time series by transforming
"parent" Gaussian time series having specific characteristics
(Papalexiou, 2018).
}
\section{Funding}{

The package was partly funded by the Global Institute for Water Security
(GIWS; \href{https://water.usask.ca/}{https://water.usask.ca/}) and the
Global Water Futures (GWF; \href{https://gwf.usask.ca/}{https://gwf.usask.ca/})
program.
}

\references{
Papalexiou, S.M. (2018). Unified theory for stochastic modelling
of hydroclimatic processes: Preserving marginal distributions, correlation
structures, and intermittency. Advances in Water Resources 115, 234-252,
\doi{10.1016/j.advwatres.2018.02.013}

Papalexiou, S.M., Markonis, Y., Lombardo, F., AghaKouchak, A.,
Foufoula-Georgiou, E. (2018). Precise Temporal Disaggregation Preserving
Marginals and Correlations (DiPMaC) for Stationary and Nonstationary
Processes. Water Resources Research, 54(10), 7435-7458,
\doi{10.1029/2018WR022726}

Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
Preserving Marginal Distributions, Correlations, and Intermittency, With
Applications From Rainfall to Humidity. Water Resources Research, 56(2),
e2019WR026331, \doi{10.1029/2019WR026331}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/TycheLab/CoSMoS}
}

}
\author{
\strong{Coded by:} Filip Strnad \email{strnadf@fzp.czu.cz} and
Francesco Serinaldi \email{francesco.serinaldi@ncl.ac.uk}

\strong{Conceptual design by:} Simon Michael Papalexiou
\email{sm.papalexiou@usask.ca}

\strong{Tested and documented by:} Yannis Markonis
\email{markonis@fzp.czu.cz}

\strong{Maintained by:} Kevin Shook
\email{kevin.shook@usask.ca}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{actfInv}
\alias{actfInv}
\title{Inverse ACTF}
\usage{
actfInv(rhoz, b, c)
}
\arguments{
\item{rhoz}{Gaussian correlation value}

\item{b}{1st parameter}

\item{c}{2nd parameter}
}
\value{
numeric vector of back-transformed marginal correlations
}
\description{
Inverse of the autocorrelation transformation function.
}
\examples{
actfInv(.4, 1, 0)

}
\seealso{
\code{\link{actf}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.checkTS}
\alias{plot.checkTS}
\title{Plot method for \code{checkTS} objects}
\usage{
\method{plot}{checkTS}(x, ...)
}
\arguments{
\item{x}{a \code{checkTS} object returned by \code{\link{checkTS}}}

\item{...}{currently unused}
}
\value{
a \code{ggplot} object (invisibly returned; also printed)
}
\description{
Displays boxplots of simulated statistics against theoretical expected values
for each statistic tracked by \code{\link{checkTS}}.
}
\examples{

library(CoSMoS)

x <- generateTS(margdist = "burrXII",
                margarg = list(scale = 1,
                               shape1 = .75,
                               shape2 = .15),
                acsvalue = acs(id = "weibull",
                               t = 0:30,
                               scale = 10,
                               shape = .75),
                n = 1000, p = 30, p0 = .25, TSn = 100)

chck <- checkTS(x)
plot(chck)

}
\seealso{
\code{\link{checkTS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seasonal.R
\name{seasonalACF}
\alias{seasonalACF}
\title{Calculate seasonal autocorrelation function}
\usage{
seasonalACF(TS, season, lag.max = 50)
}
\arguments{
\item{TS}{data frame or data table with columns \code{date} and \code{value}}

\item{season}{character; name of a date-component function
(e.g. \code{"month"})}

\item{lag.max}{integer; maximum lag for the ACF (default 50)}
}
\value{
a named list (one element per season) of numeric ACF vectors
starting at lag 0
}
\description{
Computes the empirical ACF for each season of a time series.
}
\seealso{
\code{\link{analyzeTS}}, \code{\link{fitACS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitting.R
\name{optimACS}
\alias{optimACS}
\title{Auxiliary objective function for \code{fitACS}}
\usage{
optimACS(par, id, eACS, error = "MSE")
}
\arguments{
\item{par}{numeric vector of ACS parameter values}

\item{id}{character; ACS identifier (e.g. \code{"weibull"})}

\item{eACS}{numeric vector of empirical ACS values}

\item{error}{character; error criterion — currently only \code{"MSE"}}
}
\value{
scalar numeric; the MSE between theoretical and empirical ACS
}
\description{
Computes the mean squared error between a theoretical autocorrelation
structure and the empirical ACS; passed as \code{fn} to \code{optim} inside
\code{\link{fitACS}}.
}
\seealso{
\code{\link{fitACS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seasonal.R
\name{stratifySeasonData}
\alias{stratifySeasonData}
\title{Stratify a time series by season}
\usage{
stratifySeasonData(TS, season)
}
\arguments{
\item{TS}{data frame or data table with columns \code{date} and \code{value}}

\item{season}{character; name of a date-component function
(e.g. \code{"month"}, \code{"week"})}
}
\value{
a list of two elements: \code{[[1]]} named list of full seasonal
subsets; \code{[[2]]} named list of nonzero seasonal subsets
}
\description{
Splits a two-column (date, value) time series into a list of seasonal
subsets, plus a parallel list of nonzero-only subsets.
}
\seealso{
\code{\link{analyzeTS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{fitactf}
\alias{fitactf}
\title{Fit the AutoCorrelation Transformation Function}
\usage{
fitactf(actpnts, discrete = FALSE)
}
\arguments{
\item{actpnts}{estimated ACT points (output of \code{\link{actpnts}})}

\item{discrete}{logical — is the marginal distribution discrete?}
}
\value{
An object of class \code{"acti"} with components:
\describe{
\item{actfcoef}{fitted ACTF coefficients \code{b} and \code{c}}
\item{actfpoints}{the input ACT points data frame}
}
}
\description{
Fits the ACTF to the estimated (rho_x, rho_z) points using \code{nls}.
}
\examples{

library(CoSMoS)

p   <- actpnts(margdist = "paretoII",
               margarg  = list(scale = 1, shape = .3),
               p0 = 0)
fit <- fitactf(p)
plot(fit)

}
\seealso{
\code{\link{actpnts}}, \code{\link{actf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitting.R
\name{rMSE}
\alias{rMSE}
\title{Ratio mean squared error}
\usage{
rMSE(x, y)
}
\arguments{
\item{x}{numeric vector of observed values}

\item{y}{numeric vector of reference values}
}
\value{
scalar numeric
}
\description{
Ratio mean squared error
}
\seealso{
\code{\link{MSE}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-fields.R
\name{generateMTS}
\alias{generateMTS}
\title{Simulation of multiple time series with given marginals and spatiotemporal
properties}
\usage{
generateMTS(n, STmodel)
}
\arguments{
\item{n}{number of time steps to simulate}

\item{STmodel}{list of arguments from \code{\link{fitVAR}}}
}
\value{
A matrix of class \code{"matrix"} with attribute \code{STmodel}.
Rows correspond to time steps and columns to spatial locations.
}
\description{
Generates multiple time series with given marginals and spatiotemporal
properties. Provide (1) the output of \code{\link{fitVAR}} and (2) the number
of time steps to simulate.
}
\details{
Referring to the documentation of \code{\link{fitVAR}} for details on
computational complexity, here we report indicative simulation CPU times,
assuming model parameters are already evaluated. CPU times refer to a
Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
4-core, 8 logical processors, and 32 GB RAM. \cr
CPU time: \cr
d = 900, p = 1, n = 1000: ~17s \cr
d = 900, p = 1, n = 10000: ~75s \cr
d = 900, p = 5, n = 100: ~280s \cr
d = 900, p = 5, n = 1000: ~302s \cr
d = 2500, p = 1, n = 1000: ~160s \cr
d = 2500, p = 1, n = 10000: ~570s \cr
where \eqn{d} denotes the number of spatial locations.
}
\examples{
## Simulation of a 4-dimensional vector with VAR(1) correlation structure
coord <- cbind(runif(4) * 30, runif(4) * 30)

fit <- fitVAR(
  spacepoints = coord,
  p = 1,
  margdist = "burrXII",
  margarg = list(scale = 3,
                 shape1 = .9,
                 shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(scfid = "weibull",
                 tcfid = "weibull",
                 copulaarg = 2,
                 scfarg = list(scale = 20,
                               shape = 0.7),
                 tcfarg = list(scale = 1.1,
                               shape = 0.8))
)

sim <- generateMTS(n = 100, STmodel = fit)

}
\seealso{
\code{\link{fitVAR}}, \code{\link{generateRF}}, \code{\link{generateMTSFast}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anisotropyT.R
\name{anisotropyTaffine}
\alias{anisotropyTaffine}
\title{Affine anisotropy transformation}
\usage{
anisotropyTaffine(spacepoints, phi1, phi2, phi12, theta)
}
\arguments{
\item{spacepoints}{vector of coordinates (2 x d), where d is the number of locations/grid points}

\item{phi1}{stretching parameter along the x axis}

\item{phi2}{stretching parameter along the y axis}

\item{phi12}{shear effect}

\item{theta}{rotation angle}
}
\description{
Affine anisotropy transformation.
}
\examples{

## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

at <- anisotropyTaffine(spacepoints = coord,
                       phi1 = 0.5,
                       phi2 = 2,
                       phi12 = 0,
                       theta = -pi/3)

## visualize transformed coordinate system
aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
ggplot(aux, aes(x = lon, y = lat)) +
geom_path(aes(group = id1)) +
geom_path(aes(group = id2)) +
geom_point(col = 2) +
theme_light()

}
\references{
Allard, D., Senoussi, R., Porcu, E. (2016). Anisotropy Models for
Spatial Data. Mathematical Geosciences, 48(3), 305-328, \doi{10.1007/s11004-015-9594-x}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{actpnts_cpp}
\alias{actpnts_cpp}
\title{Fast actpnts via nested C++ 1D integration}
\usage{
actpnts_cpp(
  margdist,
  margarg,
  p0,
  lower,
  upper,
  mu1,
  mu2,
  rhoz_vals,
  max_eval = 10000L,
  abs_tol = 1e-05,
  rel_tol = 1e-05
)
}
\arguments{
\item{margdist}{character distribution name}

\item{margarg}{named numeric vector of distribution parameters}

\item{p0}{probability zero}

\item{lower}{integration lower bound}

\item{upper}{integration upper bound}

\item{mu1}{first raw moment}

\item{mu2}{second central moment}

\item{rhoz_vals}{numeric vector of Gaussian correlation values}

\item{max_eval}{max evaluations per 1D integral (default 1e4)}

\item{abs_tol}{absolute tolerance (default 1e-5)}

\item{rel_tol}{relative tolerance (default 1e-5)}
}
\value{
data.frame with columns rhoz and rhox
}
\description{
Internal C++ implementation. Do not call directly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acs.R
\name{acs}
\alias{acs}
\title{AutoCorrelation Structure}
\usage{
acs(id, ...)
}
\arguments{
\item{id}{autocorrelation structure id.}

\item{...}{other arguments (\code{t} as lag and ACS parameters).}
}
\value{
A numeric vector of autocorrelation values at the supplied lags.
}
\description{
Provides a parametric function that describes the values of the linear
autocorrelation up to desired lags. For more details on the parametric
autocorrelation structures see section 3.2 in Papalexiou (2018).
}
\examples{

library(CoSMoS)

## specify lag
t <- 0:10

## get the ACS
f <- acs("fgn",     t = t, H = .75)
b <- acs("burrXII", t = t, scale = 1, shape1 = .6, shape2 = .4)
w <- acs("weibull", t = t, scale = 2, shape = 0.8)
p <- acs("paretoII", t = t, scale = 3, shape = 0.3)

## visualize the ACS
dta   <- data.table(t, f, b, w, p)
m.dta <- melt(dta, id.vars = "t")

ggplot(m.dta,
       aes(x      = t,
           y      = value,
           group  = variable,
           colour = variable)) +
  geom_point(size = 2.5) +
  geom_line(lwd = 1) +
  scale_color_manual(values = c("steelblue4", "red4", "green4", "darkorange"),
                     labels = c("FGN", "Burr XII", "Weibull", "Pareto II"),
                     name   = "") +
  labs(x = bquote(lag ~ tau),
       y = "Acf") +
  scale_x_continuous(breaks = t) +
  theme_classic()
}
\references{
Papalexiou, S.M. (2018). Unified theory for stochastic
modelling of hydroclimatic processes: Preserving marginal distributions,
correlation structures, and intermittency. Advances in Water Resources,
115, 234-252, \doi{10.1016/j.advwatres.2018.02.013}
}
\seealso{
\code{\link{actpnts}}, \code{\link{fitACS}}, \code{\link{fitactf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitting.R
\name{N}
\alias{N}
\title{Norm minimised during distribution fitting}
\usage{
N(par, val, dist, norm, n.points)
}
\arguments{
\item{par}{numeric vector of distribution parameter values}

\item{val}{numeric vector of empirical data}

\item{dist}{character; name of the distribution to fit (e.g. \code{"norm"},
\code{"ggamma"})}

\item{norm}{character; norm identifier — one of \code{"N1"}, \code{"N2"},
\code{"N3"}, \code{"N4"}}

\item{n.points}{integer; number of ECDF points used in the norm computation}
}
\value{
scalar numeric; the norm value (returns \code{10e9} if non-finite)
}
\description{
Computes one of four fitting norms used as the objective function in
\code{\link{fitDist}}.
}
\seealso{
\code{\link{fitDist}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advectionF.R
\name{advectionFrotation}
\alias{advectionFrotation}
\title{Rotational advection field}
\usage{
advectionFrotation(spacepoints, x0, y0, a, b)
}
\arguments{
\item{spacepoints}{vector of coordinates (2 x d), where d is the number of locations/grid points}

\item{x0}{x coordinate of the center of rotation}

\item{y0}{y coordinate of the center of rotation}

\item{a}{parameter controlling the x component of rotational velocity}

\item{b}{parameter controlling the y component of rotational velocity}
}
\description{
Provides an advection field corresponding to rotation around a specified center.
}
\section{Note}{
\itemize{
\item if a > 0, b > 0: clockwise rotation around (x0, y0)
\item if a < 0, b < 0: counter-clockwise rotation around (x0, y0)
}
}

\examples{

library(ggquiver)
library(ggplot2)
## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

af <- advectionFrotation(spacepoints = coord,
                        x0 = floor(m / 2),
                        y0 = floor(m / 2),
                        a = 3,
                        b = 2)

## visualize advection field
dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
geom_quiver() +
theme_light()

}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{quickTSPlot}
\alias{quickTSPlot}
\title{Quick visualisation of basic time series properties}
\usage{
quickTSPlot(TS, ci = 0.95)
}
\arguments{
\item{TS}{numeric vector (or \code{data.frame}/\code{data.table}) of time
series values}

\item{ci}{numeric; confidence level for the zero-autocorrelation band
(default \code{0.95})}
}
\value{
a \code{ggdraw} object (printed as a side effect)
}
\description{
Returns a composite figure showing the time series, empirical density
function, and empirical autocorrelation function.
}
\examples{
ggamma_sim <- rggamma(n = 1000, scale = 1, shape1 = 1, shape2 = .5)
quickTSPlot(ggamma_sim)

}
\seealso{
\code{\link{generateTS}}, \code{\link{plot.cosmosts}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{GEV}
\alias{GEV}
\alias{dgev}
\alias{pgev}
\alias{qgev}
\alias{rgev}
\alias{mgev}
\title{Generalized Extreme Value distribution}
\usage{
dgev(x, loc, scale, shape, log = FALSE)

pgev(q, loc, scale, shape, lower.tail = TRUE, log.p = FALSE)

qgev(p, loc, scale, shape, lower.tail = TRUE, log.p = FALSE)

rgev(n, loc, scale, shape)

mgev(r, loc, scale, shape)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{loc, scale, shape}{location, scale, and shape parameters.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities \code{p} are
given as \code{log(p)}.}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
\eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.}

\item{p}{vector of probabilities.}

\item{n}{number of observations. If \code{length(n) > 1}, the length
is taken to be the number required.}

\item{r}{raw moment order.}
}
\value{
\code{dgev} returns a numeric vector of density values.
\code{pgev} returns a numeric vector of cumulative probabilities.
\code{qgev} returns a numeric vector of quantiles.
\code{rgev} returns a numeric vector of random deviates.
\code{mgev} returns the raw moment of order \code{r} (via numerical
integration).
}
\description{
Provides density, distribution function, quantile function, random value
generation, and raw moments of order \emph{r} for the generalized extreme
value distribution.
}
\examples{

## plot the density

ggplot(data.frame(x = c(0, 20)),
       aes(x)) +
  stat_function(fun = dgev,
                args = list(loc = 1,
                            scale = .5,
                            shape = .15),
                colour = "royalblue4") +
  labs(x = "",
       y = "Density") +
  theme_classic()
}
\seealso{
\code{\link{fitDist}}, \code{\link{moments}}
}
\concept{Continuous}
\concept{Univariate}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precip.R
\docType{data}
\name{precip}
\alias{precip}
\title{Hourly station precipitation data}
\format{
A data.table with 79633 rows and 2 variables:
\describe{
\item{date}{POSIXct format date/time}
\item{value}{precipitation totals}
}
}
\source{
The National Oceanic and Atmospheric Administration (NOAA)
}
\usage{
precip
}
\description{
Station details
\itemize{
\item Name: Philadelphia International Airport
\item Network ID: COOP:366889
\item Latitude/Longitude:	39.87327°, -75.22678°
\item Elevation: 3m
}
}
\details{
more details can be found \href{https://www.ncdc.noaa.gov/cdo-web/datasets/PRECIP_HLY/stations/COOP:366889/detail}{here}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advectionF.R
\name{advectionFspiralCE}
\alias{advectionFspiralCE}
\title{Spiraling advection field satisfying continuity equation}
\usage{
advectionFspiralCE(spacepoints, a, C)
}
\arguments{
\item{spacepoints}{vector of coordinates (2 x d), where d is the number of locations/grid points}

\item{a}{parameter controlling the intensity of rotational velocity (a > 0 clokwise; a < 0 conter-clockwise)}

\item{C}{parameter ranging in (0, 2*pi)}
}
\description{
Provides an advection field corresponding to a spiral motion to/from a specified reference point (sink) satisfying continuity equation (from \href{https://github.com/johannesgerer/jburkardt-f/tree/2a3860f296af9ab0a07c6dc805b496bc75ad3935/spiral_data}{Git Mirror of John Burkardt's collection of FORTRAN 90 Software}).
}
\examples{
library(ggquiver)
library(ggplot2)
## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

af <- advectionFspiralCE(spacepoints = coord,
                        a = 5,
                        C = 1)

## visualize advection field
dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
geom_quiver() +
theme_light()

}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.fitACS}
\alias{plot.fitACS}
\title{Plot method for \code{fitACS} objects}
\usage{
\method{plot}{fitACS}(x, ...)
}
\arguments{
\item{x}{a \code{fitACS} object returned by \code{\link{fitACS}}}

\item{...}{currently unused}
}
\value{
a \code{ggplot} object (invisibly returned; also printed)
}
\description{
Displays the empirical ACF alongside the fitted theoretical autocorrelation
structure.
}
\examples{

x <- arima.sim(model = list(ar = 0.8), n = 1000)
acsfit <- fitACS(acf(x, plot = FALSE)$acf, "weibull", c(1, 1))
plot(acsfit)

}
\seealso{
\code{\link{fitACS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkRF.r
\name{checkRF}
\alias{checkRF}
\title{Numerical and visual check of generated random fields}
\usage{
checkRF(RF, lags = 30, nfields = 49, method = "stat")
}
\arguments{
\item{RF}{output of \code{\link{generateRF}}}

\item{lags}{number of lags of empirical STCF to be considered in the
graphical output (default set to 30)}

\item{nfields}{number of fields to be used in the numerical and graphical
output (default set to 49). As the plots are arranged in a matrix with nrows as close as possible to ncol, we suggest using values such as 3x3, 3x4, 7x8, etc.}

\item{method}{report method - \code{"stat"} for basic statistical report,
\code{"statplot"} for graphical check of lagged SCS, target STCS, and marginal
distribution, \code{"field"} for plotting a matrix of the first \code{nfields},
and \code{"movie"} to save the first \code{nfields} as a GIF file named "movieRF.gif"
in the current working directory}
}
\description{
Compares generated random fields sample statistics with the theoretically
expected values (similar to \code{\link{checkTS}}). It also returns graphical output for
visual check.
}
\examples{
## The example below refers to the fitting and simulation of 10 random fields
## of size 10x10 with AR(1) temporal correlation. As the fitting algorithm has
## O((mxm)^3) complexity for a mxm field, this setting allows for quick fitting
## and simulation (short CPU time). However, for a more effective visualization
## and reliable performance assessment, we suggest to generate a larger number
## of fields (e.g. 100 or more) of size about 30X30. This setting needs more
## CPU time but enables more effective comparison of theoretical and
## empirical statistics. Sizes larger than about 50x50 can be unpractical
##  on standard machines.

fit <- fitVAR(
  spacepoints = 10,
  p = 1,
  margdist ="burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(scfid = "weibull", tcfid = "weibull",
                 copulaarg = 2,
                 scfarg = list(scale = 20, shape = 0.7),
                tcfarg = list(scale = 1.1, shape = 0.8))
)

sim <- generateRF(n = 12,
                    STmodel = fit)
checkRF(RF = sim,
          lags = 10,
          nfields = 12)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecdf.R, R/fitting.R
\name{ECDF}
\alias{ECDF}
\title{Empirical cumulative distribution function}
\usage{
ECDF(x)

ECDF(x)
}
\arguments{
\item{x}{numeric vector of values}
}
\value{
A \code{data.table} with columns \code{p} (plotting positions)
and \code{value} (sorted unique values).

a \code{data.table} with columns \code{p} (plotting positions) and
\code{value} (sorted unique values), keyed on \code{value}
}
\description{
Calculates the ECDF using the Weibull plotting position formula.

Computes the ECDF using the Weibull plotting position.
}
\seealso{
\code{\link{fitDist}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anisotropyT.R
\name{anisotropyTswirl}
\alias{anisotropyTswirl}
\title{Swirl anisotropy transformation}
\usage{
anisotropyTswirl(spacepoints, x0, y0, b, alpha)
}
\arguments{
\item{spacepoints}{vector of coordinates (2 x d), where d is the number of locations/grid points}

\item{x0}{x coordinate of the center of the swirl deformation}

\item{y0}{y coordinate of the center of the swirl deformation}

\item{b}{scaling parameter controlling the swirl deformation}

\item{alpha}{rotation angle}
}
\description{
Swirl anisotropy transformation.
}
\examples{

## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

at <- anisotropyTswirl(spacepoints = coord,
                      x0 = floor(m / 2),
                      y0 = floor(m / 2),
                      b = 10,
                      alpha = 1.5 * pi)

## visualize transformed coordinate system
aux = data.frame(lon = at[ ,1], lat = at[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
ggplot(aux, aes(x = lon, y = lat)) +
geom_path(aes(group = id1)) +
geom_path(aes(group = id2)) +
geom_point(col = 2) +
theme_light()

}
\references{
Ligas, M., Banas, M., Szafarczyk, A. (2019). A method for local
approximation of a planar deformation field. Reports on Geodesy and
Geoinformatics, 108(1), 1-8, \doi{10.2478/rgg-2019-0007}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ar-models.R
\name{YW}
\alias{YW}
\title{Yule-Walker solver}
\usage{
YW(ACS)
}
\arguments{
\item{ACS}{Numeric vector. ACS values starting from lag 0
(\code{ACS[1] = 1}). The AR order is \code{length(ACS) - 1}.}
}
\value{
Numeric matrix of AR coefficients, shape \code{(p, 1)}.
}
\description{
Solves the Yule-Walker equations to compute AR(\eqn{p}) coefficients from
an autocorrelation structure (ACS) vector. Used internally by
\code{\link{ARp}} and \code{\link{seasonalAR}}.
}
\seealso{
\code{\link{ARp}}, \code{\link{seasonalAR}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advectionF.R
\name{advectionF2}
\alias{advectionF2}
\title{Advection fields}
\usage{
advectionF2(id, arglist)
}
\arguments{
\item{id}{advection type id}

\item{arglist}{list of additional arguments (vector of coordinates and parameters of advection field functions)}
}
\description{
Provides parametric functions that describe different types of advection fields.
}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-ts.R
\name{generateTS}
\alias{generateTS}
\title{Generate time series}
\usage{
generateTS(
  n,
  margdist,
  margarg,
  p = NULL,
  p0 = 0,
  TSn = 1,
  distbounds = c(-Inf, Inf),
  acsvalue = NULL
)
}
\arguments{
\item{n}{Positive integer. Length of the generated time series.}

\item{margdist}{target marginal distribution}

\item{margarg}{list of marginal distribution arguments}

\item{p}{Positive integer or \code{NULL}. AR model order. When
\code{NULL} (default), the order is chosen automatically as the number of
lags where the transformed ACS exceeds 0.01, capped at 1000.}

\item{p0}{probability zero}

\item{TSn}{number of time series to generate}

\item{distbounds}{numeric vector of length 2; distribution bounds (default
\code{c(-Inf, Inf)})}

\item{acsvalue}{Numeric vector. Target autocorrelation structure starting
from lag 0 (i.e. \code{acsvalue[1] = 1}).}
}
\value{
An object of class \code{'cosmosts'}: a list of \code{TSn} numeric
vectors, each of length \code{n}, with per-series attributes recording the
fitted model parameters.
}
\description{
Generates time series with given properties. Provide (1) the target marginal
distribution and its parameters, (2) the target autocorrelation structure or
individual autocorrelation values up to a desired lag, and (3) the probability
zero if you wish to simulate an intermittent process.
}
\details{
A step-by-step guide:
\itemize{
\item First define the target marginal (\code{margdist}), that is, the probability
distribution of the generated data. For example set \code{margdist = 'ggamma'}
for the Generalised Gamma distribution, \code{margdist = 'burrXII'} for Burr
type XII etc. For a full list of supported distributions see the help
\href{https://CRAN.R-project.org/package=CoSMoS/vignettes/vignette.html}{vignette}.
In general, the package supports all built-in distribution functions of R and
of other packages.
\item Define the parameters (\code{margarg}) of the selected distribution.
For example the Generalised Gamma has one scale and two shape parameters,
e.g. \code{margarg = list(scale = 2, shape1 = 0.9, shape2 = 0.8)}.
See the help vignette for details on each distribution's parameters.
\item If you wish your time series to be intermittent (e.g. precipitation), define
the probability zero. For example \code{p0 = 0.9} produces 90\\% zero values.
\item Define your linear autocorrelations.
\itemize{
\item Supply specific lag autocorrelations starting from lag 0 up to a desired
lag, e.g. \code{acsvalue = c(1, 0.9, 0.8, 0.7)}; this preserves lag-1,
lag-2 and lag-3 autocorrelations equal to 0.9, 0.8 and 0.7.
\item Alternatively, use a parametric autocorrelation structure (see section 3.2
in Papalexiou (2018)). Supported structures: \code{weibull}, \code{paretoII},
\code{fgn} and \code{burrXII}. See also \code{\link[CoSMoS]{acs}}.
}
\item Define the AR model order \code{p}. For example if you aim to preserve the
first 10 lag autocorrelations then set \code{p = 10}. Set \code{p = NULL} and
the model will choose \code{p} to preserve the whole autocorrelation structure.
\item Set the time series length, e.g. \code{n = 1000}, and the number of time
series to generate, e.g. \code{TSn = 10}.
}
}
\examples{

library(CoSMoS)

## Case 1:
## Generate 3 time series of length 1000 following the Generalised Gamma
## distribution with scale = 1, shape1 = 0.8, shape2 = 0.8 and ParetoII
## autocorrelation structure with scale = 1 and shape = 0.75.
x <- generateTS(margdist = "ggamma",
                margarg = list(scale = 1,
                               shape1 = .8,
                               shape2 = .8),
                acsvalue = acs(id = "paretoII",
                               t = 0:30,
                               scale = 1,
                               shape = .75),
                n = 1000,
                p = 30,
                TSn = 3)

## see the results
plot(x)

\donttest{

## Case 2:
## Same as Case 1 but intermittent with probability zero equal to 90\%.
y <- generateTS(margdist = "ggamma",
                margarg = list(scale = 1,
                               shape1 = .8,
                               shape2 = .8),
                acsvalue = acs(id = "paretoII",
                               t = 0:30,
                               scale = 1,
                               shape = .75),
                p0 = .9,
                n = 1000,
                p = 30,
                TSn = 3)

## see the results
plot(y)

## Case 3:
## Generate a time series of length 1000 following the Beta distribution
## (e.g. relative humidity in [0, 1]) with shape1 = 0.6, shape2 = 0.8
## and ParetoII autocorrelation structure.
z <- generateTS(margdist = "beta",
                margarg = list(shape1 = .6,
                               shape2 = .8),
                distbounds = c(0, 1),
                acsvalue = acs(id = "paretoII",
                               t = 0:30,
                               scale = 1,
                               shape = .75),
                n = 1000,
                p = 20)

## see the results
plot(z)

## Case 4:
## Same as Case 3 but providing specific autocorrelation values for the
## first three lags (lag 1 to 3 equal to 0.9, 0.8, 0.7).
z <- generateTS(margdist = "beta",
                margarg = list(shape1 = .6,
                               shape2 = .8),
                distbounds = c(0, 1),
                acsvalue = c(1, .9, .8, .7),
                n = 1000,
                p = NULL)

## see the results
plot(z)

}

}
\references{
Papalexiou, S.M. (2018). Unified theory for stochastic modelling of
hydroclimatic processes: Preserving marginal distributions, correlation structures,
and intermittency. Advances in Water Resources, 115, 234-252,
\doi{10.1016/j.advwatres.2018.02.013}
}
\seealso{
\code{\link{regenerateTS}}, \code{\link{ARp}}, \code{\link{actpnts}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{ParetoII}
\alias{ParetoII}
\alias{dparetoII}
\alias{pparetoII}
\alias{qparetoII}
\alias{rparetoII}
\alias{mparetoII}
\title{Pareto Type II distribution}
\usage{
dparetoII(x, scale, shape, log = FALSE)

pparetoII(q, scale, shape, lower.tail = TRUE, log.p = FALSE)

qparetoII(p, scale, shape, lower.tail = TRUE, log.p = FALSE)

rparetoII(n, scale, shape)

mparetoII(r, scale, shape)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{scale, shape}{scale and shape parameters; the shape argument cannot
be a vector (must have length one).}

\item{log, log.p}{logical; if \code{TRUE}, probabilities \code{p} are
given as \code{log(p)}.}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
\eqn{P[X \le x]}, otherwise \eqn{P[X > x]}.}

\item{p}{vector of probabilities.}

\item{n}{number of observations. If \code{length(n) > 1}, the length
is taken to be the number required.}

\item{r}{raw moment order.}
}
\value{
\code{dparetoII} returns a numeric vector of density values.
\code{pparetoII} returns a numeric vector of cumulative probabilities.
\code{qparetoII} returns a numeric vector of quantiles.
\code{rparetoII} returns a numeric vector of random deviates.
\code{mparetoII} returns the raw moment of order \code{r}.
}
\description{
Provides density, distribution function, quantile function, random value
generation, and raw moments of order \emph{r} for the Pareto type II
distribution.
}
\examples{

## plot the density

ggplot(data.frame(x = c(0, 20)),
       aes(x)) +
  stat_function(fun = dparetoII,
                args = list(scale = 1,
                            shape = .3),
                colour = "royalblue4") +
  labs(x = "",
       y = "Density") +
  theme_classic()
}
\seealso{
\code{\link{fitDist}}, \code{\link{moments}}
}
\concept{Continuous}
\concept{Univariate}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stcs.r
\name{stcs2}
\alias{stcs2}
\title{SpatioTemporal Correlation Structure}
\usage{
stcs2(id, arglist)
}
\arguments{
\item{id}{spatiotemporal correlation structure ID}

\item{arglist}{list of additional arguments (t as time lag, s as spatial lag (distance),  and stcs parameters)}
}
\description{
Provides a parametric function that describes the values of the linear
spatiotemporal autocorrelation up to desired lags.
For more details on the parametric spatiotemporal correlation structures see
section 2.3 and 2.4 in Papalexiou and Serinaldi (2020).
}
\references{
Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
Preserving Marginal Distributions, Correlations, and Intermittency,
With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
e2019WR026331, \doi{10.1029/2019WR026331}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moments.R
\name{moments}
\alias{moments}
\title{Numerical estimation of moments}
\usage{
moments(
  dist,
  distarg,
  p0 = 0,
  raw = TRUE,
  central = TRUE,
  coef = TRUE,
  distbounds = c(-Inf, Inf),
  order = 1:4
)
}
\arguments{
\item{dist}{character; distribution name (e.g. \code{"norm"},
\code{"paretoII"})}

\item{distarg}{list of distribution arguments}

\item{p0}{numeric; probability zero (default 0)}

\item{raw}{logical; compute raw moments?}

\item{central}{logical; compute central moments?}

\item{coef}{logical; compute standardised coefficients (CV, skewness,
kurtosis)?}

\item{distbounds}{numeric vector of length 2; distribution bounds (default
\code{c(-Inf, Inf)})}

\item{order}{integer vector; raw moment orders (default \code{1:4})}
}
\value{
a named list with zero or more of:
\describe{
\item{\code{m}}{raw moments}
\item{\code{mu}}{central moments}
\item{\code{coefficients}}{CV, skewness, kurtosis}
}
}
\description{
Uses numerical integration to compute the theoretical raw or central moments
of the specified distribution.
}
\examples{

library(CoSMoS)

## Normal distribution
moments("norm", list(mean = 2, sd = 1))

## Pareto type II
moments(dist    = "paretoII",
        distarg = list(shape = 0.2, scale = 1))

}
\seealso{
\code{\link{sample.moments}}, \code{\link{populationstat}}
}
\keyword{moments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getArg.R
\name{getDistArg}
\alias{getDistArg}
\title{Get argument names of a distribution function}
\usage{
getDistArg(dist)
}
\arguments{
\item{dist}{distribution name (character), e.g. \code{"paretoII"}.}
}
\value{
A character vector of argument names, or \code{NULL} (with a
message) if the distribution is not defined.
}
\description{
Returns the non-standard argument names of the quantile function
\code{q<dist>} (i.e. everything except \code{p}, \code{lower.tail},
and \code{log.p}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getArg.R
\name{getACSArg}
\alias{getACSArg}
\title{Get argument names of an ACS function}
\usage{
getACSArg(id)
}
\arguments{
\item{id}{ACS id (character), e.g. \code{"weibull"}.}
}
\value{
A character vector of parameter names, or \code{NULL} (with a
message) if the ACS is not defined.
}
\description{
Returns the non-\code{t} argument names of the ACS function
\code{acf<id>} (i.e. the parameter names only).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{actpnts}
\alias{actpnts}
\title{AutoCorrelation Transformed Points}
\usage{
actpnts(margdist, margarg, p0 = 0, distbounds = c(-Inf, Inf))
}
\arguments{
\item{margdist}{target marginal distribution}

\item{margarg}{list of marginal distribution arguments}

\item{p0}{probability zero}

\item{distbounds}{numeric vector of length 2; distribution bounds (default
\code{c(-Inf, Inf)})}
}
\value{
A data frame with columns \code{rhoz} (Gaussian correlations) and
\code{rhox} (corresponding target marginal correlations).
}
\description{
Evaluates the (rho_x, rho_z) mapping between the target marginal
autocorrelation and the underlying Gaussian autocorrelation, using a
double numerical integral.
}
\details{
When the package is compiled with Rcpp support (i.e., \code{actpnts_cpp}
is available), the double integral is evaluated in C++ via the Cubature
algorithm, which is substantially faster than the nested base-R
\code{integrate()} fallback. The C++ path supports the following
distributions natively: \code{ggamma}, \code{paretoII}, \code{burrXII},
\code{burrIII}, \code{gev}, \code{norm}, \code{beta}, \code{gamma},
\code{exp}, \code{weibull}, \code{lnorm}, \code{unif}. Any other
distribution falls back to the R quantile function automatically, so
correctness is always preserved.
}
\examples{

library(CoSMoS)

## Pareto type II marginal
x <- actpnts(margdist = "paretoII",
             margarg  = list(scale = 1, shape = .3),
             p0 = 0)
x

}
\references{
Papalexiou, S.M. (2018). Unified theory for stochastic modelling
of hydroclimatic processes: Preserving marginal distributions, correlation
structures, and intermittency. Advances in Water Resources, 115, 234-252,
\doi{10.1016/j.advwatres.2018.02.013}
}
\seealso{
\code{\link{fitactf}}, \code{\link{acti}}, \code{\link{generateTS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advectionF.R
\name{advectionF}
\alias{advectionF}
\title{Advection fields}
\usage{
advectionF(id, ...)
}
\arguments{
\item{id}{advection type id (\code{uniform}, \code{rotation}, \code{spiral}, \code{spiralCE}, \code{radial}, and \code{hyperbolic})}

\item{...}{other arguments (vector of coordinates and parameters of advection field functions)}
}
\description{
Provides parametric functions that describe different types of advection fields.
}
\examples{

library(ggquiver)
library(ggplot2)

## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

## get the advection field
af <- advectionF("spiral",
                 spacepoints = coord,
                 x0 = floor(m / 2),
                 y0 = floor(m / 2),
                 a = 3,
                 b = 2,
                 rotation = 1)

## visualize advection field
dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
geom_quiver() +
theme_light()

}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-fields.R
\name{generateRF}
\alias{generateRF}
\title{Simulation of random fields with given marginals and spatiotemporal properties}
\usage{
generateRF(n, STmodel)
}
\arguments{
\item{n}{number of fields (time steps) to simulate}

\item{STmodel}{list of arguments from \code{\link{fitVAR}}}
}
\value{
A matrix of class \code{"matrix"} with attribute \code{STmodel}.
Rows correspond to spatial locations and columns to time steps.
}
\description{
Generates a random field with given marginals and spatiotemporal properties.
Provide (1) the output of \code{\link{fitVAR}} and (2) the number of time
steps to simulate.
}
\details{
Referring to the documentation of \code{\link{fitVAR}} for details on
computational complexity, here we report indicative simulation CPU times,
assuming model parameters are already evaluated. CPU times refer to a
Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
4-core, 8 logical processors, and 32 GB RAM. \cr
CPU time: \cr
m = 30, p = 1, n = 1000: ~17s \cr
m = 30, p = 1, n = 10000: ~75s \cr
m = 30, p = 5, n = 100: ~280s \cr
m = 30, p = 5, n = 1000: ~302s \cr
m = 50, p = 1, n = 1000: ~160s \cr
m = 50, p = 1, n = 10000: ~570s \cr
where m denotes the side length of a square field (m x m).
}
\examples{
## The example below simulates few random fields of size 10x10 with AR(1)
## temporal correlation for illustration. For reliable performance assessment
## generate a larger number of fields (e.g. 100 or more) of size ~30x30.
## See 'Details' for running times with different settings.

fit <- fitVAR(
  spacepoints = 10,
  p = 1,
  margdist = "burrXII",
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(scfid = "weibull", tcfid = "weibull",
                 copulaarg = 2,
                 scfarg = list(scale = 20, shape = 0.7),
                 tcfarg = list(scale = 1.1, shape = 0.8))
)

sim <- generateRF(n = 12, STmodel = fit)
checkRF(sim, lags = 10, nfields = 12)

}
\seealso{
\code{\link{fitVAR}}, \code{\link{checkRF}}, \code{\link{generateMTS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stcs.r
\name{stcfclayton}
\alias{stcfclayton}
\title{Clayton SpatioTemporal Correlation Structure}
\usage{
stcfclayton(t, s, scfid, tcfid, copulaarg, scfarg, tcfarg)
}
\arguments{
\item{t}{time lag}

\item{s}{spatial lag (distance)}

\item{scfid}{ID of the spatial (marginal) correlation structure (e.g. weibull)}

\item{tcfid}{ID of the temporal (marginal) correlation structure (e.g. weibull)}

\item{copulaarg}{parameter of the Clayton copula linking the marginal correlation structures}

\item{scfarg}{parameters of spatial (marginal) correlation structure}

\item{tcfarg}{parameters of temporal (marginal) correlation structure}
}
\description{
Provides spatiotemporal correlation structure function based on Clayton copula.
For more details on the parametric spatiotemporal correlation structures
see section 2.3 and 2.4 in Papalexiou and Serinaldi (2020).
}
\examples{

library(plot3D)

## specify grid of spatial and temporal lags
d <- 31
st <- expand.grid(0:(d - 1),
                  0:(d - 1))

## get the STCS
wc <- stcfclayton(t = st[, 1],
                  s = st[, 2],
                  scfid = "weibull",
                  tcfid = "weibull",
                  copulaarg = 2,
                  scfarg = list(scale = 20,
                                shape = 0.7),
                  tcfarg = list(scale = 1.1,
                                shape = 0.8))

## visualize the STCS
wc.m <- matrix(wc,
               nrow = d)

persp3D(z = wc.m, x = 1: nrow(wc.m), y = 1:ncol(wc.m),
        expand = 1, main = "", scale = TRUE, facets = TRUE,
        xlab="Time lag", ylab = "Distance", zlab = "STCF",
        colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
        resfac = 5,  col= gg2.col(100))

}
\references{
Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
Preserving Marginal Distributions, Correlations, and Intermittency,
With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
e2019WR026331, \doi{10.1029/2019WR026331}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/population.R
\name{PopulationStat}
\alias{PopulationStat}
\alias{populationstat}
\alias{popmean}
\alias{popsd}
\alias{popvar}
\alias{popcvar}
\alias{popskew}
\alias{popkurt}
\title{Population statistics}
\usage{
populationstat(stat = "mean", dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popmean(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popsd(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popvar(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popcvar(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popskew(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))

popkurt(dist, distarg, p0 = 0, distbounds = c(-Inf, Inf))
}
\arguments{
\item{stat}{character; statistic to compute — one of \code{"mean"},
\code{"sd"}, \code{"var"}, \code{"cvar"}, \code{"skew"}, \code{"kurt"}}

\item{dist}{character; distribution name (e.g. \code{"norm"},
\code{"paretoII"})}

\item{distarg}{list of distribution arguments}

\item{p0}{numeric; probability zero (default 0)}

\item{distbounds}{numeric vector of length 2; distribution bounds (default
\code{c(-Inf, Inf)})}
}
\value{
scalar numeric
}
\description{
Computes theoretical descriptive statistics of a distribution.
}
\examples{

library(CoSMoS)

populationstat("mean", "norm", list(mean = 2, sd = 1))
populationstat("sd",   "norm", list(mean = 2, sd = 1))
populationstat("var",  "norm", list(mean = 2, sd = 1))
populationstat("cvar", "norm", list(mean = 2, sd = 1))
populationstat("skew", "norm", list(mean = 2, sd = 1))
populationstat("kurt", "norm", list(mean = 2, sd = 1))

}
\seealso{
\code{\link{moments}}, \code{\link{checkTS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-internal.R
\name{DHMgenSim}
\alias{DHMgenSim}
\title{Simulate one realisation using DHM circulant-embedding}
\usage{
DHMgenSim(Sj, rn = NULL)
}
\arguments{
\item{Sj}{numeric vector of spectral values from \code{DHMgenSj}}

\item{rn}{optional vector of standard normal random numbers of length
\code{2 * length(Sj) - 2}; generated internally if \code{NULL}}
}
\value{
numeric vector of length \code{length(Sj) - 1}
}
\description{
Simulates one realisation of a stationary Gaussian process from precomputed
spectral values \eqn{S_j} via the Dietrich-Newsam-Haskard algorithm.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-ts.R
\name{regenerateTS}
\alias{regenerateTS}
\title{Bulk time series generation}
\usage{
regenerateTS(ts, TSn = 1)
}
\arguments{
\item{ts}{a \code{cosmosts} object returned by \code{\link{generateTS}}}

\item{TSn}{number of time series to generate}
}
\value{
An object of class \code{'cosmosts'}: a list of \code{TSn} numeric
vectors of the same length as those in \code{ts}.
}
\description{
Generates additional time series using parameters already fitted by
\code{\link{generateTS}}, avoiding recomputation of the ACTF.
}
\details{
After calling \code{\link{generateTS}}, use \code{regenerateTS} to generate
more time series with the same fitted parameters. This is faster than
re-running \code{\link{generateTS}} because the ACTF fitting step is skipped.
}
\examples{

library(CoSMoS)

## Fit once
x <- generateTS(margdist = "burrXII",
                margarg = list(scale = 1,
                               shape1 = .75,
                               shape2 = .25),
                acsvalue = acs(id = "weibull",
                               t = 0:30,
                               scale = 10,
                               shape = .75),
                n = 1000, p = 30, p0 = .5, TSn = 3)

## Generate more realisations with the same parameters
r <- regenerateTS(x)

plot(r)

}
\seealso{
\code{\link{generateTS}}, \code{\link{ARp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anisotropyT.R
\name{anisotropyT}
\alias{anisotropyT}
\title{Anisotropy transformation}
\usage{
anisotropyT(id, ...)
}
\arguments{
\item{id}{anisotropy type id (\code{affine}, \code{swirl}, and \code{wave})}

\item{...}{additional arguments (vector of coordinates and parameters of the anisotropy transformations)}
}
\description{
Provides parametric functions that describe different types of planar deformation fields, including affine (rotation and stretching), and swirl-like deformation. For more details see Papalexiou et al.(2021) and references therein.
}
\examples{

library(CoSMoS)

## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

## get the anisotropy field
at1 <- anisotropyT("affine",
                 spacepoints = coord,
                 phi1 = 0.5,
                 phi2 = 2,
                 phi12 = 0,
                 theta = -pi/3)
at2 <- anisotropyT("swirl",
                 spacepoints = coord,
                 x0 = floor(m / 2),
                 y0 = floor(m / 2),
                 b = 10,
                 alpha = 1.5 * pi)
at3 <- anisotropyT("wave",
                 spacepoints = coord,
                 phi1 = 0.5,
                 phi2 = 2,
                 beta = 3,
                 theta = 0)

## visualize anisotropy field
aux = data.frame(lon = at2[ ,1], lat = at2[ ,2], id1 = rep(1:m, each = m), id2 = rep(1:m, m))
ggplot(aux, aes(x = lon, y = lat)) +
geom_path(aes(group = id1)) +
geom_path(aes(group = id2)) +
geom_point(col = 2) +
theme_light()

}
\references{
Papalexiou, S. M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond,
Water Resources Research, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.cosmosts}
\alias{plot.cosmosts}
\title{Plot method for \code{cosmosts} objects}
\usage{
\method{plot}{cosmosts}(x, ...)
}
\arguments{
\item{x}{a \code{cosmosts} object returned by \code{\link{generateTS}}}

\item{...}{currently unused}
}
\value{
a \code{ggplot} object (invisibly returned; also printed)
}
\description{
Visualises time series generated by \code{\link{generateTS}} as bar charts,
one panel per series.
}
\examples{

library(CoSMoS)

ts <- generateTS(margdist = "ggamma",
                 margarg  = list(scale = 1, shape1 = .8, shape2 = .8),
                 acsvalue = acs(id = "paretoII", t = 0:30,
                                scale = 1, shape = .75),
                 n = 1000, p = 30, TSn = 2)
plot(ts)

}
\seealso{
\code{\link{generateTS}}, \code{\link{regenerateTS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-internal.R
\name{lmom}
\alias{lmom}
\title{L-moments}
\usage{
lmom(x)
}
\arguments{
\item{x}{numeric vector of values}
}
\value{
named numeric vector of length 4: \code{l1}, \code{l2},
\code{l3} (L-skewness ratio), \code{l4} (L-kurtosis ratio)
}
\description{
Computes the first four L-moments of a sample using probability-weighted
moments.
}
\seealso{
\code{\link{reportTS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate-fields.R
\name{generateMTSFast}
\alias{generateMTSFast}
\title{Faster simulation of multiple time series with approximately separable
spatiotemporal correlation structure}
\usage{
generateMTSFast(
  n,
  spacepoints,
  margdist,
  margarg,
  p0,
  distbounds = c(-Inf, Inf),
  stcsarg,
  scalefactor = 1,
  anisotropyid = "affine",
  anisotropyarg = list(phi1 = 1, phi2 = 1, phi12 = 0, theta = 0),
  dsid = "gauss",
  dsarg = NULL
)
}
\arguments{
\item{n}{number of time steps to simulate}

\item{spacepoints}{matrix (d x 2) of coordinates (e.g. longitude and
latitude) for d spatial locations (e.g. gauge stations)}

\item{margdist}{target marginal distribution}

\item{margarg}{list of marginal distribution arguments; consult the
documentation of the selected distribution for the required parameters}

\item{p0}{probability zero}

\item{distbounds}{distribution bounds (default \code{c(-Inf, Inf)})}

\item{stcsarg}{list of spatiotemporal correlation structure arguments;
consult the documentation of the selected structure for required parameters}

\item{scalefactor}{factor specifying the distance between pixel centres
(default 1)}

\item{anisotropyid}{spatial anisotropy ID (\code{"affine"} by default;
\code{"swirl"} or \code{"wave"} also available)}

\item{anisotropyarg}{list of arguments for \code{\link{anisotropyT}};
isotropic fields by default}

\item{dsid}{dependence structure ID (\code{"gauss"} by default;
\code{"student"}, \code{"bardossy"}, or \code{"bardossyF"})}

\item{dsarg}{argument for the dependence structure: \code{NULL} for
\code{"gauss"}, degrees of freedom for \code{"student"}, or parameter
\code{m} in \eqn{(-\infty, \infty)} for \code{"bardossy"}}
}
\value{
A matrix of class \code{c("matrix", "cosmosts")} with attribute
\code{STmodel} containing the fitted model components.
}
\description{
For more details see section 6 in Serinaldi and Kilsby (2018) and section 2.4
in Papalexiou and Serinaldi (2020).
}
\details{
\code{\link{generateMTSFast}} provides faster multivariate simulation than
\code{\link{generateMTS}} by exploiting circulant-embedding fast Fourier
transformation. This approach is feasible only for approximately separable
target spatiotemporal correlation functions.
\code{\link{generateMTSFast}} combines fitting and simulation in a single call.
Indicative CPU times (Windows 10 Pro x64, Intel Core i7-6700HQ, 32 GB RAM): \cr
d = 2500, n = 1000: ~58s \cr
d = 2500, n = 10000: ~160s \cr
d = 10000, n = 1000: ~2955s (~50 min) \cr
where \eqn{d} denotes the number of spatial locations.
}
\examples{
coord <- cbind(runif(4) * 30, runif(4) * 30)

sim <- generateMTSFast(
    n = 50,
    spacepoints = coord,
    p0 = 0.7,
    margdist = "paretoII",
    margarg = list(scale = 1,
                   shape = .3),
    stcsarg = list(scfid = "weibull",
                   tcfid = "weibull",
                   scfarg = list(scale = 20,
                                 shape = 0.7),
                   tcfarg = list(scale = 1.1,
                                 shape = 0.8))
)

}
\references{
Serinaldi, F., Kilsby, C.G. (2018). Unsurprising Surprises:
The Frequency of Record-breaking and Overthreshold Hydrological Extremes Under
Spatial and Temporal Dependence. Water Resources Research, 54(9), 6460-6487,
\doi{10.1029/2018WR023055}

Papalexiou, S.M., Serinaldi, F. (2020). Random Fields Simplified:
Preserving Marginal Distributions, Correlations, and Intermittency,
With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
e2019WR026331, \doi{10.1029/2019WR026331}
}
\seealso{
\code{\link{generateMTS}}, \code{\link{generateRFFast}},
\code{\link{fitVAR}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkTS.R
\name{checkTS}
\alias{checkTS}
\title{Check generated time series}
\usage{
checkTS(TS, distbounds = c(-Inf, Inf))
}
\arguments{
\item{TS}{a \code{cosmosts} object from \code{\link{generateTS}}, or a list
of numeric vectors, or a single numeric vector}

\item{distbounds}{numeric vector of length 2; distribution bounds (default
\code{c(-Inf, Inf)})}
}
\value{
An object of class \code{c("checkTS", "matrix")} with rows
\code{"expected"} and one row per simulated series, and columns for
\code{mean}, \code{sd}, \code{skew}, \code{p0}, \code{acf_t1},
\code{acf_t2}, \code{acf_t3}. Attributes \code{margdist}, \code{margarg},
and \code{p0} are attached for use by \code{\link{plot.checkTS}}.
}
\description{
Compares sample statistics of generated time series against theoretically
expected values.
}
\examples{

library(CoSMoS)

x <- generateTS(margdist = "burrXII",
                margarg = list(scale = 1,
                               shape1 = .75,
                               shape2 = .25),
                acsvalue = acs(id = "weibull",
                               t = 0:30,
                               scale = 10,
                               shape = .75),
                n = 1000, p = 30, p0 = .5, TSn = 5)

checkTS(x)

}
\seealso{
\code{\link{generateTS}}, \code{\link{plot.checkTS}},
\code{\link{moments}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advectionF.R
\name{advectionFuniform}
\alias{advectionFuniform}
\title{Uniform advection field}
\usage{
advectionFuniform(spacepoints, u, v)
}
\arguments{
\item{spacepoints}{vector of coordinates (2 x d), where d is the number of locations/grid points}

\item{u}{velocity component along the x axis}

\item{v}{velocity component along the y axis}
}
\description{
Provides an advection field with constant orthogonal (u and v) components at each grid point. This mimics rigid translation in a given direction according to the components u and v of the velocity vector.
}
\examples{

library(ggquiver)
library(ggplot2)
## specify coordinates
m = 25
aux <- seq(0, m - 1, length = m)
coord <- expand.grid(aux, aux)

af <- advectionFuniform(spacepoints = coord,
                       u = 2,
                       v = 6)

## visualize advection field
dta <- data.frame(lon = coord[ ,1], lat = coord[ ,2], u = af[ ,1], v = af[ ,2])
ggplot(dta, aes(x = lon, y = lat, u = u, v = v)) +
geom_quiver() +
theme_light()

}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stcs.r
\name{stcfgneiting16}
\alias{stcfgneiting16}
\title{Gneiting-16 SpatioTemporal Correlation Structure}
\usage{
stcfgneiting16(t, s, a, c, alpha, beta, nu, tau)
}
\arguments{
\item{t}{time lag}

\item{s}{spatial lag (distance)}

\item{a}{nonnegative scaling parameter of time}

\item{c}{nonnegative scaling parameter of space}

\item{alpha}{smoothness parameter of time. Valid range: \eqn{(0,1]}}

\item{beta}{space-time interaction parameter. Valid range: \eqn{[0,1]}}

\item{nu}{smoothness parameter of space. Valid range: \eqn{>0}}

\item{tau}{space-time interaction parameter. Valid range: \eqn{\ge 1} (for 2-dimensional fields)}
}
\description{
Provides spatiotemporal correlation structure function proposed by
Gneiting (2002; Eq.16 at p. 594).
}
\examples{

library(plot3D)

## specify grid of spatial and temporal lags
d <- 31
st <- expand.grid(0:(d - 1),
                  0:(d - 1))

## get the STCS
g16 <- stcfgneiting16(t = st[, 1],
                      s = st[, 2],
                      a = 1/50,
                      c = 1/10,
                      alpha = 1,
                      beta = 1,
                      nu = 0.5, tau = 1)

## visualize the STCS

g16.m <- matrix(g16,
                nrow = d)

persp3D(z = g16.m, x = 1: nrow(g16.m), y = 1:ncol(g16.m),
        expand = 1, main = "", scale = TRUE, facets = TRUE,
        xlab="Time lag", ylab = "Distance", zlab = "STCF",
        colkey = list(side = 4, length = 0.5), phi = 20, theta = 120,
        resfac = 5,  col= gg2.col(100))

}
\references{
Gneiting, T. (2002). Nonseparable, Stationary Covariance Functions
for Space-Time Data, Journal of the American Statistical Association,
97:458, 590-600, \doi{10.1198/016214502760047113}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seasonal.R
\name{analyzeTS}
\alias{analyzeTS}
\alias{reportTS}
\alias{simulateTS}
\title{Analyse, report, and simulate seasonal time series}
\usage{
analyzeTS(
  TS,
  season = "month",
  dist = "ggamma",
  acsID = "weibull",
  norm = "N1",
  n.points = 30,
  lag.max = 30,
  constrain = FALSE,
  opts = NULL
)

reportTS(aTS, method = "dist")

simulateTS(aTS, from = NULL, to = NULL)
}
\arguments{
\item{TS}{data frame or data table with columns \code{date} and \code{value}}

\item{season}{character; name of a date-component function
(e.g. \code{"month"}, \code{"week"})}

\item{dist}{character; name of the distribution to fit (e.g. \code{"norm"},
\code{"ggamma"})}

\item{acsID}{character; ACS identifier passed to \code{\link{fitACS}}}

\item{norm}{character; norm identifier — one of \code{"N1"}, \code{"N2"},
\code{"N3"}, \code{"N4"}}

\item{n.points}{integer; number of ECDF points used in the norm computation}

\item{lag.max}{integer; maximum lag for the empirical ACF}

\item{constrain}{logical; if \code{TRUE}, constrains \code{shape2} parameters
to \code{(0, 0.48)} to enforce finite upper tails}

\item{opts}{list of \code{nloptr} minimisation options}

\item{aTS}{an \code{analyzeTS} result object}

\item{method}{character; report type — \code{"dist"} for distribution fits,
\code{"acs"} for ACS fits, \code{"stat"} for descriptive statistics table}

\item{from}{POSIXct; start of simulation period (defaults to start of
observed series)}

\item{to}{POSIXct; end of simulation period (defaults to end of observed
series)}
}
\value{
\itemize{
\item \code{analyzeTS}: a list with elements \code{data}, \code{dfits},
\code{afits}, and attributes \code{season}, \code{dist}, \code{acsID},
\code{date}
\item \code{reportTS}: a \code{ggplot} object (\code{"dist"} or \code{"acs"}
method) or a \code{data.frame} (\code{"stat"} method)
\item \code{simulateTS}: a \code{data.table} with columns \code{date} and
\code{value}
}
}
\description{
\code{analyzeTS} automatically performs seasonal analysis, fits distributions
and correlation structures. \code{reportTS} visualises the fitted
distributions and correlation structures, or returns a table of fitted
parameters and descriptive statistics. \code{simulateTS} takes the result of
\code{analyzeTS} and generates synthetic realisations.
}
\details{
In practice, we typically want to simulate a natural process from observed
data. \code{analyzeTS} fits a marginal distribution and autocorrelation
structure for each season; \code{reportTS} lets you inspect the fit;
\code{simulateTS} generates synthetic time series with the same seasonal
statistical properties.

Recommended distributions by variable type:
\itemize{
\item \emph{precipitation / streamflow}: \code{ggamma}, \code{burrXII}, \code{burrIII}
\item \emph{relative humidity}: \code{beta}
\item \emph{temperature}: \code{norm}
}
}
\examples{

library(CoSMoS)

## Load data included in the package
data("precip")
\donttest{
## Fit seasonal ACSs and distributions to the data
a <- analyzeTS(precip)

reportTS(a, "dist")  ## seasonal distribution fits
reportTS(a, "acs")   ## seasonal ACS fits
reportTS(a, "stat")  ## descriptive statistics

## Simulate a time series of the same length
sim <- simulateTS(a)

precip[, id := "observed"]
sim[, id := "simulated"]
dta <- rbind(precip, sim)

ggplot(dta) +
  geom_line(aes(x = date, y = value)) +
  facet_wrap(~id, ncol = 1) +
  theme_classic()

## Simulate a time series of different length
sim <- simulateTS(a,
                  from = as.POSIXct("1978-12-01 00:00:00"),
                  to   = as.POSIXct("2008-12-01 00:00:00"))
}
\dontshow{
precip <- precip[between(date,
  as.POSIXct("1990-1-01", format("\%Y-\%m-\%d"), tz = "America/Regina"),
  as.POSIXct("1990-1-5",  format("\%Y-\%m-\%d"), tz = "America/Regina"))]
a <- analyzeTS(precip, opts = list("algorithm" = "NLOPT_LN_NELDERMEAD",
                                   "maxeval" = 10))
}

}
\seealso{
\code{\link{fitDist}}, \code{\link{fitACS}}, \code{\link{generateTS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitVAR.r
\name{fitVAR}
\alias{fitVAR}
\title{VAR model parameters to simulate correlated parent Gaussian random vectors and fields}
\usage{
fitVAR(
  spacepoints,
  p,
  margdist,
  margarg,
  p0,
  distbounds = c(-Inf, Inf),
  stcsid,
  stcsarg,
  scalefactor = 1,
  anisotropyid = "affine",
  anisotropyarg = list(phi1 = 1, phi2 = 1, phi12 = 0, theta = 0),
  advectionid = "uniform",
  advectionarg = list(u = 0, v = 0),
  dsid = "gauss",
  dsarg = NULL
)
}
\arguments{
\item{spacepoints}{it can be a numeric integer, which is interpreted as the side length m of the square field (m x m), or a matrix (d x 2) of coordinates (e.g. longitude and latitude) of d spatial locations (e.g. d gauge stations)}

\item{p}{order of VAR(p) model}

\item{margdist}{target marginal distribution of the field}

\item{margarg}{list of marginal distribution arguments. Please consult the documentation of the selected marginal distribution indicated in the argument \code{margdist} for the list of required parameters}

\item{p0}{probability zero}

\item{distbounds}{distribution bounds (default set to \code{c(-Inf, Inf)})}

\item{stcsid}{spatiotemporal correlation structure ID}

\item{stcsarg}{list of spatiotemporal correlation structure arguments. Please consult the documentation of the selected spatiotemporal correlation structure indicated in the argument \code{stcsid} for the list of required parameters}

\item{scalefactor}{factor specifying the distance between the centers of two pixels (default set to 1)}

\item{anisotropyid}{spatial anisotropy ID (\code{affine} by default, \code{swirl} or \code{wave})}

\item{anisotropyarg}{list of arguments characterizing the spatial anisotropy according to the syntax of the function \code{\link{anisotropyT}}. Isotropic fields by default}

\item{advectionid}{advection field ID (\code{uniform} by default, \code{rotation}, \code{spiral}, \code{spiralCE}, \code{radial}, or \code{hyperbolic})}

\item{advectionarg}{list of arguments characterizing the advection field according to the syntax of \code{advectionF}. No advection by default}

\item{dsid}{dependence structure ID (\code{gauss} by default, \code{student}, \code{bardossy}, and \code{bardossyF})}

\item{dsarg}{argument characterizing the dependence structure: \code{NULL} for \code{gauss} by default, number of degrees of freedom for \code{student} or parameter \code{m} in \code{(-Inf, Inf)} for \code{bardossy} (see Note section for more details)}
}
\description{
Compute VAR model parameters to simulate parent Gaussian random vectors with specified spatiotemporal correlation structure using the method described by Biller and Nelson (2003).
}
\details{
The fitting algorithm has \eqn{O(m*m)^3} complexity for a \eqn{(m*m)} field
or equivalently \eqn{O(d^3)} complexity for a \eqn{d}-dimensional vector.
Very large values of \eqn{(m*m)} (or \eqn{d}) and high order AR correlation
structures can be unpractical on standard machines.

Here, we give indicative CPU times for some settings, referring to a
Windows 10 Pro x64 laptop with Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz,
4-core, 8 logical processors, and 32GB RAM. \cr:
CPU time:\cr
d = 100 or m = 10, p = 1: ~  0.4s \cr
d = 900 or m = 30, p = 1: ~  6.0s \cr
d = 900 or m = 30, p = 5: ~ 47.0s \cr
d = 2500 or m = 50, p = 1: ~100.0s
}
\note{
While all the advection types can be applied to isotropic random fields,
anisotropic random fields require more care. We suggest combining \code{affine}
anysotropy with \code{uniform} advection, and \code{swirl} anisotropy
with \code{rotation} or \code{spiral} advection with the same rotation center..\cr
Concerning the Bardossy model, the increase of the parameter \code{m}
leads to a more and more symmetrical copula, and if \code{m} tends to \code{Inf},
then the copula converges to the Gaussian copula. The \code{bardossy} model is
characterized by lower tail dependence weaker than the upper tail dependence, while the
flipped Bárdossy dependence structure, denoted as  \code{bardossyF}, has lower tail
dependence stronger than the upper tail dependence.
See Bárdossy (2006) for more details about the properties and
parametrization of the multivariate Bardossy distribution
}
\examples{
## for multivariate simulation
coord <- cbind(runif(4)*30, runif(4)*30)

fit <- fitVAR(
  spacepoints = coord,
  p = 1,
  margdist ='burrXII',
  margarg = list(scale = 3,
                 shape1 = .9,
                 shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(scfid = "weibull",
                 tcfid = "weibull",
                 copulaarg = 2,
                 scfarg = list(scale = 20,
                               shape = 0.7),
                 tcfarg = list(scale = 1.1,
                               shape = 0.8))
)

dim(fit$alpha)
dim(fit$res.cov)

fit$m
fit$margarg
fit$margdist

## for random fields simulation
fit <- fitVAR(
  spacepoints = 10,
  p = 1,
  margdist ='burrXII',
  margarg = list(scale = 3, shape1 = .9, shape2 = .2),
  p0 = 0.8,
  stcsid = "clayton",
  stcsarg = list(scfid = "weibull", tcfid = "weibull",
                 copulaarg = 2,
                 scfarg = list(scale = 20, shape = 0.7),
                 tcfarg = list(scale = 1.1, shape = 0.8))
)

dim(fit$alpha)
dim(fit$res.cov)

fit$m
fit$margarg
fit$margdist

}
\references{
Bárdossy, A. (2006), Copula-based geostatistical models for groundwater
quality parameters, Water Resour. Res., 42, W11416, \doi{10.1029/2005WR004754}

Biller, B., Nelson, B.L. (2003). Modeling and generating multivariate
time-series input processes using a vector autoregressive technique.
ACM Trans. Model. Comput. Simul. 13(3), 211-237, \doi{10.1145/937332.937333}

Papalexiou, S.M. (2018). Unified theory for stochastic modelling of
hydroclimatic processes: Preserving marginal distributions, correlation structures,
and intermittency. Advances in Water Resources, 115, 234-252,
\doi{10.1016/j.advwatres.2018.02.013}

Papalexiou, S.M.,  Serinaldi, F. (2020). Random Fields Simplified:
Preserving Marginal Distributions, Correlations, and Intermittency,
With Applications From Rainfall to Humidity. Water Resources Research, 56(2),
e2019WR026331, \doi{10.1029/2019WR026331}

Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ar-models.R
\name{ARp}
\alias{ARp}
\title{Autoregressive model of order \emph{p}}
\usage{
ARp(margdist, margarg, acsvalue, actfpara, n, p = NULL, p0 = 0)
}
\arguments{
\item{margdist}{Character. Name of the target marginal distribution
(e.g. \code{"ggamma"}, \code{"paretoII"}). Must have a quantile function
\code{q<margdist>} available.}

\item{margarg}{Named list. Parameters of the marginal distribution passed
to \code{q<margdist>}.}

\item{acsvalue}{Numeric vector. Target autocorrelation structure starting
from lag 0 (i.e. \code{acsvalue[1] = 1}).}

\item{actfpara}{List returned by \code{\link{fitactf}}, containing the
fitted ACTF coefficients in \code{actfpara$actfcoef}.}

\item{n}{Positive integer. Length of the generated time series.}

\item{p}{Positive integer or \code{NULL}. AR model order. When
\code{NULL} (default), the order is chosen automatically as the number of
lags where the transformed ACS exceeds 0.01, capped at 1000.}

\item{p0}{Numeric in \eqn{[0, 1)}. Probability of zero values
(intermittency). Default \code{0}.}
}
\value{
A numeric vector of length \code{n} with the following attributes:
\describe{
\item{margdist}{Target marginal distribution name.}
\item{margarg}{Target marginal distribution parameters.}
\item{acsvalue}{Original (untransformed) target ACS.}
\item{p0}{Probability of zeros.}
\item{ar_coef}{AR(\eqn{p}) coefficients (Yule-Walker solution).}
\item{noise_sd}{Standard deviation of the Gaussian innovations.}
\item{gaussian}{The underlying Gaussian process (length \code{n}).}
\item{transformed_acs}{ACTF-transformed ACS used for the AR model.}
}
}
\description{
Generates a time series from an AR(\eqn{p}) model. The parent Gaussian
process is constructed to match a target autocorrelation structure via the
Yule-Walker equations, then transformed to a target marginal distribution
and optional intermittency (probability of zeros).
}
\seealso{
\code{\link{generateTS}}, \code{\link{actpnts}},
\code{\link{fitactf}}, \code{\link{acs}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seasonal.R
\name{seasonalAR}
\alias{seasonalAR}
\title{Seasonal AR model}
\usage{
seasonalAR(x, ACS, season = "month")
}
\arguments{
\item{x}{POSIXct vector of dates for which to generate the process}

\item{ACS}{named list of ACS vectors (one per season, starting at lag 0)}

\item{season}{character; name of a date-component function (default
\code{"month"})}
}
\value{
a \code{data.table} with columns \code{date}, \code{gauss}
(Gaussian process values), and \code{season} (season index)
}
\description{
Generates a seasonal Gaussian AR process over a vector of dates, one season
at a time, using precomputed ACS-derived AR coefficients.
}
\seealso{
\code{\link{simulateTS}}, \code{\link{analyzeTS}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acs.R
\name{ACSfunctions}
\alias{ACSfunctions}
\alias{acfburrXII}
\alias{acfparetoII}
\alias{acffgn}
\alias{acfweibull}
\title{Parametric autocorrelation structure functions}
\usage{
acfburrXII(t, scale, shape1, shape2)

acfparetoII(t, scale, shape)

acffgn(t, H)

acfweibull(t, scale, shape)
}
\arguments{
\item{t}{lag value (non-negative numeric).}

\item{scale}{scale parameter (must be > 0).}

\item{shape, shape1, shape2}{shape parameters (must be > 0).}

\item{H}{Hurst exponent for the FGN model (0.5 < H < 1).}
}
\value{
A numeric value (or \code{NaN} if parameters are invalid).
}
\description{
Parametric functions used internally by \code{\link{acs}} to evaluate
autocorrelation at lag \code{t}.  These are internal helpers and are not
part of the public API; call them via \code{acs(id, ...)}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ar-models.R
\name{AR1}
\alias{AR1}
\title{Autoregressive model of order 1}
\usage{
AR1(n, alpha, mean = 0, sd = 1)
}
\arguments{
\item{n}{Positive integer. Number of values to generate.}

\item{alpha}{Numeric in \eqn{(-1, 1)}. Lag-1 autocorrelation coefficient.}

\item{mean}{Numeric. Process mean. Default \code{0}.}

\item{sd}{Positive numeric. Process standard deviation. Default \code{1}.}
}
\value{
A numeric vector of length \code{n}.
}
\description{
Generates a Gaussian time series from a first-order autoregressive (AR(1))
model with specified lag-1 autocorrelation, mean, and standard deviation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moments.R
\name{sample.moments}
\alias{sample.moments}
\title{Sample moments}
\usage{
sample.moments(
  x,
  na.rm = FALSE,
  raw = TRUE,
  central = TRUE,
  coef = TRUE,
  order = 1:4
)
}
\arguments{
\item{x}{numeric vector of values}

\item{na.rm}{logical; strip \code{NA} values before computation?}

\item{raw}{logical; compute raw moments?}

\item{central}{logical; compute central moments?}

\item{coef}{logical; compute standardised coefficients (CV, skewness,
kurtosis)?}

\item{order}{integer vector; raw moment orders (default \code{1:4})}
}
\value{
a named list with zero or more of:
\describe{
\item{\code{m}}{raw moments}
\item{\code{mu}}{central moments}
\item{\code{coefficients}}{CV, skewness, kurtosis}
}
}
\description{
Computes raw moments, central moments, and standardised coefficients (CV,
skewness, kurtosis) from a numeric sample.
}
\examples{

library(CoSMoS)

x <- rnorm(1000)
sample.moments(x)

y <- rparetoII(1000, 10, .1)
sample.moments(y)

}
\seealso{
\code{\link{moments}}, \code{\link{checkTS}}
}
\keyword{moments}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitting.R
\name{fitACS}
\alias{fitACS}
\title{Autocorrelation structure fitting}
\usage{
fitACS(acf, ID, start = NULL, lag = NULL)
}
\arguments{
\item{acf}{numeric vector of autocorrelation function values from lag 0}

\item{ID}{character; ACS identifier (e.g. \code{"weibull"}, \code{"paretoII"})}

\item{start}{numeric vector of starting parameter values; if \code{NULL},
all parameters start at 1}

\item{lag}{integer; number of lags to use; if \code{NULL}, lags up to the
first value \eqn{\le 0.01} are used (or all lags if none drops below 0.01)}
}
\value{
An object of class \code{"fitACS"}: a named list of fitted ACS
parameters with attributes \code{ID} and \code{eACS} (empirical ACS used
for fitting).
}
\description{
Fits a parametric autocorrelation structure (ACS) to empirical ACF values
using Nelder-Mead optimisation with MSE criterion.
}
\examples{

x <- arima.sim(model = list(ar = 0.8), n = 1000)

acsfit <- fitACS(acf(x, plot = FALSE)$acf, "weibull", c(1, 1))

}
\seealso{
\code{\link{fitDist}}, \code{\link{plot.fitACS}}, \code{\link{acs}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disch.R
\docType{data}
\name{disch}
\alias{disch}
\title{Daily streamflow data data}
\format{
A data.table with 23315 rows and 2 variables:
\describe{
\item{date}{POSIXct format date/time}
\item{value}{daily avarage values}
}
}
\source{
The United States Geological Survey (USGS) National Water Information System (NWIS)
}
\usage{
disch
}
\description{
Station details
\itemize{
\item Name: Nassawango Creek near Snow Hill, Worcester County, Maryland, Hydrologic Unit 02080111
\item Network Id: , USGS 01485500
\item Latitude/Longitude: 38°13'44.1", 75°28'17.2"
\item Elevation: 11.49 ft above North American Vertical Datum of 1988.
\item Measurement unit: cubic feet per second
}
}
\details{
more details can be found \href{https://waterdata.usgs.gov/monitoring-location/USGS-01485500/}{here}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{actpntsB6}
\alias{actpntsB6}
\title{AutoCorrelation Transformed Points for Bardossy dependence structure}
\usage{
actpntsB6(margdist, margarg, m, p0 = 0)
}
\arguments{
\item{margdist}{target marginal distribution}

\item{margarg}{list of marginal distribution arguments}

\item{m}{mean of the parent Gaussian processes controlling asymmetry}

\item{p0}{probability of zero values}
}
\value{
A data frame with columns \code{rhoz} and \code{rhox}.
}
\description{
Evaluates the (rho_x, rho_z) mapping using Monte Carlo integration for the
Bardossy copula dependence structure.
}
\examples{

library(CoSMoS)

x <- actpntsB6(margdist = "paretoII",
               margarg  = list(scale = 1, shape = .3),
               m = 1,
               p0 = 0)
x

}
\references{
Bardossy, A. (2006). Copula-based geostatistical models for
groundwater quality parameters. Water Resources Research, 42(11),
\doi{10.1029/2005WR004754}
}
\seealso{
\code{\link{actpnts}}, \code{\link{generateMTSFast}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-methods.R
\name{plot.fitDist}
\alias{plot.fitDist}
\title{Plot method for \code{fitDist} objects}
\usage{
\method{plot}{fitDist}(x, ...)
}
\arguments{
\item{x}{a \code{fitDist} object returned by \code{\link{fitDist}}}

\item{...}{currently unused}
}
\value{
a \code{ggplot} object (invisibly returned; also printed)
}
\description{
Displays the empirical CDF against the fitted theoretical CDF on a
log-exceedance-probability scale.
}
\examples{

x <- fitDist(rnorm(1000), "norm", 30, "N1", FALSE)
plot(x)

}
\seealso{
\code{\link{fitDist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-internal.R
\name{erfc}
\alias{erfc}
\alias{inv.erfc}
\title{Complementary error function}
\usage{
erfc(x)

inv.erfc(x)
}
\arguments{
\item{x}{numeric vector}
}
\value{
numeric vector
}
\description{
Complementary error function

Inverse complementary error function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{acti}
\alias{acti}
\title{ACTI — autocorrelation transformation integral function}
\usage{
acti(x, y, dist, distarg, rhoz, p0)
}
\arguments{
\item{x}{x-plane value}

\item{y}{y-plane value}

\item{dist}{distribution name}

\item{distarg}{a list of distribution arguments}

\item{rhoz}{Gaussian correlation}

\item{p0}{probability of zero values}
}
\value{
numeric scalar — integrand value at (x, y)
}
\description{
Expression supplied to the double integral inside \code{\link{actpnts}}.
}
\examples{
acti(1, -1, "norm", list(mean = 0, sd = 1), .3, 0)

}
\seealso{
\code{\link{actpnts}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitting.R
\name{fitDist}
\alias{fitDist}
\title{Distribution fitting}
\usage{
fitDist(
  data,
  dist,
  n.points,
  norm,
  constrain,
  opts = list(algorithm = "NLOPT_LN_NELDERMEAD", xtol_rel = 1e-08, maxeval = 10000)
)
}
\arguments{
\item{data}{numeric vector of values to fit}

\item{dist}{character; distribution name (e.g. \code{"norm"}, \code{"ggamma"})}

\item{n.points}{integer; number of ECDF points used in norm computation}

\item{norm}{character; norm identifier — one of \code{"N1"} (ratio RMSE on
quantiles), \code{"N2"} (MSE on quantiles), \code{"N3"} (ratio RMSE on
probabilities), \code{"N4"} (MSE on probabilities)}

\item{constrain}{logical; if \code{TRUE}, constrains \code{shape2} parameters
to \code{(0, 0.48)} to enforce finite upper tails}

\item{opts}{list of \code{nloptr} minimisation options}
}
\value{
An object of class \code{"fitDist"}: a named list of fitted
distribution parameters with attributes \code{dist}, \code{edf} (empirical
CDF), and \code{nfo} (full \code{nloptr} output).
}
\description{
Fits a parametric distribution to data using the Nelder-Mead simplex
algorithm to minimise one of four fitting norms.
}
\examples{

x <- fitDist(rnorm(1000), "norm", 30, "N1", FALSE)
x

}
\seealso{
\code{\link{fitACS}}, \code{\link{plot.fitDist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-internal.R
\name{DHMgenSj}
\alias{DHMgenSj}
\title{Compute spectral values for DHM circulant-embedding simulation}
\usage{
DHMgenSj(acvf)
}
\arguments{
\item{acvf}{numeric vector of autocovariance values starting at lag 0}
}
\value{
numeric vector of \eqn{S_j} values (length = \code{length(acvf)})
}
\description{
Computes the one-sided spectral values \eqn{S_j} from an autocovariance
function (ACVF) vector using the Dietrich-Newsam-Haskard circulant-embedding
approach.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actf.R
\name{actfdiscrete}
\alias{actfdiscrete}
\title{ACTF auto-correlation transformation function for discrete distributions}
\usage{
actfdiscrete(rhox, b, c)
}
\arguments{
\item{rhox}{marginal correlation value}

\item{b}{1st parameter}

\item{c}{2nd parameter}
}
\value{
numeric vector of transformed Gaussian correlations
}
\description{
Provides transformation for discrete distributions, based on two parameters.
}
\examples{
actfdiscrete(.4, .2, 1)

}
\seealso{
\code{\link{actpnts}}, \code{\link{fitactf}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anisotropyT.R
\name{anisotropyT2}
\alias{anisotropyT2}
\title{Anisotropy transformation}
\usage{
anisotropyT2(id, arglist)
}
\arguments{
\item{id}{anisotropy type id}

\item{arglist}{list of additional arguments (vector of coordinates and parameters of the anisotropy transformations)}
}
\description{
Provides parametric functions that describe different types of planar deformation fields, including affine (rotation and stretching), and swirl-like deformation. For more details see Papalexiou et al.(2021) and references therein.
}
\references{
Papalexiou, S.M., Serinaldi, F., Porcu, E. (2021). Advancing
Space-Time Simulation of Random Fields: From Storms to Cyclones and Beyond.
Water Resources Research, 57, e2020WR029466, \doi{10.1029/2020WR029466}
}
\keyword{internal}
